package au.com.explodingsheep.diskDOM;



import org.w3c.dom.*;
import au.com.explodingsheep.diskDOM.identifier.*;
import au.com.explodingsheep.diskDOM.stringStore.*;


/**
 * My version of the CharacterData interface
 */

public class DefaultMyCharacterData extends DefaultMyNode implements MyCharacterData, HasValueIdentifier
{
  /* Points to the value of this Node as stored in the StringStore. */
  protected Identifier value = null;
  
  
  /**
   * Constructor.
   */
  
  public DefaultMyCharacterData( Identifier instanceID,
                           Identifier identifier, 
                           IdentifierIdentifier parent, 
                           IdentifierList children, 
                           String prefix,
                           String namespaceURI,
                           String nodeName,
                           int nodeType,
                           Identifier value,
                           Identifier document,
                           boolean readOnly )
  {
    super( instanceID,
            identifier, 
            parent, 
            children, 
            prefix,
            namespaceURI,
            nodeName,
            nodeType,
            document,
            readOnly );
    this.value = value;
  }


  /**
   * Confirms that the passed offset value (used in many CharacterData 
   * methods) is >= 0.
   * Also makes sure the offset is less than the length of the passed string.
   */
  
  protected void validateParams( String s, int offset, int count ) throws DOMException
  {
    if ( s != null )
    {
      int length = s.length();
      if ( offset < 0 )
      {
        throw new DOMException( DOMException.INDEX_SIZE_ERR, "Offset must be >= 0." );
      }
      else if ( count < 0 )
      {
        throw new DOMException( DOMException.INDEX_SIZE_ERR, "Count must be >= 0." );
      }
      else if ( offset > length )
      {
        throw new DOMException( DOMException.INDEX_SIZE_ERR, "Offset is greater than the length of the string." );
      }
      else if ( offset + count > length )
      {
        throw new DOMException( DOMException.INDEX_SIZE_ERR, "Offset + count is greater than the length of the string." );
      }
    }
    else
    {
      throw new DOMException( DOMException.INDEX_SIZE_ERR, "String is null." );
    }
  }
  
  
  /*------------------------- Overridden Node methods -------------------------*/
  
  
  /**
   * Returns the value of this node as text.
   */
  
  public String getNodeValue() throws DOMException
  {
    String result = null;
    try
    {
      if ( this.value != null )
      {
        result = super.stringStore.getString( this.value );
      }
    }
    catch ( StringStoreException sse )
    {
      throw new DOMException( super.DOM_EXCEPTION_NUM, sse.getMessage() );
    }
    return result;
  }  
  
  
  /**
   * Sets the value of this node as text.
   */
  
  public void setNodeValue( String nodeValue ) throws DOMException
  {
    super.confirmWriteAccess();
    
    try
    {
      super.stringStore.setString( this.value, nodeValue );
    }
    catch ( StringStoreException sse )
    {
      throw new DOMException( super.DOM_EXCEPTION_NUM, sse.getMessage() );
    }
  }


  /*------------------------ CharacterData methods -------------------------*/
  
  
  /**
   * Returns the value of this node as text.
   */
  
  public String getData() throws DOMException
  {
    String result = this.getNodeValue();
    return result;
  }
  
  
  /**
   * Sets the value of this node.
   */
  
  public void setData(String data) throws DOMException
  {
    /* This confirms we have write permission on this node and throws */
    /* an exception if not.                                            */
    super.confirmWriteAccess();

    this.setNodeValue( data );
  }
  
  
  /**
   * Returns the size of the string.
   */
  
  public int getLength()
  {
    int result = 0;
    String s = this.getData();
    if ( s != null )
    {
      result = s.length();
    }
    return result;
  }
  
  
  /**
   * Returns a sub-string of the value in this node.
   */
  
  public String substringData(int offset, int count) throws DOMException
  {
    String result = null;
    String s = this.getData();

    /* We don't want the count + offset to be greater than the length */
    /* of the string we want to replace to in, so "validateParams"    */
    /* will not fail.                                                  */
    int stringLength = s.length();
    if ( offset + count > stringLength )
      count = stringLength - offset;

    this.validateParams( s, offset, count );
//    this.validateCount( count );
    result = s.substring( offset, offset + count );
    return result;
  }
  
  
  /**
   * Appends a string to the current data in this node.
   */
  
  public void appendData(String arg) throws DOMException
  {
    /* This confirms we have write permission on this node and throws */
    /* an exception if not.                                            */
    super.confirmWriteAccess();
    
    String s = this.getData();
    s += arg;
    this.setData( s );
  }
  
  
  /**
   * Inserts a string into the current data in this node.
   */
  
  public void insertData(int offset, String arg) throws DOMException
  {
    /* This confirms we have write permission on this node and throws */
    /* an exception if not.                                            */
    super.confirmWriteAccess();

    String newString = null;
    String s = this.getData();
    this.validateParams( s, offset, 0 );
    String s1 = s.substring( 0, offset );
    String s2 = s.substring( offset );
    newString = s1 + arg + s2;
    this.setData( newString );
  }
  
  
  /**
   * Removes text from the current data in this node.
   */
  
  public void deleteData(int offset, int count) throws DOMException
  {
    /* This confirms we have write permission on this node and throws */
    /* an exception if not.                                            */
    super.confirmWriteAccess();

    String s = this.getData();

    /* We don't want the count + offset to be greater than the length */
    /* of the string we want to replace to in, so "validateParams"    */
    /* will not fail.                                                  */
    int stringLength = s.length();
    if ( offset + count > stringLength )
      count = stringLength - offset;

    this.validateParams( s, offset, count );

    String s1 = s.substring( 0, offset );
    String s2 = s.substring( offset + count );
    String newString = s1 + s2;
    this.setData( newString );
  }
  
  
  /**
   * Replaces text with new text.
   */
  
  public void replaceData(int offset, int count, String arg) throws DOMException
  {
    /* This confirms we have write permission on this node and throws */
    /* an exception if not.                                            */
    super.confirmWriteAccess();

    String s = this.getData();
    
    /* We don't want the count + offset to be greater than the length */
    /* of the string we want to replace to in, so "validateParams"    */
    /* will not fail.                                                  */
    int stringLength = s.length();
    if ( offset + count > stringLength )
      count = stringLength - offset;
      
    this.validateParams( s, offset, count );
    String s1 = s.substring( 0, offset );
    String s2 = s.substring( offset + count );
    String newString = s1 + arg + s2;
    this.setData( newString );
  }
  
  
  /*----------------------- HasValueIdentifier methods -----------------------*/
  
  
  /**
   * Returns the Identifier to the value of the node.
   */

  public au.com.explodingsheep.diskDOM.identifier.Identifier getValueIdentifier()
  {
    return this.value;
  }
}
