package au.com.explodingsheep.diskDOM;


import java.util.*;
import java.io.*;
import org.w3c.dom.*;
import au.com.explodingsheep.diskDOM.helpers.IntToBytes;
import au.com.explodingsheep.diskDOM.identifier.*;
import au.com.explodingsheep.diskDOM.nodeStore.*;
import au.com.explodingsheep.diskDOM.stringStore.*;
import au.com.explodingsheep.diskDOM.simpleFileSystem.*;


/**
 * This class is what DefaultMyNode and all its sub-classes
 * (and pretty well every other class) refers to via its
 * DOMImplementationIdentifier property. 
 * It holds the shared nodestore, and all the other resource
 * manager objects that the MyNodes need to operate.
 * 
 * All the resource managers (StringStore, NodeStore, IdentifierStore etc)
 * all read and store their data to a SimpleFileSystem passed in the
 * constructor.
 */

public class DefaultMyDOMImplementation implements MyDOMImplementation
{
  /* Where everything reads and writes its data to. */
  protected SimpleFileSystem sfs = null;
  
  /* The underlying object that reads and writes fixed sized */
  /* blocks of data to some permanent place, used by the     */
  /* SimpleFileSystem above.                                 */
  protected BlockIO blockIO = null;

  /* All the various resource managers. */  
  protected IdentifierStore identifierStore = null;
  protected NodeStore nodeStore = null;
  protected StringStore stringStore = null;
  protected IdentifierListStore identifierListStore = null;
  protected IdentifierGenerator identifierGenerator = null;

  /* Identifies this DOMImplementation instance. */
  protected Identifier domImplementationIdentifier = null;

  /* Cache used when reading and writing integers to disk. */
  protected byte[] fourBytes = new byte[4];

  /* Holds all the documents created by this DOMImplementation. */
  protected List cachedDocumentList = new ArrayList();

  /* We cannot load the already existing Documents in a parsed node */
  /* file in the constructor, so we must load them once during the  */
  /* "getDocument" method.                                          */
  protected boolean loadedDocuments = false;
  
    
  /**
   * Constructor.
   * 
   * Note that although we are passed an opened BlockIO by DocumentBuilder
   * we close that BlockIO in our "close" method.
   */
  
  public DefaultMyDOMImplementation( BlockIO blockIO ) throws MyDOMImplementationException
  {
    try
    {
      /* Create the SimpleFileSystem that stores all the   */
      /* data for everything - the nodestore, identifiers, */
      /* strings, everything.                               */
      this.blockIO = blockIO;
      this.sfs = new DefaultSimpleFileSystem( this.blockIO );


      this.identifierGenerator = new SimpleFileSystemIntIdentifierGenerator( this.sfs );

      
      /* If the sfs we just created is not empty - ie thie BlockIO */
      /* points to an already existing data file - then load the   */
      /* Identifier that uniquely identifies this                    */
      /* DOMImplementation.                                          */
      if ( this.sfs.doesFileExist( "DefaultMyDOMImplementation.identifier" ) == true )
      {
        /* Load the identifier. */
        SimpleFileSystemHandle handle = this.sfs.openFile( "DefaultMyDOMImplementation.identifier" );
        try
        {
          this.sfs.read( handle, this.fourBytes );
          int i = IntToBytes.bytesToInt( this.fourBytes );
          this.domImplementationIdentifier = new IntIdentifier( i );
        }
        finally
        {
          this.sfs.closeFile( handle );
        }
      }
      else
      {
        /* This is a brand new file so create a new Identifier. */
        this.domImplementationIdentifier = this.identifierGenerator.createIdentifier();
      }


      /* Create all the other resource managers. */      
      this.identifierStore = new SimpleFileSystemIdentifierStore( this.sfs );
      
      this.nodeStore = new ExperimentalNodeStore( this.sfs );
      
      this.stringStore = new SimpleFileSystemStringStore( this.sfs );

      this.identifierListStore = new SimpleFileSystemIntIdentifierListStore( this.sfs, this.domImplementationIdentifier );
    }
    catch ( Exception e )
    {
      throw new MyDOMImplementationException( e );
    }  
  }
  

  /**
   * Loads the MyDocuments that are in the parsed node file
   * and puts them in the "cachedDocumentList" List.
   */
  
  protected void loadDocumentNodesFromCurrentPDOM() throws MyDOMImplementationException
  {
    try
    {
      /* Load each Document that exists in the parsed data       */
      /* node file.                                               */
      /* To do this we first load a list of Document Identifiers */
      /* then create a Document for each.                         */
      if ( this.sfs.doesFileExist( "DefaultMyDOMImplementation.DocumentIdentifiers" ) == true )
      {
        SimpleFileSystemHandle handle = this.sfs.openFile( "DefaultMyDOMImplementation.DocumentIdentifiers" );
        try
        {
          int fileSize = this.sfs.size( handle );
          byte[] data = new byte[fileSize];
          this.sfs.read( handle, data );
          ByteArrayInputStream bais = new ByteArrayInputStream( data );
          ObjectInputStream ois = new ObjectInputStream( bais );
          List documentIdentifiers = ( List ) ois.readObject();
    
          /* Now create a MyDocument instance for each identifier. */
          int numDocIdentifiers = documentIdentifiers.size();
          for ( int loop = 0; loop < numDocIdentifiers; loop++ )
          {
            Identifier documentIdentifier = ( Identifier ) documentIdentifiers.get( loop );
            MyDocument document = ( MyDocument ) this.nodeStore.loadNode( documentIdentifier );
            this.addCachedDocumentToList( document );
          }
        }
        finally
        {
          this.sfs.closeFile( handle );
        }
      }
    }
    catch ( SimpleFileSystemException sfse)
    {
      throw new MyDOMImplementationException( sfse );
    }
    catch ( NodeStoreException nse )
    {
      throw new MyDOMImplementationException( nse );
    }
    catch ( IOException ioe )
    {
      throw new MyDOMImplementationException( ioe );
    }
    catch ( ClassNotFoundException cnfe )
    {
      throw new MyDOMImplementationException( cnfe );
    }
  }
  
  
  /*------------------------- MyDOMImplementation methods -----------------------*/


  /**
   * Closes all the resource managers and the BlockIO it
   * is all written to.
   */
  
  public void close() throws MyDOMImplementationException
  {
    try
    {
      if ( this.loadedDocuments == false )
      {
        this.loadDocumentNodesFromCurrentPDOM();
        this.loadedDocuments = true;
      }

      /* Write this DOMImplementation's identifier to disk. */
      byte[] fourBytes = new byte[4];
      SimpleFileSystemHandle handle = this.sfs.openFile( "DefaultMyDOMImplementation.identifier" );
      try
      {
        IntIdentifier identifier = ( IntIdentifier ) this.domImplementationIdentifier;
        int i = identifier.getValue();
        IntToBytes.intToBytes( i, fourBytes );
        this.sfs.write( handle, fourBytes );
      }
      finally
      {
        this.sfs.closeFile( handle );
      }

      /* Create a list of Identifiers for each Document in this */
      /* DOMImplementation and write them to disk.              */
      List documentIdentifiers = new ArrayList();
      int numDocuments = this.cachedDocumentList.size();
      for ( int loop = 0; loop < numDocuments; loop++ )
      {
        MyDocument document = ( MyDocument ) this.cachedDocumentList.get( loop );
        Identifier identifier = document.getIdentifier();
        documentIdentifiers.add( identifier );
      }
      
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      ObjectOutputStream oos = new ObjectOutputStream( baos );
      oos.writeObject( documentIdentifiers );
      oos.close();
      byte[] data = baos.toByteArray();
      handle = this.sfs.openFile( "DefaultMyDOMImplementation.DocumentIdentifiers" );
      try
      {
        this.sfs.write( handle, data );
      }
      finally
      {
        this.sfs.closeFile( handle );
      }

      
      this.nodeStore.close();
      this.identifierGenerator.close();
      this.identifierListStore.close();
      this.identifierStore.close();
      this.stringStore.close();
      this.sfs.close();

      this.blockIO.close();
    }
    catch ( Exception e )
    {
      throw new MyDOMImplementationException( e );
    }
  }
  
  
  /**
   * Returns the Identifier for this Instance.
   */
  
  public Identifier getIdentifier()
  {
    return this.domImplementationIdentifier;
  }  
  
  
  /**
   * Sets the Identifier to associate with this DOMImplementation.
   * All the MyNodes created will bet set with this identifier.
   */
  
  public void setIdentifier( Identifier identifier )
  {
    this.domImplementationIdentifier = identifier;
  }

  
  /**
   * Returns a NodeStore.
   */
  
  public NodeStore getNodeStore()
  {
    return this.nodeStore;
  }
  
  
  /**
   * Returns a IdentifierListManager.
   */
  
  public IdentifierListStore getIdentifierListStore()
  {
    return this.identifierListStore;
  }
  
  
  /**
   * Returns a StringStore.
   */
  
  public StringStore getStringStore()
  {
    return this.stringStore;
  }
  
  
  /**
   * Returns an IdentifierIdentifierStore.
   */
  
  public IdentifierStore  getIdentifierStore()
  {
    return this.identifierStore;
  }
  
  
  /**
   * Returns a IdentifierGenerator used to create Identifiers for
   * the Nodes in the Document.
   */
  
  public IdentifierGenerator getIdentifierGenerator()
  {
    return this.identifierGenerator;
  }
  
  
  
  /**
   * If a Document has been created by an instance of this class
   * then it will have been cached in "cachedDocumentList". If so it
   * can be returned here given its identifier. Returns null if
   * the document is not cached, possibly because a previously 
   * processed DOM file is being opened.
   */
  
  public MyDocument getCachedDocument( Identifier documentIdentifier )
  {
    MyDocument result = null;
    int numDocuments = this.cachedDocumentList.size();
    int loop = 0;
    while ( loop < numDocuments )
    {
      MyDocument document = ( MyDocument ) this.cachedDocumentList.get( loop );
      Identifier identifier = document.getIdentifier();
      if ( documentIdentifier.equals( identifier ) )
      {
        result = document;
        loop = numDocuments;
      }
      else
      {
        loop++;
      }
    }
    return result;
  }

  
  /**
   * Adds a Document to the list of cached Documents.
   */
  
  protected void addCachedDocumentToList( MyDocument document )
  {
    Identifier identifier = document.getIdentifier();
    if ( this.getCachedDocument( identifier ) == null )
    {
      this.cachedDocumentList.add( document );
    }
  }


  /**
   * Loads an already existing MyDocument from the parsed node data file
   * and returns it.
   * 
   * @param The name of the MyDocument to load.
   */
  
  public MyDocument getDocument( String documentName ) throws NodeStoreException, MyDOMImplementationException
  {
    MyDocument result = null;
    
    if ( this.loadedDocuments == false )
    {
      this.loadDocumentNodesFromCurrentPDOM();
      this.loadedDocuments = true;
    }
    
    int numDocuments = this.cachedDocumentList.size();
    int loop = 0;
    while ( loop < numDocuments )
    {
      MyDocument document = ( MyDocument ) this.cachedDocumentList.get( loop );
      String documentName2 = document.getDocumentName();
      if ( documentName.equals( documentName2 ) )
      {
        result = document;
        loop = numDocuments;
      }
      else
      {
        loop++;
      }
    }
    
    return result;
  }
  
    
  /**
   * Returns the names of all the MyDocuments in this
   * parsed node file being controlled by this DOMImplementation.
   */
  
  public String[] getDocumentNames() throws MyDOMImplementationException
  {
    if ( this.loadedDocuments == false )
    {
      this.loadDocumentNodesFromCurrentPDOM();
      this.loadedDocuments = true;
    }
    
    List documentNames = new ArrayList();
    int numDocuments = this.cachedDocumentList.size();
    for ( int loop = 0; loop < numDocuments; loop++ )
    {
      MyDocument document = ( MyDocument ) this.cachedDocumentList.get( loop );
      String documentName = document.getDocumentName();
      documentNames.add( documentName );
    }
    
    String[] result = ( String[] ) documentNames.toArray( new String[0] );
    return result;
  }

  
  /*-------------------------- DOMImplementation methods -------------------------*/


  /**
   * Test if the DOM implementation implements a specific feature. 
   * By default we support only core features.
   */
    
  public boolean hasFeature(String feature, String version)
  {
    boolean result = false;
    
    if ( feature.equalsIgnoreCase( "core" ) ) 
    {
//      if ( ( version.equalsIgnoreCase( "2.0" ) ) || ( version.equals( "" ) ) )
      {
        result = true;
      }
    }

    if ( feature.equalsIgnoreCase( "XML" ) ) 
    {
//      if ( ( version.equalsIgnoreCase( "1.0" ) )  || ( version.equals( "" ) ) )
      {
        result = true;
      }
    }
    
    return result;
  }


  /**
   * Creates an empty DocumentType node. Entity declarations 
   * and notations are not made available. Entity reference 
   * expansions and default attribute additions do not occur. 
   * It is expected that a future version of the DOM will 
   * provide a way for populating a DocumentType.
   * 
   * !!! Note that I am passing null as the Document.
   * Practically a DocumentType is created first by DocumentBuilder,
   * then passed to a Document.
   */
  
  public DocumentType createDocumentType(String qualifiedName, String publicId, String systemId) throws DOMException
  {
    DefaultMyDocumentType result = null;
    Exception e = null;
    try
    {
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierIdentifier parentIdentifierIdentifier = this.identifierStore.appendIdentifier( null );
      IdentifierList entityList = this.identifierListStore.createIdentifierList();
      IdentifierList notationList = this.identifierListStore.createIdentifierList();
  
      result = new DefaultMyDocumentType( this.domImplementationIdentifier, 
                                           null, 
                                           parentIdentifierIdentifier, 
                                           qualifiedName, 
                                           null,
                                           entityList,
                                           notationList );
      result.init();
  
      /* Store the new node in the NodeStore. */
      Identifier identifier = this.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
    }
    catch ( MyNodeException mne )
    {
      e = mne;
    }
    catch ( IdentifierListException ile )
    {
      e = ile;
    }
    catch ( IdentifierStoreException iise )
    {
      e = iise;
    }
    catch ( NodeStoreException nse )
    {
      e = nse;
    }

    return result;
  }
  

  /**
   * Creates a DOM Document object.
   */
  
  public Document createDocument(String namespaceURI, String qualifiedName, DocumentType doctype) throws DOMException
  {
    DefaultMyDocument result = null;
    try
    {
      IdentifierListStore ils = this.getIdentifierListStore();
      IdentifierList childIdentifierList = ils.createIdentifierList();
       IdentifierList attributeIdentifierList = ils.createIdentifierList();
       NodeStore nodeStore = this.getNodeStore();
       IdentifierGenerator ig = this.getIdentifierGenerator();

      Identifier docTypeIdentifier = null;
      if ( doctype != null )
      {
         MyDocumentType myDocType = ( MyDocumentType ) doctype;
         docTypeIdentifier = myDocType.getIdentifier();
      }

      /* Create the Document. */
      result = new DefaultMyDocument( this.domImplementationIdentifier, 
                                      null, 
                                      childIdentifierList, 
                                      attributeIdentifierList, 
                                      docTypeIdentifier,
                                      qualifiedName );
      result.init();

      /* Store the new node in the NodeStore. */
      Identifier identifier = this.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
      
      /* Add this new Document instance to cachedDocumentList".    */
      /* MyNode instances will go to here to look for an     */
      /* already existing Document instance.                 */
      this.addCachedDocumentToList( result );
    }
    catch ( Exception e )
    {
      throw new DOMException( DefaultMyNode.DOM_EXCEPTION_NUM, e.getMessage() );
    }
    return result;
  }
}
