package au.com.explodingsheep.diskDOM;


import org.w3c.dom.*;
import au.com.explodingsheep.diskDOM.identifier.*;
import au.com.explodingsheep.diskDOM.nodeStore.*;
import au.com.explodingsheep.diskDOM.stringStore.*;
import au.com.explodingsheep.diskDOM.dtdParser.*;


/**
 * Implementation of the MyDocument interface.
 * 
 * These are created by a DefaultMyDOMImplementation,
 * which in turn is created by a DefaultMyDocumentBuilder.
 */

public class DefaultMyDocument extends DefaultMyNode implements MyDocument
{
  /* This is the DOMException code thrown when "setDocumentName"   */
  /* is called when a document name is passed that already exists. */
  public static final short DOCUMENT_NAME_ALREADY_EXISTS = 1234;

  protected String namespaceURI = null;
  protected String qualifiedName = null;
  protected Identifier docTypeIdentifier = null;
  protected MyDocumentType docType = null;

  /* This is the name of a document used when retrieving a pre-parsed */
  /* Document from a DocumentBuilder.                                  */
  protected String documentName = null;
    
  
  /**
   * Constructor.
   */
  
  public DefaultMyDocument( Identifier domImplementationIdentifier,
                             Identifier identifier,
                             IdentifierList children,
                             IdentifierList attributes,
                             Identifier docTypeIdentifier,
                              String documentName ) throws DOMException
  {
    super( domImplementationIdentifier,
            identifier,
            null,
            children, 
            null,
            null,
            null,
            Node.DOCUMENT_NODE,
            null,
            false );
            
    super.documentIdentifier =  identifier;
    this.documentName = documentName;
    this.docTypeIdentifier = docTypeIdentifier;
  }


  /**
   * Constructor.
   */
  
  public DefaultMyDocument( Identifier domImplementationIdentifier,
                             Identifier identifier,
                             IdentifierList children,
                             IdentifierList attributes,
                             Identifier docTypeIdentifier ) throws DOMException
  {
    super( domImplementationIdentifier,
            identifier, 
            null,
            children, 
            null,
            null,
            null,
            Node.DOCUMENT_NODE,
            null,
            false );
            
    super.documentIdentifier =  identifier;
    this.docTypeIdentifier = docTypeIdentifier;
  }


  /**
   * Fills the passed NodeList with those Nodes whose names match
   * the name passed. Recurse into itself to fill the passed NodeList.
   */
  
  protected void fillNodeListWithNodesWithMatchingName( UnChangingNodeList nodeList, Node parent, String namespaceURI, String name )
  {
    if ( parent != null )
    {
      NodeList children = parent.getChildNodes();
      if ( children != null )
      {
        int numChildren = children.getLength();
        for( int loop = 0; loop < numChildren; loop++ )
        {
          boolean matches = false;
          Node node = children.item( loop );
          if ( node instanceof Element )
          {
            String nodeNamespaceURI = node.getNamespaceURI();
            String nodeName = node.getNodeName();
            
            /* !!! Note that the DOM core test              */
            /* "documentGetElementsByTagNameTotalLength"    */
            /* expects that a passed tag name of " " should */
            /* match all names, just like "*".              */
            if ( ( name.equals( "*" ) == false )
              && ( name.equals( " " ) == false ) )
            {
              if ( name.equals( nodeName ) == true )
              {
                if ( namespaceURI != null )
                {
                  if ( ( namespaceURI.equals( nodeNamespaceURI ) == true )
                    || ( namespaceURI.equals( "*" ) == true ) )
                  {
                    matches = true;
                  }
                }
                else
                {
                  matches = true;
                }
              }
            }
            else
            {
              matches = true;
            }
          }
          
          if ( matches == true )
          {
            nodeList.add( node );
          }
          
          /* Recurse into this node. */
          this.fillNodeListWithNodesWithMatchingName( nodeList, node, namespaceURI, name );
        }
      }
    }
  }


  /**
   * Confirms that the passed node name has no illegal characters
   * in it. If so throw an exception.
   */
  
  protected void validateTagName( String name ) throws DOMException
  {
    if ( name != null )
    {
      if ( name.equals( "" ) == false )
      {
        int length = name.length();
        int loop = 0;
        while ( loop < length )
        {
          char ch =  name.charAt( loop );
          if ( ( ch != '^' )
            && ( ch != '!' ) )
          {
            loop++;
          }
          else
          {
            throw new DOMException( DOMException.INVALID_CHARACTER_ERR, "Illegal character \"" + ch + "\" in node name \"" + name + "\"." );
          }      
        }
      }
      else
      {
        throw new DOMException( DOMException.INVALID_CHARACTER_ERR, "Node name cannot be \"\"." );
      }
    }
    else
    {
      throw new DOMException( DOMException.INVALID_CHARACTER_ERR, "Node name cannot be null." );
    }
  }


  /**
   * Extracts the localname from the passed qualified name.
   */
  
  protected String getLocalNameFromQualifiedName( String qualifiedName )
  {
    String result = null;
    int index = qualifiedName.indexOf( ':' );
    if ( index > 0 )
      result = qualifiedName.substring( index + 1 );
    else
      result = qualifiedName;
    return result;
  }
  

  /**
   * Extracts the prefix from the passed qualified name.
   */
  
  protected String getPrefixFromQualifiedName( String qualifiedName )
  {
    String result = null;
    int index = qualifiedName.indexOf( ':' );
    if ( index > 0 )
      result = qualifiedName.substring( 0, index );
    return result;
  }
  

  /*-------------------------- MyDocument methods --------------------------*/
  

  /**
   * Returns an identifier to the DocumentType associated
   * with this document.
   */
  
  public Identifier getDocumentTypeIdentifier()
  {
    return this.docTypeIdentifier;
  }
  
    
  /**
   * Associates a DocumentType with this Document.
   */
  
  public void setDocumentType( MyDocumentType documentType )
  {
    this.docType = documentType;
    if ( this.docType != null )
    {
      this.docTypeIdentifier = this.docType.getIdentifier();
    }
    else
    {
      this.docTypeIdentifier = null;
    }
  }


  /**
   * Creates an Entity object. This method is not part of the
   * Document interface, it is part of the MyDocument interface.
   */

  public Entity createEntity(String entityName, String publicID, String systemID, String notationName ) throws DOMException
  {
    DefaultMyEntity result = null;

    try
    {
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierList childIdentifierList = super.ilm.createIdentifierList();
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );


      result = new DefaultMyEntity( super.domImplementationIdentifier,
                              null,       // Identifier
                              childIdentifierList, 
                              null,       // prefix,
                              null,       // namespaceURI,
                              entityName,
                              super.identifier,
                              publicID,
                              systemID,
                              notationName );
      result.init();
                   
       /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( IdentifierListException ile )
    {
      super.throwException( ile );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }


  /**
   * Creates an Entity object. This method is not part of the
   * Document interface, it is part of MyDocument.
   */

  public Notation createNotation( String notationName, String publicID, String systemID ) throws DOMException
  {
    DefaultMyNotation result = null;

    try
    {
      /* First we must create the IdentifierList for the parent and children. */
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );

      result = new DefaultMyNotation( super.domImplementationIdentifier,
                              null,       // Identifier
                              parentIdentifierIdentifier,
                              prefix,
                              namespaceURI,
                              notationName,
                              publicID,
                              systemID,
                              super.identifier );
      result.init();
                   
       /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }
  

  /**
   * When a MyDocumentBuilder is closed, all the MyDocuments in it
   * are also closed. 
   * When a MyDocumentBuilder is re-opened, and its parsed node file
   * refers to an already existing file, then the MyDocuments in
   * that file can be retrieved. This document-name is how you 
   * retrieve each MyDocument.
   */
  
  public String getDocumentName()
  {
    return this.documentName;
  }
    
  
  /**
   * When a MyDocumentBuilder is closed, all the MyDocuments in it
   * are also closed. 
   * When a MyDocumentBuilder is re-opened, and its parsed node file
   * refers to an already existing file, then the MyDocuments in
   * that file can be retrieved. This document-name is how you 
   * retrieve each MyDocument.
   */

  public void setDocumentName( String documentName ) throws NodeStoreException, DOMException, MyDOMImplementationException
  {
    /* Make sure a Document with this name does not already exist. */
    boolean alreadyExists = false;
    MyDOMImplementation di = ( MyDOMImplementation ) this.getImplementation();
    MyDocument alreadyExistingDocument = di.getDocument( documentName );
    if ( alreadyExistingDocument != null )
    {
      Identifier alreadyExistingDocumentIdentifier = alreadyExistingDocument.getIdentifier();
      if ( alreadyExistingDocumentIdentifier.equals( this.identifier ) == false )
      {
        alreadyExists = true;
      }
    }
    
    if ( alreadyExists == false )
    {    
      this.documentName = documentName;
      this.saveNode();
    }
    else
    {
      throw new DOMException( DefaultMyDocument.DOCUMENT_NAME_ALREADY_EXISTS, "A document with the name \"" + documentName + "\" already exists." );
    }
  }


  /*-------------------- !!! The below methods need sorting --------------------*/


  /**
   * Marks the passed node and all its children as read-only.
   */
  
  protected void markNodesAsReadOnly( MyNode node )
  {
    if ( node != null )
    {
      node.setReadOnly( true );
      
      NamedNodeMap attributes = node.getAttributes();
      if ( attributes != null )
      {
        int numAttributes = attributes.getLength();
        for ( int loop = 0; loop < numAttributes; loop++ )
        {
          MyAttr attr = ( MyAttr) attributes.item( loop );
          attr.setReadOnly( true );
        }
      }
      
      NodeList children = node.getChildNodes();
      if ( children != null )
      {
        int numChildren = children.getLength();
        for ( int loop = 0; loop < numChildren; loop++ )
        {
          MyNode child = ( MyNode ) children.item( loop );
          this.markNodesAsReadOnly( child );
        }
      }
    }
  }

  
  /*------------------------- Overridden Node methods -----------------------*/

  
  /**
   * Returns "#document".
   */
  
  public String getNodeName()
  {
    return "#document";
  }
  
  
  /**
   * We set the Document Identifier to be the passed identifier before 
   * calling the super "setIdentifier" method. We do this as we are a 
   * Document so the Identifier has two purposes.
   * 
   * !!! Boy is this a bad comment.
   */
  
  public void setIdentifer( Identifier identifier )
  {
    super.documentIdentifier = identifier;
    super.setIdentifer( identifier );
  }
  
  
  /**
   * Returns the main parent document this node is part of.
   * Because this is a Document itself, it has no owner so
   * returns null.
   */
  
  public Document getOwnerDocument()
  {
    return null;
  }
  
  
  /*------------------------- Document methods ------------------------------*/
    
  
  /**
   * The Document Type Declaration (see DocumentType) associated 
   * with this document. For HTML documents as well as XML 
   * documents without a document type declaration this returns 
   * null. The DOM Level 2 does not support editing the Document 
   * Type Declaration. docType cannot be altered in any way, 
   * including through the use of methods inherited from the Node 
   * interface, such as insertNode or removeNode.
   */
  
  public DocumentType getDoctype() 
  {
    MyDocumentType result = null;
    
    if ( this.docType != null )
    {
      result = this.docType;
    }
    else
    if ( this.docTypeIdentifier != null )
    {
      try
      {
        result = ( MyDocumentType ) this.nodeStore.loadNode( this.docTypeIdentifier );
        this.docType = result;
      }
      catch ( NodeStoreException nse )
      {
        super.throwException( nse );
      }
    }
    
    return result;
  }
  
  
  /**
   * Returns the DOMImplementation object that created this Document.
   */
  
  public DOMImplementation getImplementation()
  {
    DOMImplementation result = MyDOMImplementationManager.getInstance( super.domImplementationIdentifier );
    return result;
  }
  
  
  /**
   * This is a convenience attribute that allows direct access 
   * to the child node that is the root element of the document. 
   * For HTML documents, this is the element with the tagName "HTML". 
   * 
   * !!! I don't know if what we are returning is correct.
   *     We want to return the first Element, not the first node,
   *     loop through all this Document's child nodes to find it.
   */
  
  public Element getDocumentElement()
  {
    Element result = null;
    NodeList children = super.getChildNodes();
    if ( ( children != null ) && ( children.getLength() > 0 ) )
    {
      int numNodes = children.getLength();
      int loop = 0;
      while ( loop < numNodes )
      {
        Node node = children.item( loop );
        if ( node instanceof Element )
        {
          result = ( Element ) node;
          loop = numNodes;
        }
        else
        {
          loop++;
        }
      }
    }
    return result;
  }
  
  
  /**
   * Creates an element of the type specified. Note that the 
   * instance returned implements the Element interface, so 
   * attributes can be specified directly on the returned 
   * object. In addition, if there are known attributes with 
   * default values, Attr nodes representing them are 
   * automatically created and attached to the element. 
   * To create an element with a qualified name and namespace 
   * URI, use the createElementNS method.
   */
  
  public Element createElement(String tagName) throws DOMException
  {
    DefaultMyElement result = null;
    
    this.validateTagName( tagName );

    try
    {
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierList childIdentifierList = super.ilm.createIdentifierList();
      IdentifierList attributeIdentifierList = super.ilm.createIdentifierList();
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );
      
      
      result = new DefaultMyElement( super.domImplementationIdentifier,
                              null,       // Identifier
                              parentIdentifierIdentifier, 
                              childIdentifierList, 
                              attributeIdentifierList,
                              null,       // prefix,
                              null,       // namespaceURI,
                              tagName,
                              Node.ELEMENT_NODE,
                              super.identifier,
                              false );
      result.init();
  
       /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
      
      /* If the element has attributes with default values then they */
      /* need to be created now.                                     */
      /* The MyDocumentType associated with this MyDocument has the  */
      /* Attr DTD information for this Element. Find it and create   */
      /* whatever Attrs have default values.                         */
      MyDocumentType docType = ( MyDocumentType ) this.getDoctype();
      if ( docType != null )
      {
        DTDElement dtdElement = docType.getDTDElement( tagName );
        if ( dtdElement != null )
        {
          DTDAttributeList attributeList = dtdElement.getDTDAttributes();
          if ( attributeList != null )
          {
            int numAttributes = attributeList.size();
            for ( int loop = 0; loop < numAttributes; loop++ )
            {
              DTDAttribute attribute = attributeList.getDTDAttribute( loop );
              String defaultValue = attribute.getDefaultValue();
              if ( defaultValue != null )
              {
                String attributeName = attribute.getName();
                MyAttr myAttr = ( MyAttr ) this.createAttribute( attributeName );
                myAttr.setDefaultValue( defaultValue );
                result.setAttributeNode( myAttr );
              }
            }
          }
        }
      }
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( IdentifierListException ile )
    {
      super.throwException( ile );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }
  

  /**
   * Creates a Text node given the specified string.   
   */
  
  public Text createTextNode(String data)
  {
    DefaultMyText result = null;

    try
    {
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );
      Identifier value = super.stringStore.appendString( data );
      
      
      result = new DefaultMyText( super.domImplementationIdentifier,
                           null,
                           parentIdentifierIdentifier, 
                           null,
                           null,
                           null,
                           Node.TEXT_NODE,
                           value,
                           super.identifier,
                           false );
      result.init();
                   
       /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( StringStoreException sse )
    {
      super.throwException( sse );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }


  /**
   * Creates a Text node given the specified string.   
   */
  
  public Comment createComment(String data)
  {
    DefaultMyComment result = null;

    try
    {
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );
      Identifier value = super.stringStore.appendString( data );
      
      
      result = new DefaultMyComment( super.domImplementationIdentifier,
                              null,       // Identifier
                               parentIdentifierIdentifier, 
                              null,       // prefix,
                              null,       // namespaceURI,
                              null,
                              Node.COMMENT_NODE,
                              value,
                              super.identifier );
      result.init();
                   
       /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( StringStoreException sse )
    {
      super.throwException( sse );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }


  /**
   * Creates a Text node given the specified string.   
   */
  
  public CDATASection createCDATASection(String data)
  {
    DefaultMyCDATASection result = null;

    try
    {
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );
      Identifier value = super.stringStore.appendString( data );
      
      
      result = new DefaultMyCDATASection( super.domImplementationIdentifier,
                                   null,
                                    parentIdentifierIdentifier, 
                                   null,
                                   null,
                                   null,
                                   Node.CDATA_SECTION_NODE,
                                   value,
                                   super.identifier );
      result.init();
                   
       /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( StringStoreException sse )
    {
      super.throwException( sse );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }


  /**
   * Creates a Text node given the specified string.   
   */
  
  public ProcessingInstruction createProcessingInstruction(String target, String data)
  {
    DefaultMyProcessingInstruction result = null;
    
    this.validateTagName( target );

    try
    {
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );
      Identifier value = super.stringStore.appendString( data );
      
      
      result = new DefaultMyProcessingInstruction( super.domImplementationIdentifier,
                                            null,
                                             parentIdentifierIdentifier, 
                                            null,
                                            null,
                                            target,
                                            Node.PROCESSING_INSTRUCTION_NODE,
                                            value,
                                             super.identifier,
                                            false );
      result.init();
                   
       /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( StringStoreException sse )
    {
      super.throwException( sse );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }


  /**
   * Creates a Text node given the specified string.   
   */
  
  public Attr createAttribute(String name)
  {
    DefaultMyAttr result = null;
    
    this.validateTagName( name );
    
    try
    {  
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierList childIdentifierList = super.ilm.createIdentifierList();
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );
      
      result = new DefaultMyAttr( super.domImplementationIdentifier,
                           null,
                           parentIdentifierIdentifier, 
                           childIdentifierList, 
                           null,
                           null,
                           name,
                           Node.ATTRIBUTE_NODE,
                           super.identifier,
                           false,
                           false );
      result.init();
                   
       /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( IdentifierListException ile )
    {
      super.throwException( ile );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }


  /**
   * Creates an EntityReference object. In addition, if the 
   * referenced entity is known, the child list of the 
   * EntityReference node is made the same as that of the 
   * corresponding Entity node.If any descendant of the 
   * Entity node has an unbound namespace prefix, the 
   * corresponding descendant of the created EntityReference 
   * node is also unbound; (its namespaceURI is null). 
   * The DOM Level 2 does not support any mechanism to 
   * resolve namespace prefixes. 
   */

  public EntityReference createEntityReference(String name) throws DOMException
  {
    DefaultMyEntityReference result = null;

    this.validateTagName( name );
        
    try
    {
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierList childIdentifierList = super.ilm.createIdentifierList();
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );
      
      result = new DefaultMyEntityReference( super.domImplementationIdentifier,
                                      null,       // Identifier
                                      parentIdentifierIdentifier, 
                                      childIdentifierList, 
                                      null,       // prefix,
                                      null,       // namespaceURI,
                                      name,
                                      Node.ENTITY_REFERENCE_NODE,
                                      super.identifier,
                                      false );
      result.init();
                   
       /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
      
      /* If the entity is known then create all the child nodes for it. */
      DocumentType docType = this.getDoctype();
      if ( docType != null )
      {
        NamedNodeMap entities = docType.getEntities();
        Entity entity = ( Entity ) entities.getNamedItem( name );
        if ( entity != null )
        {
          /* The entity contains one or more element nodes. */
          /* Clone the nodes and make them the child of the */
          /* entity reference.                              */
          /* Remember that the first node is an Entity,     */
          /* don't clone it, only its children.              */
          NodeList dataChildren = entity.getChildNodes();
          if ( dataChildren != null )
          {
            int numDataChildren = dataChildren.getLength();
            for ( int loop = 0; loop < numDataChildren; loop++ )
            {
              Node entityDataNode = dataChildren.item( loop );
              Node clonedNodes = entityDataNode.cloneNode( true );
              result.appendChild( clonedNodes );
            }
            
            this.markNodesAsReadOnly( result ); 
          }
          else
          {
            /* ??? What happens here? It is assumed an Entity */
            /* has children.                                  */
            super.throwException( "The entity \"" + name + "\" has no child nodes." );
          }
        }
        else
        {
          /* !!! What do we do here? We cannot find the entity */
          /* referred to in the XML file.                       */
        }
      }
      else
      {
        /* !!! What do we do here? No DocumentType so we can't */
        /* find the Entity.                                    */
      }
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( IdentifierListException ile )
    {
      super.throwException( ile );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }
  

 /**
  * Returns a NodeList of all descendant Elements with a given tag name, 
  * in the order in which they are encountered in a preorder traversal 
  * of this Element tree. 
  */
 
  public NodeList getElementsByTagName(String name) 
  {
    /* We need to recurse through all child elements looking for */
    /* nodes that have the passed name. Note that we must return */
    /* a NodeList.                                                */
    UnChangingNodeList result = new UnChangingNodeList();
    this.fillNodeListWithNodesWithMatchingName( result, this, null, name );
    return result;  
  }
  
 
 /**
  * Returns a NodeList of all the descendant Elements with a given 
  * local name and namespace URI in the order in which they are 
  * encountered in a preorder traversal of this Element tree. 
  */
 
  public NodeList getElementsByTagNameNS(String namespaceURI, String localName) 
  {
    /* We need to recurse through all child elements looking for */
    /* nodes that have the passed name. Note that we must return */
    /* a NodeList. Curses.                                       */
    /* To start with we will make a NodeList that does not        */
    /* automatically reflect changes to the underlying tree       */
    /* structure.                                                 */
    UnChangingNodeList result = new UnChangingNodeList();
    this.fillNodeListWithNodesWithMatchingName( result, this, namespaceURI, localName );
    return result;
  }
   
  
  /**
   * Imports the passed node and optionally its children
   * into this document.
   */
  
  public Node importNode(Node importedNode, boolean deep) throws DOMException
  {
    throw new DOMException( DOMException.NOT_SUPPORTED_ERR, "Feature not supported." );
  }


  /**
   * Creates an element of the given qualified name and namespace URI. 
   */
  
  public Element createElementNS(String namespaceURI, String qualifiedName) throws DOMException
  {
    DefaultMyElement result = null;

    this.validateTagName( qualifiedName );
    
    /* Extract the local name from the qualified name. */
    String localName = this.getLocalNameFromQualifiedName( qualifiedName );
    String prefix = this.getPrefixFromQualifiedName( qualifiedName );
        
    try
    {
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierList childIdentifierList = super.ilm.createIdentifierList();
      IdentifierList attributeIdentifierList = super.ilm.createIdentifierList();
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );
      
      
      result = new DefaultMyElement( super.domImplementationIdentifier,
                              null,
                              parentIdentifierIdentifier, 
                              childIdentifierList, 
                              attributeIdentifierList,
                              prefix,
                              namespaceURI,
                              localName,
                              Node.ELEMENT_NODE,
                              super.identifier,
                              false );
      result.init();
  
      /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( IdentifierListException ile )
    {
      super.throwException( ile );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }
  

  /**
   * Creates an attribute of the given qualified name and namespace URI. 
   */
    
  public Attr createAttributeNS(String namespaceURI, String qualifiedName) throws DOMException
  {
    DefaultMyAttr result = null;

    this.validateTagName( qualifiedName );

    /* Extract the local name from the qualified name. */
    String localName = this.getLocalNameFromQualifiedName( qualifiedName );
    String prefix = this.getPrefixFromQualifiedName( qualifiedName );
            
    try
    {
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierList childIdentifierList = super.ilm.createIdentifierList();
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );
      
      result = new DefaultMyAttr( super.domImplementationIdentifier,
                           null,
                           parentIdentifierIdentifier, 
                           childIdentifierList, 
                           prefix,
                           namespaceURI,
                           localName,
                           Node.ATTRIBUTE_NODE,
                           super.identifier,
                           false,
                           false );
      result.init();
                   
       /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( IdentifierListException ile )
    {
      super.throwException( ile );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }
  

  /**
   * Returns the Element whose ID is given by elementId. 
   * If no such element exists, returns null. Behavior 
   * is not defined if more than one element has this ID. 
   * The DOM implementation must have information that 
   * says which attributes are of type ID. Attributes 
   * with the name "ID" are not of type ID unless so 
   * defined. Implementations that do not know whether 
   * attributes are of type ID or not are expected to 
   * return null.   
   */
  
  public Element getElementById(String elementId)
  {
    return null;
  }
  
  
  /**
   * Creates an empty DocumentFragment object. 
   */
  
  public DocumentFragment createDocumentFragment()
  {
    DefaultMyDocumentFragment result = null;

    try
    {
      /* First we must create the IdentifierList for the children */
      /* and attributes.                                          */
      IdentifierList childIdentifierList = super.ilm.createIdentifierList();
      IdentifierIdentifier parentIdentifierIdentifier = super.iis.appendIdentifier( null );
      
      result = new DefaultMyDocumentFragment( super.domImplementationIdentifier,
                                       null,       // Identifier
                                       parentIdentifierIdentifier, 
                                       childIdentifierList, 
                                       null,       // prefix,
                                       namespaceURI,
                                       qualifiedName,
                                       Node.DOCUMENT_FRAGMENT_NODE,
                                       super.identifier );
      result.init();
                   
       /* Store the new node in the NodeStore. */
      Identifier identifier = super.nodeStore.appendNode( result );
      result.setIdentifer( identifier );
    }
    catch ( MyNodeException mne )
    {
      super.throwException( mne );
    }
    catch ( IdentifierListException ile )
    {
      super.throwException( ile );
    }
    catch ( IdentifierStoreException iise )
    {
      super.throwException( iise );
    }
    catch ( NodeStoreException nse )
    {
      super.throwException( nse );
    }
    
    return result;
  }


  /*--------------------- Overridden MyNode methods ----------------------------*/
  

  /**
   * Clones this node.
   * !!! Finish this.
   */
  
  public Node cloneThisNode()
  {
    throw new DOMException( ( short ) 0, "This method not supported." );
  }
}
