package au.com.explodingsheep.diskDOM;


import java.util.*;
import org.w3c.dom.*;
import au.com.explodingsheep.diskDOM.identifier.*;
import au.com.explodingsheep.diskDOM.dtdParser.*;


/**
 * My version of the DocumentType interface.
 * 
 * !!! We are handling the storing of DTDElements and DTDAttributeLists
 * badly. We are currently serializing the entire DocumentType to disk.
 * The problem is that if there is more than one DocumentType in existence
 * they won't stay synchronized - one will have newer information than
 * the other. The current solution is to assume that a DocumentType's
 * element and attribute data is only changed when a DTD is being
 * parsed, and there is only one one DocumentType instance in existence.
 * We need a more permanent solution.
 */

public class DefaultMyDocumentType extends DefaultMyNode implements MyDocumentType, java.io.Serializable
{
  protected String publicID = null;
  protected String systemID = null;
  protected String dtdName = null;
  
  /* Holds the list of identifiers to all the entities. */
  protected IdentifierList entityList = null;
  protected transient NamedNodeMap entityNamedNodeMap = null;

  /* Holds the list of identifiers to all the notations. */
  protected IdentifierList notationList = null;
  protected transient NamedNodeMap notationNamedNodeMap = null;

  /* The list of DTDElements we parse. */
  protected List elements = new ArrayList();

  /* List of DTDAttributeList instances we parse. */
   protected List attributeLists = new ArrayList();


  /**
   * Constructor.
   */
  
  public DefaultMyDocumentType( Identifier instanceID,
                          Identifier identifier, 
                          IdentifierIdentifier parent, 
                          String nodeName,
                          Identifier document,
                          IdentifierList entityList,
                          IdentifierList notationList )
  {
    super( instanceID,
            identifier, 
            parent, 
            null,
            null,
            null,
            nodeName,
            Node.DOCUMENT_TYPE_NODE,
            document,
            true );
            
    this.entityList = entityList;
    this.notationList = notationList;
  }


  /*---------------------- MyDocumentType methods -----------------------*/
  

  /**
   * Called by DTDParser when it parses an entity.
   */
  
  public void entityParsed( Entity entity ) 
  {
    String entityName = entity.getNodeName();
    if ( this.entityNamedNodeMap.getNamedItem( entityName ) == null )
    {
      Identifier identifier = ( ( MyNode ) entity ).getIdentifier();
      
      try
      {
        this.entityList.add( identifier );
      }
      catch ( IdentifierListException ile )
      {
        super.throwException( ile );
      }
    }
  }
  
  
  /**
   * Called by DTDParser when it parses a notation.
   */
  
  public void notationParsed( Notation notation )
  {
    String notationName = notation.getNodeName();
    if ( this.notationNamedNodeMap.getNamedItem( notationName ) == null )
    {
      Identifier identifier = ( ( MyNode ) notation ).getIdentifier();
      
      try
      {
        this.notationList.add( identifier );
      }
      catch ( IdentifierListException ile )
      {
        super.throwException( ile );
      }
    }
  }
  
  
  /**
   * Returns the IdentifierList that holds all the entities.
   */
  
  public IdentifierList getEntityIdentifierList()
  {
    return this.entityList;
  }
  
  
  /**
   * Returns the IdentifierList that holds all the notations.
   */
  
  public IdentifierList getNotationIdentifierList()
  {
    return this.notationList;
  }


  /**
   * Called by DTDParser when it parses an element declaration.
   * Add the DTDElement to the DTD elements list that is part 
   * of this class, then re-save this Node to disk. 
   * This serializes the node's data, including the dtd element 
   * and attribute information. Repeatedly saving this node every
   * time a DTDElement is parsed is not the fastest way to do
   * things but it only happens once when we are parsing a DTD
   * file.
   */
  
  public void dtdElementParsed( DTDElement dtdElement )
  {
    this.elements.add( dtdElement );
    super.saveNode();
  }

  
  /**
   * Called by DTDParser when it parses an attribute list declaration.
   * See the notes above for the "dtdElementParsed" method.
   */
  
  public void dtdAttributeListParsed( DTDAttributeList dtdAttributeList )
  {
    super.saveNode();
  }
    
  
  /**
   * Called when the name of a DTD has been parsed.
   * See the notes above for the "dtdElementParsed" method.
   */
  
  public void dtdNameParsed( String dtdName )
  {
    this.dtdName = dtdName;
    super.saveNode();
  }
  
  
  /**
   * Returns the number of parsed DTDElements.
   */
  
  public int getElementCount()
  {
    int result = this.elements.size();
    return result;
  }
  
  
  /**
   * Returns a DTDElement via its offset into the list.
   */
  
  public DTDElement getDTDElement( int offset )
  {
    DTDElement result = null;
    int numElements = this.getElementCount();
    if ( ( offset >= 0 ) && ( offset < numElements ) )
    {
      result = ( DTDElement ) this.elements.get( offset );
    }
    return result;
  }
  
    
  /**
   * Returns a DTDElement via its name.
   */
  
  public DTDElement getDTDElement( String elementName )
  {
    DefaultDTDElement result = null;
    int loop = 0;
    int numElements = this.elements.size();
    while ( loop < numElements )
    {
      DefaultDTDElement element = ( DefaultDTDElement ) this.elements.get( loop );
      String s = element.getName();
      if ( s.equals( elementName ) )
      {
        result = element;
        loop = numElements;
      }
      else
      {
        loop++;
      }
    }
    return result;
  }


  /*---------------------- MyNode overridden methods -----------------------*/    

  
  /**
   * If an instance of a MyNode is serialized this method
   * should be called after it has been de-serialized.
   */
  
  public void init() throws MyNodeException
  {
    super.init();
    
    Identifier instanceIdentifier = super.getDOMImplementationIdentifier();
    Identifier identifier = super.getIdentifier();

    /* "entityList" and "notationList" are IdentifierList instances. */
    /* IdentifierLists need to be "init"ed as well.                   */
    try
    {
      this.entityList.init();
      this.notationList.init();
    }
    catch ( IdentifierListException ile )
    {
      throw new MyNodeException( ile );
    }
    
    /* Create the named node maps for the entities and notations. */
    /* These named node maps are not serialized, only the         */
    /* IdentifierLists that hold the lists of nodes are.          */
    this.entityNamedNodeMap = new IdentifierListNamedNodeMap( super.nodeStore, this.entityList, identifier, null );
    this.notationNamedNodeMap = new IdentifierListNamedNodeMap( super.nodeStore, this.notationList, identifier, null );
  }
  
  
  /*------------------------- DocumentType methods ------------------------*/
  
  
  /**
   * The name of DTD; i.e., the name immediately following the DOCTYPE keyword. 
   */
   
  public String getName()
  {
    return this.dtdName;
  }
  
  
  /**
   * A NamedNodeMap containing the general entities, both external and internal, 
   * declared in the DTD. Parameter entities are not contained. Duplicates are 
   * discarded. For example in: 
   *   <!DOCTYPE ex SYSTEM "ex.dtd" [ <!ENTITY foo "foo"> <!ENTITY bar "bar"> 
   *   <!ENTITY bar "bar2"> <!ENTITY % baz "baz">]> <ex/>
   * the interface provides access to foo and the first declaration of bar but 
   * not the second declaration of bar or baz. Every node in this map also 
   * implements the Entity interface. 
   * The DOM Level 2 does not support editing entities, therefore entities 
   * cannot be altered in any way. 
   */
   
  public NamedNodeMap getEntities()
  {
    return this.entityNamedNodeMap;
  }
  
  
  /**
   * A NamedNodeMap containing the notations declared in the DTD. Duplicates 
   * are discarded. Every node in this map also implements the Notation interface. 
   * The DOM Level 2 does not support editing notations, therefore notations 
   * cannot be altered in any way. 
   */
  
  public NamedNodeMap getNotations()
  {
    return this.notationNamedNodeMap;
  }
  
  
  /**
   * The public identifier of the external subset. 
   */
   
  public String getPublicId()
  {
    return null;
  }
  
    
  /**
   * The system identifier of the external subset. 
   */
   
  public String getSystemId()
  {
    return null;
  }
  
  
  /**
   * The internal subset as a string, or null if there is none. 
   * This is does not contain the delimiting square brackets.
   * The actual content returned depends on how much information 
   * is available to the implementation. This may vary depending 
   * on various parameters, including the XML processor used to 
   * build the document. 
   */
   
  public String getInternalSubset()
  {
    return null;
  }

  
  /*------------------ Overridden org.w3c.dom.Node methods -----------------------*/

  
  /**
   * Returns the name of this node.
   */
  
  public String getNodeName()
  {
    return this.dtdName;
  }
}
