package au.com.explodingsheep.diskDOM;


import org.w3c.dom.*;
import au.com.explodingsheep.diskDOM.identifier.*;
import au.com.explodingsheep.diskDOM.dtdParser.*;


/**
 * Implementation of the MYElement interface.
 */

public class DefaultMyElement extends DefaultMyNode implements MyElement
{
  /* A live list of attributes of this node. When an attribute is */
  /* added or removed this list will automatically update itself. */
  protected IdentifierList attributes = null; 


  /**
   * Constructor.
   */
  
  public DefaultMyElement( Identifier instanceID,
                     Identifier identifier, 
                     IdentifierIdentifier parent, 
                     IdentifierList children, 
                     IdentifierList attributes,
                     String prefix,
                     String namespaceURI,
                     String nodeName,
                     int nodeType,
                     Identifier document,
                     boolean readOnly )
  {
    super( instanceID,
            identifier, 
            parent, 
            children, 
            prefix,
            namespaceURI,
            nodeName,
            nodeType,
            document,
            readOnly );
    this.attributes = attributes;
  }
  

  /**
   * Confirm that we can delete the passed node.
   * We must make sure that the attribute is not required,
   * or if it is then it has a default value so the attribute
   * is re-created.
   */
  
  protected void canWeDeleteAttribute( Attr attr ) throws DOMException
  {
    /* Get the DTD parser, which contains the DTD information for all the */
    /* elements and attributes in the DTD definition. It may be that      */
    /* there is no DTD parser as no DTD was declared, or that there is no */
    /* attribute DTD data for the passed attribute as it may be a new     */
    /* attribute created by the user. In either of these cases we allow   */
    /* the attribute to be deleted.                                       */
    DefaultMyDocument document = ( DefaultMyDocument) super.getOwnerDocument();
    MyDocumentType docType = ( MyDocumentType ) document.getDoctype();
    if ( docType != null )
    {
      String elementName = super.getNodeName();
      DTDElement dtdElement = docType.getDTDElement( elementName );
      DTDAttributeList dtdAttributes = dtdElement.getDTDAttributes();
      String attributeName = attr.getNodeName();
      DTDAttribute dtdAttribute = dtdAttributes.getDTDAttribute( attributeName );
      if ( dtdAttribute != null )
      {
        if ( dtdAttribute.isRequired() == true )
        {    
          String defaultValue = dtdAttribute.getDefaultValue();
          if ( defaultValue == null )
          {
            /* The attribute is required (and thus read only) so we */
            /* cannot delete it.                                     */
            throw new DOMException( DOMException.NO_MODIFICATION_ALLOWED_ERR, "This node is required." );
          }
        }
      }
    }
  }
  
  
  /**
   * Fills the passed NodeList with those Nodes whose names match
   * those passed. Recurse into itself to fill the NodeList.
   */
  
  protected void fillNodeListWithNodesWithMatchingName( UnChangingNodeList nodeList, Node parent, String namespaceURI, String name )
  {
    if ( parent != null )
    {
      NodeList children = parent.getChildNodes();
      if ( children != null )
      {
        int numChildren = children.getLength();
        for( int loop = 0; loop < numChildren; loop++ )
        {
          Node node = children.item( loop );
          
          /* If this node is an element then don't include child text or */
          /* attribute nodes.                                             */
          boolean ignoreNode = false;
          if ( this instanceof Element ) 
          {
            if ( ( node instanceof Text )
              || ( node instanceof Attr ) )
            {
              ignoreNode = true;
            }
          }
          
          if ( ignoreNode == false )
          {
            String nodeNamespaceURI = node.getNamespaceURI();
            String nodeName = node.getNodeName();
            boolean matches = false;
            if ( name.equals( "*" ) == false )
            {
              if ( name.equals( nodeName ) == true )
              {
                if ( namespaceURI != null )
                {
                  if ( ( namespaceURI.equals( nodeNamespaceURI ) == true )
                    || ( namespaceURI.equals( "*" ) == true ) )
                  {
                    matches = true;
                  }
                }
                else
                {
                  matches = true;
                }
              }
            }
            else
            {
              matches = true;
            }
            
            if ( matches == true )
            {
              nodeList.add( node );
            }
            
            /* Recurse into this node. */
            this.fillNodeListWithNodesWithMatchingName( nodeList, node, namespaceURI, name );
          }
        }
      }
    }
  }

  
  /**
   * Returns the list if attribute Identifiers for this node.
   */
  
  public IdentifierList getAttributeIdentifiers()
  {
    return this.attributes;
  }


  /*------------------------ MyElement methods ----------------------------*/

    
  /**
   * Creates any attributes that have default values and don't already
   * exist.
   */
  
  public void createDefaultAttributes()
  {
    String elementName = super.getNodeName();
    DefaultMyDocument document = ( DefaultMyDocument) super.getOwnerDocument();
    MyDocumentType docType = ( MyDocumentType ) document.getDoctype();
    DTDElement dtdElement = docType.getDTDElement( elementName );
    if ( dtdElement != null )
    {
      DTDAttributeList attributeList = dtdElement.getDTDAttributes();
      if ( attributeList != null )
      {
        int numAttributes = attributeList.size();
        for ( int loop = 0; loop < numAttributes; loop++ )
        {
          DTDAttribute dtdAttribute = attributeList.getDTDAttribute( loop );
          String attributeName = dtdAttribute.getName();
          if ( this.getAttributeNode( attributeName ) == null )
          {
            String defaultValue = dtdAttribute.getDefaultValue();
            if ( defaultValue != null )
            {
              MyAttr myAttr = ( MyAttr) document.createAttribute( attributeName );
              myAttr.setDefaultValue( defaultValue );
              this.setAttributeNode( myAttr );
            }
          }
        }
      }
    }
  }

  
  /*-------------------------- Overridden Node methods ---------------------------*/
  

  /**
   * Returns a NamedNodeMap containing the attributes of 
   * this node (if it is an Element) or null otherwise.
   */
  
  public NamedNodeMap getAttributes()
  {
    NamedNodeMap result = null;

    // !!! Potential bug here. We are passing the Parent identifier, 
    //    not the IdentifierIdentifier.
    Identifier parentIdentifier = this.getParentIdentifier();

    Document ownerDocument = super.getOwnerDocument();
    result = new IdentifierListNamedNodeMap( this.nodeStore, this.attributes, parentIdentifier, ownerDocument );

     return result;
  }


  /**
   * Returns whether this node (if it is an element) has any attributes.   
   */
  
  public boolean hasAttributes()
  {
    boolean result = false;
    try
    {
      int numAttributes = this.attributes.size();
      result = ( numAttributes > 0 );
    }
    catch ( IdentifierListException ile )
    {
      throw new DOMException( DefaultMyNode.DOM_EXCEPTION_NUM, ile.getMessage() );
    }
     return result;
  }
  
  
  /**
   * Clones this element node.
   */
  
  public Node cloneNode(boolean deep)
  {
    Element result = null;
    
    if ( deep == true )
    {
      result = ( Element ) super.cloneNode( true );
    }
    else
    {
      Document document = super.getOwnerDocument();
      String name = super.getNodeName();
      result = ( Element ) document.createElement( name );
      
      /* Also clone the element's attributes but not its text value. */
      NamedNodeMap attributes = this.getAttributes();
      int numAttributes = attributes.getLength();
      for ( int loop = 0; loop < numAttributes; loop++ )
      {
        Attr attribute = ( Attr ) attributes.item( loop );
        Attr clonedAttribute = ( Attr ) attribute.cloneNode( false );
        result.setAttributeNode( clonedAttribute );
      }
    }
    
    return result;
  }


  /*-------------------------- Overridden MyNode methods ------------------------*/
  
  
  /**
   * If an instance of a MyNode is serialized this method
   * should be called after it has been de-serialized.
   */
  
  public void init() throws MyNodeException
  {
    super.init();

    /* Initialize the Attributes list. */
    try
    {
      if ( this.attributes != null )
      {
        this.attributes.init();
      }
    }
    catch ( IdentifierListException ile )
    {
      throw new MyNodeException( ile );
    }
  }


  /**
   * Returns the value of this node.
   * We return null here - sub-classes like Attr and Element 
   * should handle this appropriately.
   */
  
  public String getNodeValue() throws DOMException
  {
    return null;
  }  


  /**
   * Sets the value of this node as text.
   * We do nothing here - sub-classes like Attr and Element 
   * should handle this appropriately.
   */
  
  public void setNodeValue(String nodeValue) throws DOMException
  {
  }


  /*----------------------------- Element methods --------------------------------*/
  

 /**
  * Retrieves an attribute node by name. 
  */
 
  public Attr getAttributeNode(String name) 
  {
    Attr result = null;
    
    /* Load each attribute node and compare its name to that passed. */
    try
    {
      int numAttributes = this.attributes.size();
      int loop = 0;
      while ( loop < numAttributes )
      {
        Identifier i = this.attributes.get( loop );
        MyNode node = super.loadNode( i );
        String s = node.getNodeName();
        if ( s.equals( name ) == true )
        {
          result = ( Attr ) node;
          loop = numAttributes;
        }
        else
        {
          loop++;
        }
      }
    }
    catch ( IdentifierListException ile )
    {
      super.throwException( ile );
    }
    return result;
  }
  

 /**
  * Retrieves an attribute value by name. 
  */
 
  public String getAttribute(String name) 
  {
    String result = null;    
    Attr attr = this.getAttributeNode( name );
    if ( attr != null )
    {
      result = attr.getValue();
    }
    return result;
  }
   
 
 /**
  * Retrieves an Attr node by local name and namespace URI. 
  */
 
  public Attr getAttributeNodeNS(String namespaceURI, String localName) 
  {
    Attr result = null;
    
    /* Load each attribute node and compare its name to that passed. */
    try
    {
      int numAttributes = this.attributes.size();
      int loop = 0;
      while ( loop < numAttributes )
      {
        Identifier i = this.attributes.get( loop );
        MyNode node = super.loadNode( i );
        String s1 = node.getNodeName();
        String s2 = node.getNamespaceURI();
        if ( ( s1.equals( localName ) == true )
          && ( s2.equals( namespaceURI ) == true ) )
        {
          result = ( Attr ) node;
          loop = numAttributes;
        }
        else
        {
          loop++;
        }
      }
    }
    catch ( IdentifierListException ile )
    {
      super.throwException( ile );
    }
    
    return result;
  }
  
  
 
 /**
  * Retrieves an attribute value by local name and namespace URI. 
  */
 
  public String getAttributeNS(String namespaceURI, String localName) 
  {
    String result = null;    
    Attr attr = this.getAttributeNodeNS( namespaceURI, localName );
    if ( attr != null )
    {
      result = attr.getValue();
    }
    return result;
  }
  
 
 /**
  * Returns a NodeList of all descendant Elements with a given tag name, 
  * in the order in which they are encountered in a preorder traversal 
  * of this Element tree. 
  */
 
  public NodeList getElementsByTagName(String name) 
  {
    /* We need to recurse through all child elements looking for */
    /* nodes that have the passed name. Note that we must return */
    /* a NodeList. Curses.                                       */
    UnChangingNodeList result = new UnChangingNodeList();
    this.fillNodeListWithNodesWithMatchingName( result, this, null, name );
    return result;  
  }
  
 
 /**
  * Returns a NodeList of all the descendant Elements with a given 
  * local name and namespace URI in the order in which they are 
  * encountered in a preorder traversal of this Element tree. 
  */
 
  public NodeList getElementsByTagNameNS(String namespaceURI, String localName) 
  {
    /* We need to recurse through all child elements looking for */
    /* nodes that have the passed name. Note that we must return */
    /* a NodeList. Curses.                                       */
    /* To start with we will make a NodeList that does not        */
    /* automatically reflect changes to the underlying tree       */
    /* structure.                                                 */
    UnChangingNodeList result = new UnChangingNodeList();
    this.fillNodeListWithNodesWithMatchingName( result, this, namespaceURI, localName );
    return result;
  }
  
 
 /**
  * The name of the element. 
  */
 
  public String getTagName()
  {
    String result = super.getNodeName();
    return result;
  }
  
 
 /**
  * Returns true when an attribute with a given name is specified on this element or has a default value, false otherwise. 
  */
 
  public boolean hasAttribute(String name) 
  {
    Attr attr = this.getAttributeNode( name );
    boolean result = ( attr != null );
    return result;
  }
    
 
 /**
  * Returns true when an attribute with a given local name and namespace URI is specified on this element or has a default value, false otherwise. 
  */
 
  public boolean hasAttributeNS(String namespaceURI, String localName) 
  {
    Attr attr = this.getAttributeNodeNS( namespaceURI, localName );
    boolean result = ( attr != null );
    return result;
  }
   
 
 /**
  * Removes an attribute by name. 
  */
 
  public void removeAttribute(String attributeToRemoveName) 
  {
    super.confirmWriteAccess();

    MyAttr attr = ( MyAttr ) this.getAttributeNode( attributeToRemoveName );
    if ( attr != null )
    {
      this.canWeDeleteAttribute( attr );

      /* We are allowed to delete the node. */
      try
      {
        Identifier i = attr.getIdentifier();
        this.attributes.remove( i );
        attr.setParentIdentifier( null );

         this.createDefaultAttributes();
      }
       catch ( IdentifierListException ile )
       {
         super.throwException( ile );
       }
    }
  }
  
 
 /**
  * Removes the specified attribute node. 
  */
 
  public Attr removeAttributeNode(Attr oldAttr) 
  {
    super.confirmWriteAccess();
    
    this.canWeDeleteAttribute( oldAttr );

    MyNode myNode = ( MyNode ) oldAttr;
     Identifier i = myNode.getIdentifier();
     try
     {
       int index = this.attributes.indexOf( i );
       if ( index >= 0 )
       {
         this.attributes.remove( index );
         myNode.setParentIdentifier( null );
         
         this.createDefaultAttributes();
       }
       else
       {
         String attributeName = oldAttr.getNodeName();
         throw new DOMException( DOMException.NOT_FOUND_ERR, "Attribute \"" + attributeName + "\" not found." );
       }
    }
     catch ( IdentifierListException ile )
     {
       super.throwException( ile );
     }
     return oldAttr;
  }
  
 
 /**
  * Removes an attribute by local name and namespace URI. 
  */
 
  public void removeAttributeNS(String namespaceURI, String localName) 
  {
    super.confirmWriteAccess();
    
    Attr attr = this.getAttributeNodeNS( namespaceURI, localName );
    if ( attr != null )
    {
      this.canWeDeleteAttribute( attr );
      
      MyNode myNode = ( MyNode ) attr;
      Identifier i = myNode.getIdentifier();
      try
      {
        this.attributes.remove( i );
        myNode.setParentIdentifier( null );
        
        this.createDefaultAttributes();
      }
       catch ( IdentifierListException ile )
       {
         super.throwException( ile );
       }
    }
  }
  
 
 /**
  * Adds a new attribute. 
  */
 
  public void setAttribute(String name, String value) 
  {
    super.confirmWriteAccess();

    /* If an attribute with this already exists then use it, otherwise */
    /* create a new one.                                               */
    MyAttr attr = ( MyAttr) this.getAttributeNode( name );
    if ( attr != null )
    {
      attr.setNodeValue( value );
    }
    else
    {
      /* Create a new attribute and add it to the list of attributes */
      /* for this node.                                               */
      Document document = super.getOwnerDocument();
       attr = ( MyAttr) document.createAttribute( name );
       attr.setNodeValue( value );
       Identifier i = attr.getIdentifier();
       try
       {
         this.attributes.add( i );
       }
       catch ( IdentifierListException ile )
       {
         super.throwException( ile );
       }
    }
  }
     
  
 /**
  * Adds a new attribute node. 
  */
 
  public Attr setAttributeNode(Attr newAttr) 
  {
    super.confirmWriteAccess();

    Attr result = null;
    if ( newAttr != null )
    {
      /* It may be that the passed attribute is already          */
      /* to this element. This is legal, the node is returned    */
      /* nothing else changes.                                   */
      boolean attributeIsAlreadyConnected = false;
      String newAttributeName = newAttr.getNodeName();
      result = this.getAttributeNode( newAttributeName );
      if ( result != null )
      {
        attributeIsAlreadyConnected = result.equals( newAttr );
      }
      if ( attributeIsAlreadyConnected == false )
      {
        /* Make sure the passed attribute is not already connected */
        /* to another node.                                         */
         MyNode newAttrAsMyNode = ( MyNode ) newAttr;
        Identifier parentIdentifier = newAttrAsMyNode.getParentIdentifier();
        if ( parentIdentifier == null )
        {
          /* Make sure the parent document of the attribute we want to append */
          /* is the same as the parent document of this element.              */
          Identifier attrDocumentIdentifier = newAttrAsMyNode.getDocumentIdentifier();
          if ( attrDocumentIdentifier.equals( this.documentIdentifier ) )
          {
            /* If an attribute already exists with the same name then */
            /* remove it. We must also return that old attribute when */
            /* we exit this method.                                    */
            if ( result != null )
            {
              /* Remove this old attribute without re-creating it if it */
              /* has a default value.                                    */
              Identifier i = ( ( MyNode ) result ).getIdentifier();
              try
              {
                this.attributes.remove( i );
              }
               catch ( IdentifierListException ile )
               {
                 super.throwException( ile );
               }
              ( ( MyNode ) result ).setParentIdentifier( null );
            }
          
            /* Add the new attribute to the list of attributes */
            /* for this node.                                    */
             Identifier i = ( ( MyAttr) newAttr ).getIdentifier();
             try
             {
               /* Add the new attribute to the list for this element. */
               this.attributes.add( i );
               
               /* Set the new attribute's parent to this node. */
               newAttrAsMyNode.setParentIdentifier( this.identifier );
             }
             catch ( IdentifierListException ile )
             {
               super.throwException( ile );
             }
          }
          else
          {
            throw new DOMException( DOMException.WRONG_DOCUMENT_ERR, "Attribute's owner Document is different to element's owner Document." );
          }
        }
        else
        /* It is not an error if the attribute is already connected */
        /* to this element.                                          */
        if ( parentIdentifier.equals( this.identifier ) == false )
        {
          throw new DOMException( DOMException.INUSE_ATTRIBUTE_ERR, "Attribute is already connected to another parent." );
        }
      }
    }
    else
    {
      throw new DOMException( DOMException.INVALID_MODIFICATION_ERR, "Passed attribute is null." );
    }
     return result;
  }    
  
 
 /**
  * Adds a new attribute. 
  */
 
  public Attr setAttributeNodeNS(Attr newAttr) 
  {
    super.confirmWriteAccess();

    Attr result = this.setAttributeNode( newAttr );
    return result;
  }
  
 
 /**
  * Adds a new attribute. 
  */
 
  public void setAttributeNS(String namespaceURI, String qualifiedName, String value)
  {
    /* !!! For now call the same method without namespace support. */
    this.setAttribute( qualifiedName, value );
  }
}
