package au.com.explodingsheep.diskDOM;


import org.w3c.dom.*;
import au.com.explodingsheep.diskDOM.helpers.ErrorHandler;
import au.com.explodingsheep.diskDOM.identifier.*;
import au.com.explodingsheep.diskDOM.nodeStore.*;
import au.com.explodingsheep.diskDOM.stringStore.*;


/**
 * My implementation of a MyNode. This class does most the work
 * of looking up and saving child nodes and attributes.
 * 
 * Methods that are used by the majority of sub-classes are 
 * implemented here - things like "getParentNode" and "getChildrenNodes".
 * Others like "getAttributes" return null and are implemented by
 * the MyAttr class only.
 */

public class DefaultMyNode implements MyNode
{
  /* Identifier that points to the DOMImplementation that created */
  /* the Document that in turn created this Node. When a MyNode   */
  /* is de-serialized it uses this Identifier to retrieve the     */
  /* DOMImplementation it needs to get at the various resource    */
  /* managers it needs to do its job.                              */
  protected Identifier domImplementationIdentifier = null;
  
  /* Identifies this node. */
  protected Identifier identifier = null;
  
  /* Identifies the parent of this node. */
  protected IdentifierIdentifier parent = null; 
  
  /* A live list of child nodes to this node. When a child node is */
  /* added or removed this list will automatically update itself.   */
  protected IdentifierList children = null;

  /* Manager the above IdentifierLists use. */
  protected transient IdentifierListStore ilm = null;
  
  /* The NodeStore we can read and write nodes to. */
  protected transient NodeStore nodeStore = null;
  
  /* The StringStore we load and save the value to. */
  protected transient StringStore stringStore = null;

  /* The name of this node. */
  protected String nodeName = null;
  
  /* What type of node this is - an element, attribute etc. */
  /* It is set in the constructor.                          */
  public short nodeType = -1;

  /* DOM exception code we use when we generate a DOM exception. */
  protected static final short DOM_EXCEPTION_NUM = 0;
  
  /* The URI associated with the current namespace. */
  protected String namespaceURI = null;

  /* The prefix that refers to the above namespace. */
  protected String prefix = null;
  
  /* The owner document. */
  protected Identifier documentIdentifier = null;

  /* Used to load and save the parent Identifier. */
  protected transient IdentifierStore iis = null;
  
  /* Is the node read only or not? */
  protected boolean readOnly = false;


  /**
   * Constructor.
   */
  
  public DefaultMyNode( Identifier domImplementationIdentifier,
                         Identifier identifier, 
                         IdentifierIdentifier parent, 
                         IdentifierList children, 
                         String prefix,
                         String namespaceURI,
                         String nodeName,
                         int nodeType,
                         Identifier documentIdentifier,
                         boolean readOnly )
  {
    this.domImplementationIdentifier = domImplementationIdentifier;
    this.identifier = identifier;
    this.parent = parent;
    this.children = children;
    this.prefix = prefix;
    this.namespaceURI = namespaceURI;
    this.nodeName = nodeName;
    this.nodeType = ( short ) nodeType;
    this.documentIdentifier = documentIdentifier;
    this.readOnly = readOnly;
  }


  /**
   * Loads the node whose Identifier is passed.
   */
  
  protected MyNode loadNode( Identifier identifier ) 
  {
    MyNode result = null;
    try
    {
      result = this.nodeStore.loadNode( identifier );
    }
    catch ( NodeStoreException nse )
    {
      this.throwException( nse );
    }
    return result;
  }


  /**
   * Saves this node to the NodeStore. It is assumed that the Identifier
   * for this node is valid.
   */
  
  protected void saveNode() throws DOMException
  {
    if ( this.identifier != null )
    {
      try
      {
        this.nodeStore.updateNode( this );
      }
      catch ( NodeStoreException nse )
      {
        this.throwException( nse );
      }
    }
    else 
    {
      this.throwException( "Identifier is null." );
    }
  }
  
  
  /**
   * This is where we handle any exceptions that happen while 
   * dealing with requests, but cannot pass them to the caller
   * as the Node specification doesn't define an exception.
   */
  
  protected void throwException( Exception e )
  {
    ErrorHandler.throwException( e );
  }

  
  /**
   * We call this if we need to throw an exception because of
   * some problem.
   */
  
  protected void throwException( String msg )
  {
    ErrorHandler.throwException( msg );
  }
  
  
  /**
   * Compares this node to another.
   */
  
  public boolean equals( Object o )
  {
    boolean result = false;
    if ( o instanceof DefaultMyNode )
    {
      DefaultMyNode dmn = ( DefaultMyNode ) o;
      Identifier i = dmn.getIdentifier();
      if ( ( i != null ) && ( this.identifier != null ) )
      {
        if ( i.equals( this.identifier ) )
        {
          result = true;
        }
      }
    }
    return result;
  }


  /**
   * Does a this node to that passed, and optionally compares
   * all the child nodes as well.
   */

  public boolean equals( MyNode node, boolean deep )
  {
    boolean result = this.equals( node );
    if ( deep == true )
    {
      NodeList myChildren = this.getChildNodes();
      NodeList nodesChildren = node.getChildNodes();
      int numMyChildren = myChildren.getLength();
      int numNodesChilden = nodesChildren.getLength();
      if ( numMyChildren == numNodesChilden )
      {
        int loop = 0;
        while ( loop < numMyChildren )
        {
          MyNode child1 = ( MyNode ) myChildren.item( loop );
          MyNode child2 = ( MyNode ) nodesChildren.item( loop );
          result = child1.equals( child2, deep );
          if ( result == true )
          {
            loop++;
          }
          else
          {
            loop = numMyChildren;
          }
        }
      }
    }
    return result;
  }


  /**
   * Returns the Identifier that references the DOMInstance this node
   * is associated with.
   */
  
  public Identifier getDOMImplementationIdentifier()
  {
    return this.domImplementationIdentifier;
  }



  /**
   * Prints one line of data with leading spaces.
   */
  
  public static void print( int indentAmount, String s )
  {
    String s2 = "";
    for ( int loop = 0; loop < indentAmount * 2; loop++ )
    {
      s2 += "  ";
    }
    s2 += s;
    System.out.println( s2 );
  }
  

  /**
   * Displays the facts about the passed node.
   */
  
  public static void printNode( int indentAmount, Node node, boolean recurse )
  {
    if ( ( node != null ) && ( indentAmount < 20 ) )
    {
      print( indentAmount, "Name     : " + node.getNodeName() );
      
      String s = "Type     : ";
      short i = node.getNodeType();
      switch ( i )
      {
        case Node.ATTRIBUTE_NODE: s += "Attribute"; break;
        case Node.CDATA_SECTION_NODE: s += "CDATA section"; break;
        case Node.COMMENT_NODE: s += "Comment"; break;
        case Node.DOCUMENT_FRAGMENT_NODE: s += "Document fragment"; break;
        case Node.DOCUMENT_NODE: s += "Document"; break;
        case Node.DOCUMENT_TYPE_NODE: s += "Document type"; break;
        case Node.ELEMENT_NODE: s += "Element"; break;
        case Node.ENTITY_NODE: s += "Entity"; break;
        case Node.ENTITY_REFERENCE_NODE: s += "Entity reference"; break;
        case Node.NOTATION_NODE: s += "Notation"; break;
        case Node.PROCESSING_INSTRUCTION_NODE: s += "Processing instruction"; break;
        case Node.TEXT_NODE: s += "Text"; break;
        default: s += "!!! Unknown !!!"; break;
      }
        
      print( indentAmount, s );
      
      print( indentAmount, "Value    : " + node.getNodeValue() );
      print( indentAmount, "Read-only: " + ( ( MyNode ) node ).isReadOnly() );

      
      NamedNodeMap nnp = node.getAttributes();
      if ( nnp != null )
      {
        int numAttributes = nnp.getLength();
        print( indentAmount, "# attributes = " + numAttributes );;
        for ( int loop = 0; loop < nnp.getLength(); loop++ )
        {
          MyAttr attr = ( MyAttr ) nnp.item( loop );
          print( indentAmount, "Attribute " + ( loop + 1 ) + ": Name = " + attr.getName() + ", value = " + attr.getValue() + ", readonly = " + attr.isReadOnly() );
        }
      }
      else
      {
        print( indentAmount, "Node has no attributes" );
      }
      
      print( indentAmount, "" );
      
      if ( recurse == true )
      {
        NodeList children = node.getChildNodes();
        if ( children != null )
        {
          indentAmount++;
          for ( int loop = 0; loop < children.getLength(); loop++ )
          {
            Node childNode = children.item( loop );
            printNode( indentAmount, childNode, recurse );
          }
        }
        else
        {
          print( indentAmount, "Node has no children" );
        }
      }
    }
    else
    {
      print( indentAmount, "Node is null" );
    }
  }


  /**
   * The "replaceChild" method doesn't want "newChild" to be 
   * an ancestor of "oldChild". Throws a DOMException if this is the case.
   * 
   * We do the check by doing a recursive check of all of "oldChild"'s parents.
   */
  
  protected void checkNewChildNotAncestorOfOldChild( Node newChild, Node oldChild ) throws DOMException
  {
    /* If "newChild" doesn't have a parent then don't bother doing */
    /* a long search.                                               */
    MyNode myNewChild = ( MyNode ) newChild;
    Identifier newChildParentIdentifier = myNewChild.getParentIdentifier();
    if ( newChildParentIdentifier != null )
    {
      /* Recurse through all the ancestors of "oldChild" and see if any */
      /* are the same as "newChild".                                    */
      Node oldChildParent = oldChild.getParentNode();
      while ( oldChildParent != null )
      {
        if ( oldChildParent.equals( newChild ) )
        {
          throw new DOMException( DOMException.HIERARCHY_REQUEST_ERR, "newChild cannot be an ancestor of oldChild." );
        }
        else
        {
          oldChildParent = oldChildParent.getParentNode();
        }
      }
    }
  }


  /**
   * Makes sure that the passed child nodetype is compatiable with
   * the passed parent nodetype. Ie if we are going to append or replace
   * the child node as a child of the parent node we must make sure the
   * parent is allowed to have the child as a child. If not then throw
   * an exception.
   */
  
  protected void checkNewChildCompatibleWithParent( short parentNodeType, short childNodeType ) throws DOMException
  {
    boolean b = false;
    switch ( parentNodeType )
    {
      case Node.DOCUMENT_NODE:
        if ( ( childNodeType == Node.ELEMENT_NODE )
          || ( childNodeType == Node.PROCESSING_INSTRUCTION_NODE )
          || ( childNodeType == Node.COMMENT_NODE )
          || ( childNodeType == Node.DOCUMENT_TYPE_NODE ) )
        {
          b = true;
        }
        break;
        
      case Node.DOCUMENT_FRAGMENT_NODE:
        if ( ( childNodeType == Node.ELEMENT_NODE )
          || ( childNodeType == Node.PROCESSING_INSTRUCTION_NODE )
          || ( childNodeType == Node.COMMENT_NODE )
          || ( childNodeType == Node.TEXT_NODE )
          || ( childNodeType == Node.CDATA_SECTION_NODE )
          || ( childNodeType == Node.ENTITY_REFERENCE_NODE ) )
        {
          b = true;
        }
        break;

      case Node.ENTITY_REFERENCE_NODE:
        if ( ( childNodeType == Node.ELEMENT_NODE )
          || ( childNodeType == Node.PROCESSING_INSTRUCTION_NODE )
          || ( childNodeType == Node.COMMENT_NODE )
          || ( childNodeType == Node.TEXT_NODE )
          || ( childNodeType == Node.CDATA_SECTION_NODE )
          || ( childNodeType == Node.ENTITY_REFERENCE_NODE ) )
        {
          b = true;
        }
        break;

      case Node.ELEMENT_NODE:
        if ( ( childNodeType == Node.ELEMENT_NODE )
          || ( childNodeType == Node.PROCESSING_INSTRUCTION_NODE )
          || ( childNodeType == Node.COMMENT_NODE )
          || ( childNodeType == Node.TEXT_NODE )
          || ( childNodeType == Node.CDATA_SECTION_NODE )
          || ( childNodeType == Node.ENTITY_REFERENCE_NODE ) )
        {
          b = true;
        }
        break;
        

      case Node.ATTRIBUTE_NODE:
        if ( ( childNodeType == Node.TEXT_NODE )
          || ( childNodeType == Node.ENTITY_REFERENCE_NODE ) )
        {
          b = true;
        }
        break;

      case Node.ENTITY_NODE:
        if ( ( childNodeType == Node.ELEMENT_NODE )
          || ( childNodeType == Node.PROCESSING_INSTRUCTION_NODE )
          || ( childNodeType == Node.COMMENT_NODE )
          || ( childNodeType == Node.TEXT_NODE )
          || ( childNodeType == Node.CDATA_SECTION_NODE )
          || ( childNodeType == Node.ENTITY_REFERENCE_NODE ) )
        {
          b = true;
        }
        break;
    }
    
    if ( b == false )
    {        
      throw new DOMException( DOMException.HIERARCHY_REQUEST_ERR, "Child cannot be a child of this." );
    }
  }
  
  
  /**
   * Confirms that we have write access on this node - we are allowed
   * to add children or modfiy this node. If not then throw an exception.
   */
  
  protected void confirmWriteAccess() throws DOMException
  {
    if ( this.isReadOnly() == true )
    {
      throw new DOMException( DOMException.NO_MODIFICATION_ALLOWED_ERR, "Node is read-only" );
    }
  }


  /**
   * Make sure we are not trying to append a child where we cannot. 
   * Ie you can't make an Attr a child of an Element.                
   */
  
  protected void confirmNodeCanBeAdded( Node nodeToAdd ) throws DOMException
  {
    if ( ( this instanceof Element ) && ( nodeToAdd instanceof Attr ) )
      throw new DOMException( DOMException.HIERARCHY_REQUEST_ERR, "Cannot append an Attr as a child of an Element." );
  }


  /**
   * Confirm the owner documents of this node and the soon to 
   * be added nodes are the same.
   */
  
  protected void confirmNodesHaveSameOwnerDocument( MyNode nodeToAdd ) throws DOMException
  {

    /* If the owner document of "newChild" is different to that which */
    /* created this node then throw an exception.                         */
    Identifier newChildDocumentIdentifier = nodeToAdd.getDocumentIdentifier();
    if ( newChildDocumentIdentifier.equals( this.documentIdentifier ) == false )
    {
      throw new DOMException( DOMException.WRONG_DOCUMENT_ERR, "Documents don't match." );
    }
  }


  /**
   * Inserts the node newChild before the existing child 
   * node refChild. If refChild is null, insert newChild 
   * at the end of the list of children. 
   * If newChild is a DocumentFragment object, all of its 
   * children are inserted, in the same order, before 
   * refChild. If the newChild is already in the tree, 
   * it is first removed
   */
  
  public Node realInsertBefore( Node newChild, Node refChild ) throws DOMException
  {
    try
    {
      if ( refChild != null )
      {
        Identifier refChildIdentifier = ( ( MyNode ) refChild ).getIdentifier();
        Identifier newChildIdentifier = ( ( MyNode ) newChild ).getIdentifier();
  
        /* If the node is already in the list then remove it first. */
        int index = this.children.indexOf( newChildIdentifier );
        if ( index >= 0 )
        {
          this.children.remove( index );
        }
        
        /* If the "refChild" is in the list of children then insert the new      */
        /* node before it, otherwise throw an exception.                          */
        index = this.children.indexOf( refChildIdentifier );
        if ( index >= 0 )
        {
          this.children.insert( index, newChildIdentifier );
          ( ( MyNode ) newChild ).setParentIdentifier( this.identifier );
        }
        else
        {
          throw new DOMException( DOMException.NOT_FOUND_ERR, "RefChild not found." );
        }
      }
      else
      {
        /* Append the new node at the end of the list. */
        Identifier newChildIdentifier = ( ( MyNode ) newChild ).getIdentifier();
        this.children.add( newChildIdentifier );
      }
    }
    catch ( IdentifierListException ile )
    {
      throw new DOMException( DefaultMyNode.DOM_EXCEPTION_NUM, ile.getMessage() );
    }
     return newChild;
  }
  
  
  /**
   * Replaces the child node oldChild with newChild in the 
   * list of children, and returns the oldChild node. 
   * If newChild is a DocumentFragment object, oldChild is 
   * replaced by all of the DocumentFragment children, 
   * which are inserted in the same order. If the newChild 
   * is already in the tree, it is first removed.
   */
  
  public Node realReplaceChild(Node newChild, Node oldChild) throws DOMException
  {
    try
    {
      MyNode myNewChild = ( MyNode ) newChild;
      MyNode myOldChild = ( MyNode ) oldChild;
      
      /* Make sure the owner document of "newChild" is the same as the */
      /* Document that created this node.                               */
      Identifier newChildDocumentIdentifier = myNewChild.getDocumentIdentifier();
      Identifier thisDocumentIdentifier = this.getDocumentIdentifier();
      if ( newChildDocumentIdentifier.equals( thisDocumentIdentifier ) == true )
      {
        /* Make sure "newChild" is not an ancestor of oldChild". */
        this.checkNewChildNotAncestorOfOldChild( newChild, oldChild );
        
        /* Make sure "newChild"'s node type is compatible with the */
        /* type of "this" - ie certain types of nodes don't allow  */
        /* certain types of children.                               */
        this.checkNewChildCompatibleWithParent( this.getNodeType(), newChild.getNodeType() );

        /* If the new child is already in the list then remove it first. */
        Identifier oldChildIdentifier = myOldChild.getIdentifier();
        Identifier newChildIdentifier = myNewChild.getIdentifier();
        int index = this.children.indexOf( newChildIdentifier );
        if ( index >= 0 )
        {
          this.children.remove( index );
        }
        
        /* If the old child we are meant to be replacing is not in the */
        /* list then throw an exception.                               */
        index = this.children.indexOf( oldChildIdentifier );
        if ( index >= 0 )
        {
          this.children.set( index, newChildIdentifier );
          myNewChild.setParentIdentifier( this.identifier );
        }
        else
        {
          /* "oldChild" is not a child of this node. Throw an exception. */
          throw new DOMException( DOMException.NOT_FOUND_ERR, "oldChild is not a child of this node." );
        }
      }
      else
      {
        throw new DOMException( DOMException.WRONG_DOCUMENT_ERR, "newChild has a different owner document to this node." );
      }
    }
    catch ( IdentifierListException ile )
    {
      throw new DOMException( DefaultMyNode.DOM_EXCEPTION_NUM, ile.getMessage() );
    }
    
     return oldChild;
  }  
  
  
  /**
   * Removes a child node.
   */
  
  public Node realRemoveChild(Node oldChild) throws DOMException
  {
    try
    {
      MyNode myOldChild = ( MyNode ) oldChild;
      Identifier oldChildIdentifier = myOldChild.getIdentifier();
      int index = this.children.indexOf( oldChildIdentifier );
      if ( index >= 0 )
      {
        this.children.remove( index );
        myOldChild.setParentIdentifier( null );
      }
      else
      {
        throw new DOMException( DOMException.NOT_FOUND_ERR, "Node to remove is not a child of this node." );
      }
    }
    catch ( IdentifierListException ile )
    {
      throw new DOMException( DefaultMyNode.DOM_EXCEPTION_NUM, ile.getMessage() );
    }

     return oldChild;
  }    
  
  
  /**
   * Adds the node newChild to the end of the list of children 
   * of this node. If the newChild is already in the tree, it 
   * is first removed. 
   */
  
  public Node realAppendChild(Node newChild) throws DOMException
  {
    if ( newChild != null )
    {
      /* Make sure we are allowed to append the passed node to us - */
      /* ie you aren't allowed to add an attribute as a child of an */
      /* element via this method - you must use "setAttributeNode"  */
      /* insead.                                                    */
      this.confirmNodeCanBeAdded( newChild );
            
      try
      {
        MyNode myNewChild = ( MyNode ) newChild;

        /* Confirm the owner documents of this node and the soon to */
        /* be added nodes are the same.                              */
        this.confirmNodesHaveSameOwnerDocument( myNewChild );

        /* If the owner document of "newChild" is different to that which */
        /* created this node then throw an exception.                         */
        Identifier newChildDocumentIdentifier = myNewChild.getDocumentIdentifier();
        if ( newChildDocumentIdentifier.equals( this.documentIdentifier ) == false )
        {
          throw new DOMException( DOMException.WRONG_DOCUMENT_ERR, "Documents don't match." );
        }

        
        Identifier newChildIdentifier = myNewChild.getIdentifier();

  
        /* If the passed child is already a child of this node */
        /* then remove it first before appending it again.      */
        /* This is what is meant to happen.                     */
        Identifier childsParentIdentifier = myNewChild.getParentIdentifier();
        if ( childsParentIdentifier != null )
        {
          if ( childsParentIdentifier.equals( this.identifier ) == true )
          {
            this.children.remove( newChildIdentifier );
          }
        }
        
        
        /* If the new child is a DocumentFragment then append all its */
        /* children, otherwise just append the child node.            */
        if ( newChild instanceof DocumentFragment )
        {
           /* Add all the child nodes of "newChild" to  this node. */
          NodeList childNodes = newChild.getChildNodes();
          if ( childNodes != null )
          {
            int numChildNodes = childNodes.getLength();
            for ( int loop = 0; loop < numChildNodes; loop++ )
            {
              Node childNode = childNodes.item( loop );
              this.appendChild( childNode );
            }
          }
        }
        else
        {
          this.children.add( newChildIdentifier );
          myNewChild.setParentIdentifier( this.identifier );
        }
      }
      catch ( IdentifierListException ile )
      {
        throw new DOMException( DefaultMyNode.DOM_EXCEPTION_NUM, ile.getMessage() );
      }
    }
     return newChild;
  }    
  

  /**
   * Makes sure the new child is not a parent or ancestor of this node. 
   * Ie if "newChild" is currently a parent of this noed we cannot also
   * make it a child.
   */
  
  protected void confirmNodeIsNotAncestor( Node node ) throws DOMException
  {
    if ( node != null )
    {
      Node parent = this.getParentNode();
      while ( parent != null )
      {
        if ( parent.equals( node ) )
        {
          throw new DOMException( DOMException.HIERARCHY_REQUEST_ERR, "Node is an ancestor." );
        }
        parent = parent.getParentNode();
      }
    }
  }


  /*--------------------------------- MyNode methods ----------------------*/
    
  
  /**
   * Returns the Identifier for this instance.
   */
  
  public Identifier getIdentifier()
  {
    return this.identifier;
  }
  
  
  /**
   * Sets the identifier for this node.
   */
  
  public void setIdentifer( Identifier identifier )
  {
    this.identifier = identifier;
    
//    try
//    {
//      this.saveNode();
//    }
//    catch ( DOMException de )
//    {
//      this.throwException( de );
//    }
  }
  

  /**
   * Returns the parent's Identifier.
   */
  
  public Identifier getParentIdentifier()
  {
    Identifier result = null;
    if ( this.parent != null )
    {
      try
      {
        result = this.iis.getIdentifier( this.parent );
      }
      catch ( IdentifierStoreException iise )
      {
        this.throwException( iise );
      }
    }
    return result;
  }


  /**
   * Sets the parent's Identifier.
   */
  
  public void setParentIdentifier( Identifier identifier )
  {
    try
    {
      this.iis.setIdentifier( this.parent, identifier );
    }
    catch ( IdentifierStoreException iise )
    {
      this.throwException( iise );
    }
  }


  /**
   * Returns the list of children Identifiers for this node.
   */
  
  public IdentifierList getChildrenIdentifiers()
  {
    return this.children;
  }
  
  
  /**
   * Returns the parent IdentifierIdentifier.
   */
  
  public IdentifierIdentifier getParentIdentifierIdentifier()
  {
    return this.parent;
  }

  
  /**
   * Returns the Identifier of the main document.
   */
  
  public Identifier getDocumentIdentifier()
  {
    return this.documentIdentifier;
  }
  
  
  /**
   * If an instance of a MyNode is serialized this method
   * should be called after it has been de-serialized.
   */
  
  public void init() throws MyNodeException
  {
    /* Initialize the two IdentifierList instances. */
    try
    {
      MyDOMImplementation domImplementation = MyDOMImplementationManager.getInstance( this.domImplementationIdentifier );
      this.nodeStore = domImplementation.getNodeStore();
      this.iis = domImplementation.getIdentifierStore();
      this.stringStore = domImplementation.getStringStore();
      this.ilm = domImplementation.getIdentifierListStore();
      
      if ( this.children != null )
      {
        this.children.init();
      }
    }
    catch ( IdentifierListException ile )
    {
      throw new MyNodeException( ile );
    }
  }
  
    
  
  /**
   * Sets whether this node is read-only or not.
   */
  
  public void setReadOnly( boolean readOnly )
  {
    this.readOnly = readOnly;
    this.saveNode();
  }
  
  
  /**
   * Returns whether this node is read-only or not.
   */
  
  public boolean isReadOnly()
  {
    return this.readOnly;
  }

  
  /*------------------------- org.w3c.dom.Node methods -----------------------*/

  
  /**
   * Returns the name of this node.
   */
  
  public String getNodeName()
  {
    String result = null;
    if ( this.prefix != null )
      result = this.prefix + this.nodeName;
    else
      result = this.nodeName;
    return result;
  }


  /**
   * Returns the value of this node as text.
   * We return null here. It is up to specific sub-classes
   * to handle this appropriately.
   */
  
  public String getNodeValue() throws DOMException
  {
    return null;
  }  
  
  
  /**
   * Sets the value of this node as text.
   * We do nothing with the passed value here. It is up to 
   * specific sub-classes to handle this appropriately.
   */
  
  public void setNodeValue(String nodeValue) throws DOMException
  {
  }

  
  /**
   * Returns the type of node this is - an element, attribute etc.
   */
  
  public short getNodeType()
  {
    return this.nodeType;
  }


  /**
   * Returns the parent node of this node.
   */
  
  public Node getParentNode()
  {
    Node result = null;
    if ( this.parent != null )
    {
      Identifier i = this.getParentIdentifier();
      if ( i != null )
      {
        if ( i.equals( this.documentIdentifier ) == false )
        {
          result = this.loadNode( i );
        }
        else
        {
          result = this.getOwnerDocument();
        }
      }
      else
      {
//        this.throwException( "Parent node is null." );
      }
    }
    else
    {
      if ( ( this instanceof DocumentFragment == false )
        && ( this instanceof Document         == false ) )
      {
        result = this.getOwnerDocument();
      }
    }
    return result;
  }
  
  
  /**
   * Returns a list of child nodes to this node.
   */
  
  public NodeList getChildNodes()
  {
    NodeList result = new IdentifierListNodeList( this.children, this.nodeStore );
    return result;
  }
  
  
  /**
   * Returns the first child node of this node.
   */
  
  public Node getFirstChild()
  {
    Node result = null;
    
    try
    {
      if ( this.children != null )
      {
        int numChildren = this.children.size();
        if ( numChildren > 0 )
        {
          Identifier identifier = this.children.get( 0 );
           result = this.loadNode( identifier );
        }
      }
    }
    catch ( IdentifierListException ile )
    {
      this.throwException( ile );
    }

    return result;
  }
  
  
  
  /**
   * Returns the last child node of this node.
   */
  
  public Node getLastChild()
  {
    Node result = null;
    
    try
    {
      if ( this.children != null )
      {
        int numChildren = this.children.size();
        if ( numChildren > 0 )
        {
          Identifier identifier = this.children.get( numChildren - 1 );
           result = this.loadNode( identifier );
        }
      }
    }
    catch ( IdentifierListException ile )
    {
      this.throwException( ile );
    }

    return result;
  }


  /**
   * The node immediately preceding this node. 
   * If there is no such node, this returns null.
   * 
   * !!! We need to speed this up. Calling "indexOf" in
   * IdentifierList is slow.
   */
  
  public Node getPreviousSibling()
  {
    Node result = null;
    
    try
    {
      MyNode parentNode = ( MyNode ) this.getParentNode();
      IdentifierList parentsChildren =  parentNode.getChildrenIdentifiers();
      int index = parentsChildren.indexOf( this.identifier );
      if ( index > 0 )
      {
        Identifier identifier = parentsChildren.get( index - 1 );
        result = this.loadNode( identifier );
      }
    }
    catch ( IdentifierListException ile )
    {
      this.throwException( ile );
    }

    return result;
  }


  /**
   * The node immediately after this node. 
   * If there is no such node, this returns null.
   *
   * !!! We need to speed this up. Calling "indexOf" in
   * IdentifierList is slow.
   */
  
  public Node getNextSibling()
  {
    Node result = null;
    
    try
    {
      MyNode parentNode = ( MyNode ) this.getParentNode();
      if ( parentNode != null )
      {
        IdentifierList parentsChildren =  parentNode.getChildrenIdentifiers();
        int numChildren = parentsChildren.size();
        int index = parentsChildren.indexOf( this.identifier );
        if ( ( index >= 0 ) && ( index < numChildren - 1 ) )
        {
          Identifier identifier = parentsChildren.get( index + 1 );
          result = this.loadNode( identifier );
        }
      }
    }
    catch ( IdentifierListException ile )
    {
      this.throwException( ile );
    }

    return result;
  }
  
  
  /**
   * Returns a NamedNodeMap containing the attributes of 
   * this node (if it is an Element) or null otherwise.
   * We return null here, the MyAttr class correctly returns
   * the list of attributes.
   */
  
  public NamedNodeMap getAttributes()
  {
    return null;
  }


  /**
   * Returns the main parent document this node is part of.
   */
  
  public Document getOwnerDocument()
  {
    /* First attempt to retrieve a cached instance of the Document. */
    /* If one doesn't exist then create one.                         */
    MyDOMImplementation domImplementation = MyDOMImplementationManager.getInstance( this.domImplementationIdentifier );
    MyDocument result = ( MyDocument ) domImplementation.getCachedDocument( this.documentIdentifier );
//    if ( result == null )
//    {
//      result = ( MyDocument ) this.loadNode( this.documentIdentifier );
//      domImplementation.addDocument( result );
//    }
    return result;
  }
  
  
  /**
   * Inserts the node newChild before the existing child 
   * node refChild. If refChild is null, insert newChild 
   * at the end of the list of children. 
   * If newChild is a DocumentFragment object, all of its 
   * children are inserted, in the same order, before 
   * refChild. If the newChild is already in the tree, 
   * it is first removed
   */
  
  public Node insertBefore( Node newChild, Node refChild ) throws DOMException
  {
    /* This confirms we have write permission on this node and throws */
    /* an exception if not.                                            */
    this.confirmWriteAccess();
    
    /* Confirm we are allowed to the new node to this node. */
    this.confirmNodeCanBeAdded( newChild );
    
    /* Confirm the new child came from the same owner document as */
    /* this node.                                                  */
    this.confirmNodesHaveSameOwnerDocument( ( MyNode ) newChild );
    
    /* Confirm the new child is not an ancestor of this node. */
    this.confirmNodeIsNotAncestor( newChild );
    
    /* If "newChild" is a document fragment then insert all its children */
    /* before "refChild", otherwise just insert the one "newChild" node. */
    short nodeType = newChild.getNodeType();
    if ( nodeType == Node.DOCUMENT_FRAGMENT_NODE )
    {
      NodeList children = newChild.getChildNodes();
      int numChildren = children.getLength();
      for ( int loop = 0; loop < numChildren; loop++ )
      {
        Node child = children.item( loop );
        this.realInsertBefore( child, refChild );
      }
    }
    else
    {
      this.realInsertBefore( newChild, refChild );
    }

     return newChild;
  }
  
  
  /**
   * Replaces the child node oldChild with newChild in the 
   * list of children, and returns the oldChild node. 
   * If newChild is a DocumentFragment object, oldChild is 
   * replaced by all of the DocumentFragment children, 
   * which are inserted in the same order. If the newChild 
   * is already in the tree, it is first removed.
   */
  
  public Node replaceChild(Node newChild, Node oldChild) throws DOMException
  {
    /* This confirms we have write permission on this node and throws */
    /* an exception if not.                                            */
    this.confirmWriteAccess();

    this.realReplaceChild( newChild, oldChild );
    
     return oldChild;
  }  
  
  
  /**
   * Removes a child node.
   */
  
  public Node removeChild(Node oldChild) throws DOMException
  {
    /* This confirms we have write permission on this node and throws */
    /* an exception if not.                                            */
    this.confirmWriteAccess();

    this.realRemoveChild( oldChild );

     return oldChild;
  }    
  
  
  /**
   * Adds the node newChild to the end of the list of children 
   * of this node. If the newChild is already in the tree, it 
   * is first removed. 
   */
  
  public Node appendChild(Node newChild) throws DOMException
  {
    /* This confirms we have write permission on this node and throws */
    /* an exception if not.                                            */
    this.confirmWriteAccess();
    
    /* Makes sure the new child is not a parent or ancestor of this node. */
    /* Ie if "newChild" is currently a parent of this noed we cannot also */
    /* make it a child.                                                    */
    this.confirmNodeIsNotAncestor( newChild );

    this.realAppendChild( newChild );

     return newChild;
  }    


  /**
   * Returns true if this node has children.    
   */
  
  public boolean hasChildNodes()
  {
    boolean result = false;
    try
    {
      if ( this.children != null )
      {
        int numChildren = this.children.size();
        result = ( numChildren > 0 );
      }
    }
    catch ( IdentifierListException ile )
    {
      throw new DOMException( DefaultMyNode.DOM_EXCEPTION_NUM, ile.getMessage() );
    }
     return result;
  }
  
  
  /**
   * Puts all Text nodes in the full depth of the sub-tree 
   * underneath this Node, including attribute nodes, into 
   * a "normal" form where only structure (e.g., elements, 
   * comments, processing instructions, CDATA sections, 
   * and entity references) separates Text nodes, i.e., 
   * there are neither adjacent Text nodes nor empty Text 
   * nodes. This can be used to ensure that the DOM view of 
   * a document is the same as if it were saved and 
   * re-loaded, and is useful when operations (such as 
   * XPointer lookups) that depend on a particular document 
   * tree structure are to be used.In cases where the 
   * document contains CDATASections, the normalize 
   * operation alone may not be sufficient, since XPointers 
   * do not differentiate between Text nodes and 
   * CDATASection nodes. 
   */
  
  public void normalize()
  {
    /* Look at all the children of this node. If any two of them are text nodes   */
    /* then merge them. Then recurse into each so they also normalize themselves. */
    NodeList children = this.getChildNodes();
    int numChildren = children.getLength();
    for ( int loop = 0; loop < numChildren; loop++ )
    {
      if ( loop < numChildren - 1 )
      {
        Node node1 = children.item( loop );
        Node node2 = children.item( loop + 1 );
        int node1Type = node1.getNodeType();
        int node2Type = node2.getNodeType();
        if ( ( node1Type == Node.TEXT_NODE ) && ( node2Type == Node.TEXT_NODE ) )
        {
          /* If the first node is not empty then append the second node to the */
          /* first node, then delete the second node. If either or both nodes  */
          /* are empty then delete them.                                       */
          Text text1 = ( Text ) node1;
          Text text2 = ( Text ) node2;
          String value1 = text1.getNodeValue();
          String value2 = text2.getNodeValue();
          if ( ( value1.equals( "" ) == false ) && ( value2.equals( "" ) == false ) )
          {
            /* Append the second node's value to the first and delete the second node. */
            text1.setNodeValue( value1 + value2 );
            this.removeChild( node2 );
            numChildren = children.getLength();
          }
          else
          if ( value1.equals( "" ) == false )
          {
            this.removeChild( node1 );
            numChildren = children.getLength();
          }
          else
          if ( value2.equals( "" ) == false )
          {
            this.removeChild( node2 );
            numChildren = children.getLength();
          }
          else
          if ( ( value1.equals( "" ) == true ) && ( value2.equals( "" ) == true ) )
          {
            this.removeChild( node1 );
            this.removeChild( node2 );
            numChildren = children.getLength();
          }
        }
        else
        if ( node1Type == Node.TEXT_NODE )
        {
          Text text1 = ( Text ) node1;
          String value1 = text1.getNodeValue();
          value1 = au.com.explodingsheep.diskDOM.helpers.TrimString.trimString( value1 );
          if ( value1.equals( "" ) == true )
          {
            this.removeChild( node1 );
          }
        }  
      }
      else
      {
        Node node1 = children.item( loop );
        int node1Type = node1.getNodeType();
    
        if ( node1Type == Node.TEXT_NODE )
        {
          Text text1 = ( Text ) node1;
          String value1 = text1.getNodeValue();
          value1 = au.com.explodingsheep.diskDOM.helpers.TrimString.trimString( value1 );
          if ( value1.equals( "" ) == true )
          {
            this.removeChild( node1 );
          }
        }
      }
    }
    
    /* Now recurse through the child list again, calling */
    /* their "normalize" method.                         */
    for ( int loop = 0; loop < numChildren - 1; loop++ )
    {
      Node node = children.item( loop );
      node.normalize();
    }
  }
  
  
  /**
   * Tests whether the DOM implementation implements a 
   * specific feature and that feature is supported by 
   * this node. 
   */
  
  public boolean isSupported(String feature, String version)
  {
    boolean result = false;
    if ( ( feature.equalsIgnoreCase( "core" ) )
      && ( version.equalsIgnoreCase( "2.0" ) ) )
    {
      result = true;
    }
    return result;
  }
  

  /**
   * The namespace URI of this node, or null if it is 
   * unspecified. This is not a computed value that is 
   * the result of a namespace lookup based on an 
   * examination of the namespace declarations in 
   * scope. It is merely the namespace URI given at 
   * creation time. For nodes of any type other than 
   * ELEMENT_NODE and ATTRIBUTE_NODE and nodes created 
   * with a DOM Level 1 method, such as createElement 
   * from the Document interface, this is always null.
   * Per the Namespaces in XML Specification an 
   * attribute does not inherit its namespace from the 
   * element it is attached to. If an attribute is not 
   * explicitly given a namespace, it simply has no 
   * namespace.   
   */
  
  public String getNamespaceURI()
  {
    return this.namespaceURI;
  }
  
  
  /**
   * The namespace prefix of this node, or null if it is unspecified. 
   * Note that setting this attribute, when permitted, changes the 
   * nodeName attribute, which holds the qualified name, as well as 
   * the tagName and name attributes of the Element and Attr 
   * interfaces, when applicable. Note also that changing the prefix 
   * of an attribute that is known to have a default value, does not 
   * make a new attribute with the default value and the original 
   * prefix appear, since the namespaceURI and localName do not 
   * change. For nodes of any type other than ELEMENT_NODE and 
   * ATTRIBUTE_NODE and nodes created with a DOM Level 1 method, 
   * such as createElement from the Document interface, this is 
   * always null
   */

  public String getPrefix()
  {
    return this.prefix;
  }
  

  /**
   * Sets the namespace prefix for this node.
   */
    
  public void setPrefix(String prefix) throws DOMException
  {
    this.prefix = prefix;
  }
  

  /**
   * Returns the local part of the qualified name of this node. 
   * For nodes of any type other than ELEMENT_NODE and 
   * ATTRIBUTE_NODE and nodes created with a DOM Level 1 method, 
   * such as createElement from the Document interface, this is 
   * always null.   
   */
  
  public String getLocalName()
  {
    return this.nodeName;
  }
  

  /**
   * Returns whether this node (if it is an element) has any attributes.
   * In this class we always return false, in MyElement it correctly
   * returns true or false.
   */
  
  public boolean hasAttributes()
  {
    return false;
  }
  
  
  /**
   * Clones a node. We only support deep cloning here. It is assumed that
   * sub-instances will do their own shallow cloning.
   */
  
  public Node cloneNode(boolean deep)
  {
    MyNode result = null;
  
    if ( deep == true )
    {
      /* First clone this node. */
      result = ( MyNode ) this.cloneNode( false );
      
      /* Now clone any of this node's children. */
      NodeList childNodes = this.getChildNodes();
      int numChildNodes = childNodes.getLength();
      if ( numChildNodes > 0  )
      {
        /* We must turn off "read-only" temporarily so we can append */
        /* any children.                                              */
        boolean readOnlyChanged = false;
        boolean readOnly = result.isReadOnly();
        if ( readOnly == true )
        {
          result.setReadOnly( false );
          readOnlyChanged = true;
        }
        
        for ( int loop = 0; loop < numChildNodes; loop++ )
        {
          Node childNode = childNodes.item( loop );
          Node clonedChildNode = childNode.cloneNode( true );
          result.appendChild( clonedChildNode );
        }
        
        /* Reset the readonly state. */
        if ( readOnlyChanged == true )
        {
          result.setReadOnly( true );
        }
      }  
    }
      
    return result;
  }
}
