package au.com.explodingsheep.diskDOM;


import java.io.Serializable;
import org.w3c.dom.*;
import au.com.explodingsheep.diskDOM.helpers.ErrorHandler;
import au.com.explodingsheep.diskDOM.identifier.*;
import au.com.explodingsheep.diskDOM.nodeStore.*;


/**
 * This class implements the NamedNodeMap interface and wraps the
 * IdentifierList class.
 */

public class IdentifierListNamedNodeMap implements NamedNodeMap, Serializable
{
  /* The nodestore that use use to retrive nodes from. */
  protected NodeStore nodeStore = null;
  
  /* The nodes in this list. */
  protected IdentifierList identifierList = null;

  /* The parent node these attribute nodes are children of. */
  protected Identifier parentIdentifier = null;
    
  /* The Document that created this object. */
  protected Document ownerDocument = null;

  
  /**
   * Constructor.
   */
  
  public IdentifierListNamedNodeMap( NodeStore nodeStore,IdentifierList identifierList, Identifier parentIdentifier, Document ownerDocument )
  {
    this.nodeStore = nodeStore;
    this.identifierList = identifierList;
    this.parentIdentifier = parentIdentifier;
    this.ownerDocument = ownerDocument;
  }
  
  
  /**
   * Returns true if a node with the passed name exists.
   */
  
  public boolean doesNodeExist(String namespaceURI, String localName) 
  {
    boolean result = false;
    
    if ( localName != null ) 
    {
       try
      {
        int numNodes = this.identifierList.size();
        int loop = 0;
        while ( loop < numNodes )
        {
          Identifier identifier = this.identifierList.get( loop );
          MyNode node = this.nodeStore.loadNode( identifier );
          
          boolean match1 = false;
          boolean match2 = false;
          String nodeName = node.getNodeName();
          String uri = node.getNamespaceURI();
          
          if ( ( namespaceURI == null ) || ( uri.equals( namespaceURI ) == true ) )
          {
            match1 = true;
          }
          
          if ( nodeName.equals( localName ) == true )
          {
            match2 = true;
          }
          
          if ( ( match1 == true ) && ( match2 == true ) )
          {
            result = true;
            loop = numNodes;
          }
          else
          {
            loop++;
          }
        }
      }
      catch ( IdentifierListException ile )
      {
        ErrorHandler.throwException( ile );
      }
      catch ( NodeStoreException nse )
      {
        ErrorHandler.throwException( nse );
      }
    }

     return result;
  }


  /**
   * Compares the passed two strings. Returns true if they are the same
   * or both are null.
   */
  
  protected boolean compareStrings( String s1, String s2 )
  {
    boolean result = false;
    if ( ( s1 != null ) && ( s2 != null ) )
    {
      result = ( s1.equals( s2 ) );
    }
    else
    if ( ( s1 == null ) && ( s2 == null ) )    
    {
      result = true;
    }
    return result;
  }    
  
  
  /*-------------------------------- NamedNodeMap method -------------------------*/
  

  /**
   * Returns the number of nodes in this map.
   */

  public int getLength()
  {
    int result = 0;
    try
    {
      result = this.identifierList.size();
    }
    catch ( IdentifierListException ile )
    {
      ErrorHandler.throwException( ile );
    }  
     return result;
  }

  
  /**
   * Retrieves a node specified by name. 
   */
  
  public Node getNamedItem( String name ) 
  {
    Node result = null;

     try
    {
      int numNodes = this.identifierList.size();
      int loop = 0;
      while ( loop < numNodes )
      {
        Identifier identifier = this.identifierList.get( loop );
        MyNode node = this.nodeStore.loadNode( identifier );
        String nodeName = node.getNodeName();
        if ( nodeName.equals( name ) == true )
        {
          result = node;
          loop = numNodes;
        }
        else
        {
          loop++;
        }
      }
    }
    catch ( IdentifierListException ile )
    {
      ErrorHandler.throwException( ile );
    }
    catch ( NodeStoreException nse )
    {
      ErrorHandler.throwException( nse );
    }

    return result;
  }
          
 
   /**
    * Retrieves a node specified by local name and namespace URI. 
    */
            
  public Node getNamedItemNS(String namespaceURI, String localName) 
  {
    Node result = null;

     try
    {
      int numNodes = this.identifierList.size();
      int loop = 0;
      while ( loop < numNodes )
      {
        Identifier identifier = this.identifierList.get( loop );
        MyNode node = this.nodeStore.loadNode( identifier );
        String nodeName = node.getNodeName();
        String nodeNamespaceURI = node.getNamespaceURI();
        
        boolean match1 = this.compareStrings( namespaceURI, nodeNamespaceURI );
        boolean match2 = this.compareStrings( localName, nodeName );
        if ( ( match1 == true ) && ( match2 == true ) )
         {
           result = node;
           loop = numNodes;
         }
         else
         {
           loop++;
         }
      }
    }
    catch ( IdentifierListException ile )
    {
      ErrorHandler.throwException( ile );
    }
    catch ( NodeStoreException nse )
    {
      ErrorHandler.throwException( nse );
    }

    return result;
  }
  
  
  /**
   * Returns the indexth item in the map. 
   */
          
  public Node item(int index) 
  {
    MyNode result = null;
    try
    {
        Identifier identifier = this.identifierList.get( index );
        result = this.nodeStore.loadNode( identifier );
    }
    catch ( IdentifierListException ile )
    {
      ErrorHandler.throwException( ile );
    }
    catch ( NodeStoreException nse )
    {
      ErrorHandler.throwException( nse );
    }
     return result;
  }


  /**
   * Removes a node specified by name. 
   */
  
  public Node removeNamedItem(String name) 
  {
    Node result = this.removeNamedItemNS( null, name );
    return result;
  }
          
         
  /**
   * Removes a node specified by local name and namespace URI. 
   */
  
  public Node removeNamedItemNS(String namespaceURI, String localName) 
  {
    Node result = null;
    
    if ( localName != null )
    {
       try
      {
        int numNodes = this.identifierList.size();
        int loop = 0;
        while ( loop < numNodes )
        {
          Identifier identifier = this.identifierList.get( loop );
          MyNode node = this.nodeStore.loadNode( identifier );
          
          boolean match1 = false;
          boolean match2 = false;
          String nodeName = node.getNodeName();
          String uri = node.getNamespaceURI();
          
          if ( ( namespaceURI == null ) || ( uri.equals( namespaceURI ) == true ) )
          {
            match1 = true;
          }
          
          if ( nodeName.equals( localName ) == true )
          {
            match2 = true;
          }
          
          if ( ( match1 == true ) && ( match2 == true ) )
          {
            result = node;
            loop = numNodes;
            
            /* We have found the node. Delete it - ie remove it */
            /* from the nodestore and the IdentifierList.        */
            this.identifierList.remove( identifier );
            this.nodeStore.deleteNode( identifier );
            
            /* If the node is an attribute then we may need to re-create */
            /* it if it has a default value.                             */
            if ( node instanceof DefaultMyAttr )
            {
              MyAttr myAttr = ( MyAttr ) node;
              Identifier parentIdentifier = myAttr.getParentIdentifier();
              
              MyElement parent = ( MyElement ) this.nodeStore.loadNode( parentIdentifier );
              parent.createDefaultAttributes();
            }
          }
          else
          {
            loop++;
          }
        }
      }
      catch ( IdentifierListException ile )
      {
        ErrorHandler.throwException( ile );
      }
      catch ( NodeStoreException nse )
      {
        ErrorHandler.throwException( nse );
      }
    }
    
    /* If we didn't find the item that we wanted to remove then */
    /* throw an exception.                                      */
    if ( result == null )
    {
      throw new DOMException( DOMException.NOT_FOUND_ERR, "Could not find a node with the name \"" + localName + "\", namespace \"" + namespaceURI + "\"." );
    }

    return result;
  }

   
  /**
   * Adds a node using its nodeName attribute. 
   * 
   * !!! What if the passed node has not yet been written to the
   * NodeStore? Or can it? For every node we create an entry should
   * exist in the nodestore.
   * 
   * We will assume that the node is already in the NodeStore and so
   * only needs adding to the IdentifierList.
   */          
  
  public Node setNamedItem(Node arg) 
  {
    Node result = null;
    if ( arg != null )
    {
      MyNode myNode = ( MyNode ) arg;
      Identifier identifier = ( Identifier ) myNode.getIdentifier();
      
      /* If "arg" is an Attr and already has a parent node - ie it is    */
      /* already connected to another element - then throw an exception. */
      if ( arg instanceof Attr ) 
      {
        Attr attr = ( Attr ) arg;
        Element ownerElement = attr.getOwnerElement();
        if ( ownerElement != null )
        {
          throw new DOMException( DOMException.INUSE_ATTRIBUTE_ERR, "Attribute is connected to another element." );
        }
      }
      
      /* If the owner document of "arg" is different to that which */
      /* created this named node map then throw an exception.       */
      Document argDocument = arg.getOwnerDocument();
      if ( argDocument.equals( this.ownerDocument ) == false )
      {
        throw new DOMException( DOMException.WRONG_DOCUMENT_ERR, "Documents don't match." );
      }
      
      
      /* If a node with the same name and namespace already exists */
      /* then replaced it with the new node.                       */
      try
      {
        String localName = arg.getLocalName();
        String namespaceURI = arg.getNamespaceURI();
        result = this.getNamedItemNS( namespaceURI, localName );
        if ( result != null )
        {
          MyNode originalNodeBeingReplaced = ( MyNode ) result;
          Identifier originalNodeIdentifier = originalNodeBeingReplaced.getIdentifier();
          this.identifierList.remove( originalNodeIdentifier );
          this.nodeStore.deleteNode( originalNodeIdentifier );
          originalNodeBeingReplaced.setParentIdentifier( null );
        }
            
        /* Add the node's identifier to the list of node numbers. */
        this.identifierList.add( identifier );

        /* Tell the new child node of its parent. */
        myNode.setParentIdentifier( this.parentIdentifier );
      }
      catch ( IdentifierListException ile )
      {
        ErrorHandler.throwException( ile );
      }
      catch ( NodeStoreException nse )
      {
        ErrorHandler.throwException( nse );
      }
    }
     return result;
  }
     
          
   /**
    * Adds a node using its namespaceURI and localName. 
    */
 
  public Node setNamedItemNS(Node arg) 
  {
    Node result = null;
    if ( arg != null )
    {
      /* If "arg" is an Attr and already has a parent node then throw */
      /* an exception.                                                 */
      if ( arg instanceof Attr ) 
      {
        Attr attr = ( Attr ) arg;
        Element ownerElement = attr.getOwnerElement();
        if ( ownerElement != null )
        {
          throw new DOMException( DOMException.INUSE_ATTRIBUTE_ERR, "Attribute is connected to another element." );
        }
      }

      /* If the owner document of "arg" is different to that which */
      /* created this named node map then throw an exception.       */
      Document argDocument = arg.getOwnerDocument();
      if ( argDocument.equals( this.ownerDocument ) == false )
      {
        throw new DOMException( DOMException.WRONG_DOCUMENT_ERR, "Documents don't match." );
      }


      /* If a node with the same name and namespace already exists */
      /* then delete it.                                           */
      String localName = arg.getLocalName();
      String namespaceURI = arg.getNamespaceURI();
      if ( this.doesNodeExist( namespaceURI, localName ) == true )
        result = this.removeNamedItemNS( namespaceURI, localName );

      /* Add the passed node to this list. */
      MyNode myNode = ( MyNode ) arg;
      Identifier identifier = ( Identifier ) myNode.getIdentifier();
      
      /* Add the node's identifier to the list of node numbers. */
      try
      {
        this.identifierList.add( identifier );

        /* Tell the new child node of its parent. */
        myNode.setParentIdentifier( this.parentIdentifier );
      }
      catch ( IdentifierListException ile )
      {
        ErrorHandler.throwException( ile );
      }
    }
     return result;
  }
}
