package au.com.explodingsheep.diskDOM.documentBuilder;


import java.net.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import java.io.*;
import au.com.explodingsheep.diskDOM.*;
import au.com.explodingsheep.diskDOM.dtdParser.*;
import au.com.explodingsheep.diskDOM.nodeStore.*;


/**
 * My version of a MyDocumentBuilder.
 */

public class DefaultMyDocumentBuilder extends MyDocumentBuilder
{
  /* The DOMImplementation we use to create and return Document instances. */
  protected MyDOMImplementation domImplementation = null;
  
  /* Used to get the DTD information from the XML document. */
  protected DTDParser dtdParser = null;

  /* When parsing an XML file do we remove any trailing and leading */
  /* spaces or not.                                                  */
  protected boolean trimSpaces = false;

  /* Used in the "newDocument" method to create a unique name for each */
  /* Document we produce.                                               */
  protected int nameCount = 1;
  
  
  /**
   * Constructor.
   */
  
  public DefaultMyDocumentBuilder( MyDOMImplementation domImplementation, boolean trimSpaces )
  {
    super();
    this.domImplementation = domImplementation;
    this.trimSpaces = trimSpaces;
  }


  /**
   * Extended version of "parse" that specifies the Document that the
   * parsed XML nodes will be added to.
   * It also specifies that first root node that all other parsed 
   * nodes will be children of.
   */  
  
  public Node parse( InputSource is, MyDocument document, Node rootNode ) throws SAXException, IOException
  {  
    Node result = null;

    /* Try an get find a "<!DTD ...>" string in the input source.    */
    /* if one is found the process the DTD declaration that is part  */
    /* of this file or in an external file. We then want to rewind    */
    /* the input stream and re-process it for the XML data.           */
    /* To do this we wrap the passed inputstream in a                */
    /* bufferedinputstream so we gain the "rewind" ability.           */
    InputStream inputStream = new BufferedInputStream( is.getByteStream() );
    
    /* Allows us to rewind a maximum of 32k (approx). If the DTD declaration */
    /* is bigger than this then we have problems.                             */
    inputStream.mark( 32000 );
  
    /* Parse the DTD document. */
    try
    {
      /* Parse the inputstream for a DTD declaration. */
      this.dtdParser = new DefaultDTDParser( document, this.domImplementation );
      boolean containsDTD = this.dtdParser.parse( inputStream );
      
      /* Now reset the input stream and parse its data. */
      inputStream.reset();
      InputSource is2 = new InputSource( inputStream );
      is2.setPublicId( is.getPublicId() );
      is2.setSystemId( is.getSystemId() );
      is2.setEncoding( is.getEncoding() );
    
      /* This catches events from the SAXParser and creates a Document */
      /* instance.                                                     */
      MyDefaultHandler dh = new MyDefaultHandler( document, dtdParser, rootNode, trimSpaces );

      /* When DTDParser.parse ran it created a DocumentType and attached it to */
      /* the Document. Get it here and connect it to the XMLReader.             */
      MyDocumentType docType = ( MyDocumentType ) document.getDoctype();
      MyXMLReader myXMLReader = new MyXMLReader( docType );
      myXMLReader.setContentHandler( dh );
      myXMLReader.setCommentHandler( dh );
      myXMLReader.setEntityHandler( dh );
      myXMLReader.setCDATAHandler( dh );
      myXMLReader.parse( is2 );
      result = dh.getFirstNode();
      
//      /* If the root node is a MyDocument then set its documentName */
//      /* to the name of the DTD.                                    */
//      if ( ( rootNode.getNodeType() == Node.DOCUMENT_NODE ) && ( docType != null ) )
//      {
//        String dtdName = docType.getName();
//        MyDocument myDocument = ( MyDocument ) rootNode;
//        try
//        {
//          myDocument.setDocumentName( dtdName );
//        }
//        catch ( MyDOMImplementationException die )
//        {
//          throw new SAXException( die.getMessage() );
//        }
//        catch ( NodeStoreException nse )
//        {
//          throw new SAXException( nse.getMessage() );
//        }
//      }
    }
    catch ( DTDParserException dpe )
    {
      // !!! This is not the best exception to throw.
      throw new SAXException( dpe.getMessage() );
    }
    
    return result;
  }


  /**
   * Extended version of "parse" that specifies the Document that the
   * parsed XML nodes will be added to.
   */  
  
  public Node parse(InputSource is, DefaultMyDocument document ) throws SAXException, IOException
  {  
    Node result = this.parse( is, document, null );
    return result;
  }
  
  
  /**
   * Returns the DTDParsre used when parsing the document.
   */
  
  public DTDParser getDTDParser()
  {
    return this.dtdParser;
  }
  
  
//  /**
//   * Loads and returns the Document instance with the passed Identifier.
//   * This is how we re-load a Document that was created and saved previously.
//   */
//  
//  public MyDocument getDocument( Identifier identifier ) throws NodeStoreException
//  {
//    NodeStore nodeStore = this.domImplementation.getNodeStore();
//    MyDocument result = ( MyDocument ) nodeStore.loadNode( identifier );
//    return result;
//  }


  /*------------------------- DocumentBuilder methods --------------------------*/
  

  /**
   * Indicates whether or not this parser is configured 
   * to understand namespaces.
   */
  
  public boolean isNamespaceAware()
  {
    return true;
  }


  /**
   * Indicates whether or not this parser is configured 
   * to validate XML documents.
   */
  
  public boolean isValidating()  
  {
    return false;
  }
  
  
  /**
   * Specify the EntityResolver to be used to resolve entities 
   * present in the XML document to be parsed. Setting this to 
   * null will result in the underlying implementation using 
   * it's own default implementation and behavior. 
   * 
   * We do nothing with this.
   */
  
  public void setEntityResolver(EntityResolver er)
  {
  }
  
  
  /**
   * Specify the ErrorHandler to be used to report errors present 
   * in the XML document to be parsed. Setting this to null will 
   * result in the underlying implementation using it's own 
   * default implementation and behavior. 
   * 
   * We do nothing with this.
   */
  
  public void setErrorHandler(ErrorHandler eh)
  {
  }
  
  
  /**
   * Obtain a new instance of a DOM Document object to build a 
   * DOM tree with. An alternative way to create a DOM 
   * Document object is to use the getDOMImplementation method 
   * to get a DOM Level 2 DOMImplementation object and then 
   * use DOM Level 2 methods on that object to create a DOM 
   * Document object.
   */
  
  public Document newDocument()
  {
    String documentName = "Document" + nameCount;
    nameCount++;

    Document result = this.domImplementation.createDocument( null, documentName, null /*docType*/ );
    return result;
  }
  
  
  /**
   * Obtain an instance of a DOMImplementation object. 
   */
  
  public DOMImplementation getDOMImplementation()  
  {
    return this.domImplementation;
  }
  

  /**
   * Parse the content of the given input source as an XML document 
   * and return a new DOM Document object.
   */  
  
  public Document parse(InputSource is) throws SAXException, IOException
  {
    /* To implement this we will use an AXL  Transformer to import */
    /* the data from the source into a Document instance.           */
    DefaultMyDocument result = ( DefaultMyDocument) this.newDocument();
    Node firstNode = this.parse( is, result, result );
    return result;
  }
  
  
  /**
   * Parse the content of the given URI as an XML document and return 
   * a new DOM Document object. 
   */
  
  public Document parse(String uri) throws SAXException, IOException
  {
    Document result = null;
    if ( uri != null )
    {
      try
      {
        URI realURI = new URI( uri );
        try
        {
          URL url = realURI.toURL();
          InputStream inputStream = url.openStream();
          try
          {
            result = this.parse( inputStream );
          }
          finally
          {
            inputStream.close();
          }
        }
        catch ( MalformedURLException mue )
        {
          throw new SAXException( mue.getMessage() );
        }
      }
      catch ( java.net.URISyntaxException use )
      {
        throw new SAXException( use.getMessage() );
      }
    }
    else
    {
      throw new IllegalArgumentException( "URI cannot be null." );
    }
    return result;
  }


  /**
   * Parse the content of the given URI as an XML document and return 
   * a new DOM Document object. 
   */
  
  public Document parse( File file ) throws SAXException, IOException
  {
    Document result = null;
    if ( file != null )
    {
      InputStream inputStream = new FileInputStream( file );
      try
      {
        result = this.parse( inputStream );
      }
      finally
      {
        inputStream.close();
      }
    }
    else
    {
      throw new IllegalArgumentException( "The passed file cannot be null." );
    }
    return result;
  }

  
  /*-------------------------- MyDocumentBuilder methods ---------------------------*/
  
  
  /**
   * Closes this class and its DOMImplementation.
   */
  
  public void close() throws MyDocumentBuilderException
  {
    try
    {
      this.domImplementation.close();
      MyDOMImplementationManager.removeInstance( this.domImplementation );
    }
    catch ( MyDOMImplementationException mdie )
    {
      throw new MyDocumentBuilderException( mdie );
    }
  }
  
  
  /**
   * Loads and returns the document with the passed Identifier.
   */
  
  public MyDocument getDocument( String documentName ) throws MyDocumentBuilderException
  {
    MyDocument result = null;
    try
    {
      result = this.domImplementation.getDocument( documentName );
    }
    catch ( MyDOMImplementationException mdie )
    {
      throw new MyDocumentBuilderException( mdie );
    }
    catch ( NodeStoreException nse )
    {
      throw new MyDocumentBuilderException( nse );
    }
    return result;
  }
  
  
  /**
   * Returns the names of all the MyDocuments in this
   * DocumentBuilder.
   */
  
  public String[] getDocumentNames() throws MyDocumentBuilderException
  {
    String[] result = null;
    try
    {
      this.domImplementation.getDocumentNames();
    }
    catch ( MyDOMImplementationException mdie )
    {
      throw new MyDocumentBuilderException( mdie );
    }
    return result;
  }
}
