package au.com.explodingsheep.diskDOM.documentBuilder;


import java.util.*;
import org.xml.sax.helpers.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import au.com.explodingsheep.diskDOM.*;
import au.com.explodingsheep.diskDOM.dtdParser.*;
import au.com.explodingsheep.diskDOM.helpers.TrimString;


/**
 * Handles events from a SAXParser and builds a Node tree of elements,
 * attributes, text nodes etc. "getFirstNode" will return the root first
 * node that was created, that all other nodes hang from. It can then
 * be connected as the root node of a Document.
 * 
 * This class is meant to be used once. If more than one SAX source
 * needs to be parsed new instances of this should be created to
 * handle each.
 */

public class MyDefaultHandler extends DefaultHandler implements CommentHandler, EntityHandler, CDATAHandler
{
  /* The Document we will return after a SAX parser has run. */
  protected Document document = null;

  /* As we are informed of each time we hit the start of a new element */
  /* we want to keep the previous element (this element's parent) in a */
  /* stack.                                                             */
  protected List elements = new ArrayList();
  
  /* The parent element we join new nodes to. */
  protected Node parentNode = null;

  /* Text in the current element. */
  protected String text = "";

  /* Holds the parsed DTD information for the XML document we are processing. */
  protected DTDParser dtdParser = null;
  
  /* The first node we create, that all other nodes hang from. */
  protected Node firstNode = null;

  /* Flag to remember if we are processing entity reference nodes or not. */
  protected boolean processingEntityNodes = false;

//  /* Flag to remember the first entity node in a tree we are creating. */
//  protected MyNode firstEntityNode = null;

  protected DefaultMyEntityReference entityReferenceNode = null;

  /* When parsing an XML file do we remove any trailing and leading */
  /* spaces or not.                                                  */
  protected boolean trimSpaces = false;
  

  /**
   * Constructor.
   */
  
  public MyDefaultHandler( Document document, DTDParser dtdParser, Node firstNode, boolean trimSpaces )
  {
    this.document = document;
    this.dtdParser = dtdParser;
    this.firstNode = firstNode;
    this.parentNode = firstNode;
    this.trimSpaces = trimSpaces;
    if ( this.firstNode != null )
    {
      this.elements.add( this.firstNode );
    }
  }
  

  /**
   * Returns the document we are writing the SAXParser data to.
   */
  
  public Document getDocument()
  {
    return this.document;
  }


  /**
   * Returns the root first node we parsed, that all other
   * nodes hang from.
   */
  
  public Node getFirstNode()
  {
    return this.firstNode;
  }


  /**
   * Marks all the nodes in the passed tree as read-only.
   */
  
  protected void markNodesAsReadOnly( MyNode node )
  {
    if ( node != null )
    {
      node.setReadOnly( true );
      
      NamedNodeMap attributes = node.getAttributes();
      if ( attributes != null )
      {
        int numAttributes = attributes.getLength();
        for ( int loop = 0; loop < numAttributes; loop++ )
        {
          MyAttr attr = ( MyAttr ) attributes.item( loop );
          attr.setReadOnly( true );
        }
      }
      
      NodeList children = node.getChildNodes();
      if ( children != null )
      {
        int numChildren = children.getLength();
        for ( int loop = 0; loop < numChildren; loop++ )
        {
          MyNode child = ( MyNode ) children.item( loop );
          this.markNodesAsReadOnly( child );
        }
      }
    }
  }


  /**
   * Creates a text node given the current contents of "this.text" and
   * appends it to the current parent.
   */
  
  protected void createTextNodeIfNecessary()
  {
    if ( this.trimSpaces == true )
    {
      this.text = TrimString.trimString( this.text );
    }

    /* If we have collected text then create a Text node and join */
    /* it to the current Element.                                  */
    if ( ( this.text != null ) && ( this.text.equals( "" ) == false ) )
    {
      MyNode textNode = ( MyNode ) this.document.createTextNode( this.text );
      this.parentNode.appendChild( textNode );
      this.text = "";
    }
  }
  

  /*-------------------------- DefaultHandler overridden methods ------------------*/


  /**
   * Called when the SAXParser hits an element.
   * We want to create a matching element in our document.
   */
  
  public void startElement(String uri,
                            String localName,
                            String qName,
                            Attributes attributes) throws SAXException
  {
    try
    {
      /* If we have collected text then create a Text node and join */
      /* it to the current Element.                                  */
      this.createTextNodeIfNecessary();
  
      MyElement e = ( MyElement ) this.document.createElement( qName );
      if ( this.firstNode == null )
      {
        this.firstNode = e;
      }
      else
      {
        this.parentNode.appendChild( e );
      }
      
      this.elements.add( e );
       this.parentNode = e;
     
      /* Create attributes for the element. */
      if ( attributes != null )
      {
        int numAttributes = attributes.getLength();
        for ( int loop = 0; loop < numAttributes; loop++ )
        {
          String attrLocalName = attributes.getLocalName( loop );
          String attrQName = attributes.getQName( loop );
          String attrValue = attributes.getValue( loop );
          
          MyAttr attr = ( MyAttr ) this.document.createAttribute( attrQName );
          e.setAttributeNode( attr );

          // ??? Should this be setNodeValue or setDefaultValue?
          // If we setDefaultValue here then all attributes will have an
          // initial specified value of false. How do we find out if an
          // attribute's value is defined or not? 
          attr.setInitialValue( attrValue );
        }
      }
    }
    catch ( DOMException de )
    {
      throw new SAXException( de );
    }
  }


  /**
   * Receive notification of the end of an element.
   */

  public void endElement(String uri, String localName, String qName) throws SAXException
  {
    /* If we have collected text then create a Text node and join */
    /* it to the current Element.                                  */
    this.createTextNodeIfNecessary();

    int numNodes = this.elements.size();
    this.elements.remove( numNodes - 1 );
    numNodes--;
    if ( numNodes > 0 )
      this.parentNode = ( Node ) this.elements.get( numNodes - 1 );
    else
      this.parentNode = null;
  }
    

  /**
   * Receive notification of character data inside an element. 
   */
    
  public void characters(char[] ch, int start, int length) throws SAXException
  {
    try
    {
      if ( this.parentNode != null )
      {
        String s = "";
        for ( int loop = start; loop < start + length; loop++ )
        {
          char c = ch[loop];
            s += c;
        }
        if ( ( s != null ) && ( s.equals( "" ) == false ) )
        {
          this.createTextNodeIfNecessary();
          this.text = s;
        }
      }
    }
    catch ( DOMException de )
    {
      throw new SAXException( de );
    }
  }
  
  
  /**
   * Called when a processing instruction is encountered.
   */
  
  public void processingInstruction(String target, String data) throws SAXException
  {
    MyProcessingInstruction pi = ( MyProcessingInstruction ) this.document.createProcessingInstruction( target, data );
    if ( this.firstNode == null )
    {
      this.firstNode = pi;
    }
    else
    if ( this.parentNode != null )
    {
      this.parentNode.appendChild( pi );
    }
  }

  
//  /**
//   * Allow the application to resolve external entities.
//   */
//  
//  public InputSource resolveEntity(String publicId, String systemId) throws SAXException
//   {
//     org.xml.sax.InputSource result = null;
//     
//    System.out.println( "resolveEntity - \"" + publicId + "\" - \"" + systemId + "\"." );
//     
//     try
//     {
//       java.net.URI uri = new URI( systemId );
//       String fileName = uri.getPath();
//       InputStream is = new FileInputStream( fileName );
//      result = new org.xml.sax.InputSource( is );
//     }
//     catch ( FileNotFoundException fnfe )
//     {
//       fnfe.printStackTrace();
//     }
//     catch ( java.net.URISyntaxException use )
//     {
//       use.printStackTrace();
//     }
//     
//    return result;
//   }
//   
//   
//   public void notationDecl(String name,
//                         String publicId,
//                         String systemId)
//                  throws SAXException
//  {
//    System.out.println( "notationDecl - \"" + publicId + "\" - \"" + systemId + "\"." );
//  }
//  
//  
//  public void unparsedEntityDecl(String name,
//                                 String publicId,
//                                 String systemId,
//                                 String notationName)
//                          throws SAXException
//  {
//    System.out.println( "unparsedEntityDecl - \"" + publicId + "\" - \"" + systemId + "\" - \"" + notationName + "\"." );
//  }


  /*------------------------- CommentHandler methods ---------------------------*/
  
  
  /**
   * Called when MyXMLReader encounters a comment.
   */
  
  public void comment( String commentString )
  {
    Comment comment = this.document.createComment( commentString );
    if ( this.parentNode != null )
    {
      this.parentNode.appendChild( comment );
    }
    else
    if ( this.document != null )
    {
      this.document.appendChild( comment );
    }
  }
  
  
  /*------------------------- EntityHandler methods ---------------------------*/
  

  /**
   * Called when the MyXMLReader has just hit an entity.
   */
  
  public void entity( String entityName )
  {
    MyEntityReference er = ( MyEntityReference) this.document.createEntityReference( entityName );

    if ( this.firstNode == null )
    {
      this.firstNode = er;
    }
    else
    {
      this.parentNode.appendChild( er );
    }
    
    this.markNodesAsReadOnly( er );
  }
  
  
  /*----------------------- CDATAHandler methods ------------------------------*/
  

  /**
   * Called by MyXMLReader when it encounters a CDATA section.
   */

  public void cdata( String cdataString )
  {
    /* If we have collected text then create a Text node and join */
    /* it to the current Element.                                  */
    this.createTextNodeIfNecessary();

    MyCDATASection cDATASection = ( MyCDATASection ) this.document.createCDATASection( cdataString );
    if ( this.parentNode != null )
    {
      this.parentNode.appendChild( cDATASection );
    }
    else
    if ( this.document != null )
    {
      this.document.appendChild( cDATASection );
    }
  }
}
