package au.com.explodingsheep.diskDOM.documentBuilder;


import java.io.*;
import org.xml.sax.*;
import org.xml.sax.helpers.AttributesImpl;
import org.w3c.dom.*;
import au.com.explodingsheep.diskDOM.*;


/**
 * My implementation of the XMLReader interface. It parses an
 * XML file and generates events as it processes elements
 * and attributes.
 * 
 * It is assumed only one thread runs through this.
 * 
 * !!! Note that the naming of "MyXMLReader" is wrong, it 
 * should be "DefaultMyXMLReader".
 */

public class MyXMLReader implements XMLReader
{
  protected ContentHandler contentHandler = null;
  protected DTDHandler dtdHandler = null;
  protected ErrorHandler errorHandler = null;
  protected EntityResolver entityResolver = null;

  protected MyDocumentType documentType = null;

  protected CommentHandler commentHandler = null;
  
  protected EntityHandler entityHandler = null;
  
  protected CDATAHandler cDATAHandler = null;

  protected InputStream inputStream = null;
  
  /* Unget character buffer. */
  protected String ungetBuffer = "";
  

  /**
   * Constructor.
   */
  
  public MyXMLReader( MyDocumentType documentType )
  {
    this.documentType = documentType;
  }

  
  /**
   * Reads a character from the input stream.
   */
  
  protected int getChar() throws IOException
  {
    int result = -1;
    if ( this.ungetBuffer.equals( "" ) )
    {
      if ( this.inputStream != null )
      {
        result = this.inputStream.read();
      }
    }
    else
    {
      result = ( int ) this.ungetBuffer.charAt( 0 );
      if ( this.ungetBuffer.length() > 1 )
      {
        this.ungetBuffer = this.ungetBuffer.substring( 1 );
      }
      else
      {
        this.ungetBuffer = "";
      }
    }
    
    return result;
  }
  
  
  /**
   * Ungets a character.
   */
  
  protected void ungetChar( int i )
  {
    this.ungetBuffer += ( char ) i;
  }


  /**
   * Ungets a string.
   */
  
  protected void ungetString( String s )
  {  
    if ( s != null )
    {
      int length = s.length();
      for ( int loop = 0; loop < length; loop++ )
      {
        char ch = s.charAt( loop );
        this.ungetChar( ( int ) ch );
      }
    }
  }
  

  /**
   * Reads and returns a string from the inputstream until we hit 
   * the passed terminating string. If the terminating string is not 
   * found before the end of the input stream then return everything 
   * collected up to that point.
   */
  
  protected String readStringUntilTerminatingStringFound( String terminatingString ) throws IOException
  {
    String result = "";
    String bfr = "";
    int matchCount = 0;
    int terminatingStringLength = terminatingString.length();
    int i = this.getChar();
    boolean go = true;
    while ( ( go == true ) && ( i != -1 ) )
    {
      int i2 = terminatingString.charAt( matchCount );
      if ( i == i2 )
      {
        bfr += ( char ) i;
        matchCount++;
        if ( matchCount == terminatingStringLength )
        {
          /* We have hit the terminating string. Exit this method. */
          go = false;
        }
      }
      else
      {
        /* Have we collected any characters into bfr? */
        if ( matchCount > 0 )
        {
          result += bfr;
          bfr = "";
          matchCount = 0;
        }
        
        result += ( char ) i;
      }
      
      if ( go == true )
      {
        i = this.getChar();
      }
    }
    
    return result;
  }  


  /**
   * Parses the passed string that contains the one element
   * and call the appropriate events. Ie "elementString" contains
   * "<abc d="efg">" or "</ghi>" or "<jkl/>".
   */
  
  protected void createElement( String elementString ) throws SAXException
  {
    int elementStringLength = elementString.length();
    int loop = 0;

    /* This is where we hold the list of attributes we are about to */
    /* collect, and pass to the content handler.                    */
    AttributesImpl attributes = new AttributesImpl();

    /* Is this element empty (ie "</abc> versus "<abc>def</abc>"? */
    boolean emptyElement = false;

    /* Is this an ending element (ie "</abc>" versus "<abc>"? */
    boolean endingElement = false;
  
    /* First collect the element name. */
    String elementName = "";
    char ch = 'a';
    while ( ( ch != ' ' ) && ( ch != '>' ) && ( loop < elementStringLength ) )
    {
      ch = elementString.charAt( loop );
      if ( ch == '/' )
      {
        if ( loop == 0 )
          endingElement = true;
        else
          emptyElement = true;
      }
      else
      if ( ch != ' ' )
      {
        elementName += ch;
      }
      
      loop++;
    }
    
    /* If we hit a space then attributes are following so process them. */
    while ( ch == ' ' )
    {
      /* First will be the name of the attribute. */
      String attributeName = "";
      ch = elementString.charAt( loop );
      while ( ch != '=' )
      {
        attributeName += ch;
        loop++;
        ch = elementString.charAt( loop );
      }
      
      /* Next is a doublequote followed by the value of the attribute. */
      String attributeValue = "";
      loop += 2;
      ch = elementString.charAt( loop );
      while ( ( ch != '\"' ) && ( ch != '\'' ) )
      {
        attributeValue += ch;
        loop++;
        ch = elementString.charAt( loop );
      }

      /* Add the newly collected attribute information to the list. */
      attributes.addAttribute( "", attributeName, attributeName, "", attributeValue );
      
      /* Next should be either a space or a '>'. */
      loop++;
      if ( loop < elementStringLength )
      {
        ch = elementString.charAt( loop );
        if ( ch == ' ' )
          loop++;
      }
    }
  
    /* Generate an event. */
    if ( this.contentHandler != null )
    {
      if ( endingElement == false )
      {
        this.contentHandler.startElement( null, elementName, elementName, attributes );
      }
      
      /* If this is an empty element then also call the "endElement" event. */
      if ( ( endingElement == true ) || ( emptyElement == true ) )
      {
        this.contentHandler.endElement( null, elementName, elementName );
      }
    }
  }


  /**
   * Called when we encounter an entity reference.
   * Looks up the Entity instance in the DocumentType
   * for the passed entity name, then calls then entity handler event.
   */
  
  protected void processEntity( String entityName ) throws SAXException
  {
    if ( this.documentType != null )
    {
      NamedNodeMap entities = this.documentType.getEntities();
      Entity entity = ( Entity ) entities.getNamedItem( entityName );
      if ( entity != null )
      {
        /* The entity contains one or more element nodes. */
        /* Generate events for each one.                  */
        
        /* Inform any listeners that we are about to handle entity nodes. */
        if ( this.entityHandler != null )
        {
          this.entityHandler.entity( entityName );
        }
      }
      else
      {
        /* !!! What do we do here? We cannot find the entity */
        /* referred to in the XML file. What event should we */
        /* call?                                             */
        if ( this.contentHandler != null )
        {
          this.contentHandler.skippedEntity( entityName );
        }
      }
    }
    else
    {
      /* Without the DTDParser we cannot resolve the entity.     */
      /* !!! We should really try to resolve the entity, but for */
      /* the moment just call the "skippedEntity" event.          */
      if ( this.contentHandler != null )
      {
        this.contentHandler.skippedEntity( entityName );
      }
    }
  }


  /**
   * Called when we entounter text between elements.
   * We look for embedded entity references and process
   * them, then generate a "text" event for the DefaultHandler.
   */
  
  protected void createTextNodes( String textString ) throws SAXException
  {
    int textStringLength = textString.length();
    if ( textStringLength > 0 )
    {
      /* We want to spin through the passed string calling */
      /* contenthandler.character events and dealing with  */
      /* any entities found.                               */
      String textNodeString = "";
      int loop = 0;
      while ( loop < textStringLength )
      {
        char ch = textString.charAt( loop );
        if ( ch != '&' )
        {
          textNodeString += ch;
        }
        else
        {
          /* We have hit the start of an element. Call the "characters" */
          /* event for any characters collected so far.                  */    
          if ( textNodeString.equals( "" ) == false )
          {
            if ( this.contentHandler != null )
            {
              char[] chars = textNodeString.toCharArray();
              this.contentHandler.characters( chars, 0, chars.length );
            }
            textNodeString = "";
          }
          
          /* Collect the name of the entity. */
          String entityName = "";
          loop++;
          ch = textString.charAt( loop );
          while ( ch != ';' )
          {
            entityName += ch;
            loop++;
            ch = textString.charAt( loop );
          }
          
          /* Handle the entity. */
          this.processEntity( entityName );
        }
        
        loop++;
      }
      
      /* Do we have any text collected that needs a "characters" */
      /* event generated for?                                     */
      if ( textNodeString.equals( "" ) == false )
      {
        if ( this.contentHandler != null )
        {
          char[] chars = textNodeString.toCharArray();
          this.contentHandler.characters( chars, 0, chars.length );
        }
      }
    }
  }
  
  
  /**
   * Handles an XML directive (what I call the contents of a "<! ... >" thing.
   * Currrently just eat it and ignore it. Because one of the can have one
   * or more sub "< ... >" things in it keep track of the number of "<" and
   * ">" encountered so we eat everything in the outer "<! ... >".
   */
  
  protected void processXMLDirective() throws IOException
  {
    /* This hold the number of "<" encountered. */
    int bracketCount = 0;
    while ( bracketCount >= 0 )
    {
      int i = this.getChar();
      if ( i == '<' )
        bracketCount++;
      else
      if ( i == '>' )
        bracketCount--;
    }
  }
  
  
  /**
   * Called when we encounter a CDATA thiing in the XML data
   * we are parsing. Calls the CDATA event in the MyDefautHandler.
   */
  
  protected void processCDATA() throws IOException
  {
    String cData = this.readStringUntilTerminatingStringFound( "]]>" );
    if ( this.cDATAHandler != null )
    {
      this.cDATAHandler.cdata( cData );
    }
  }
  
  
  /**
   * Called when we encounter a ProcessingInstruction while
   * parsing the XMl data. Calls the "processingInstruction" event.
   */
  
  protected void createProcessingInstruction() throws IOException, SAXException
  {
    /* First should be a target string, followed by a data string. */
    String target = this.readStringUntilTerminatingStringFound( " " );
    if ( target.equals( "xml" ) == false )
    {
      String data = this.readStringUntilTerminatingStringFound( "?>" );
      if ( this.contentHandler != null )
      {
        this.contentHandler.processingInstruction( target, data );
      }
    }
  }
  
   
  /**
   * This does the real parsing of the passed inputstream
   * containing the XML nodes.
   * All other parse methods call this.
   */
   
  protected void parse( InputStream inputStream ) throws IOException, SAXException
  {
    this.inputStream = inputStream;
    
    /* We want to spin through the passed input stream, handling all the  */
    /* XML types we come across. The things we can handle are:            */
    /* -Elements                                                          */
    /* -Text, including embedded entities                                  */
    /* -Entity references                                                  */
    /* -Processing instructions                                            */
    int i = this.getChar();
    while ( i != -1 )
    {
      /* The first character must either be a "<" or a text character. */
      if ( i == '<' )
      {
        /* The following character tells us if this is an element, or */
        /* either a processing instruction or a comment.              */
        i = this.getChar();
        if ( i == '!' )
        {
          /* We have either an XML directive or a comment. */
          i = this.getChar();
          if ( i == '-' )
          {
            /* We most likely have a comment. Read the next character to */
            /* make sure.                                                 */
            i = this.getChar();
            if ( i == '-' )
            {
              /* We have a comment. Read it and call the appropriate event. */
              String comment = this.readStringUntilTerminatingStringFound( "-->" );
              
              if ( this.commentHandler != null )
              {
                this.commentHandler.comment( comment );
              }
            }
            else
            {
              /* We have an element whose name I think is an illegal. */
              /* For the moment do nothing.                            */
              /* Eat it and for the moment do nothing.                 */
              this.readStringUntilTerminatingStringFound( ">" );
            }
          }
          else
          {
            /* We have an XML directive. Is it a CDATA node? */
            this.ungetChar( i );
            String bfr = "";
            bfr += ( char ) this.getChar();
            bfr += ( char ) this.getChar();
            bfr += ( char ) this.getChar();
            bfr += ( char ) this.getChar();
            bfr += ( char ) this.getChar();
            bfr += ( char ) this.getChar();
            bfr += ( char ) this.getChar();
            if ( bfr.equals( "[CDATA[" ) )
            {
              this.processCDATA();
            }
            else
            {
              /* We have some other XML directive. Eat and ignore it. */
              this.ungetString( bfr );
              this.processXMLDirective();
            }
          }
        }
        else
        if ( i == '?' )
        {
          /* We have a processing instruction.     */
          /* Create a processing instruction node. */
          this.createProcessingInstruction();
        }
        else
        {
          /* We have an element. Collect it up and process it. */
          this.ungetChar( i );
          String elementString = this.readStringUntilTerminatingStringFound( ">" );
          this.createElement( elementString );
        }
      }
      else
      {
        /* We have text. There may be an entity amongst it so look out for */
        /* and handle it.                                                   */
        String textString = "";
        while ( ( i != '<' ) && ( i != -1 ) )
        {
          textString += ( char ) i;
          i = this.getChar();
        }
        if ( i == '<' )
        {
          this.ungetChar( i );
        }
        
        /* Create the text nodes. */
        this.createTextNodes( textString );
      }
  
      i = this.getChar();
    }
  }
   
   
   /**
    * Sets a comment handler.
    */
   
   public void setCommentHandler( CommentHandler commentHandler )
   {
     this.commentHandler = commentHandler;
   }
   
   
   /**
    * Returns the comment handler.
    */
   
   public CommentHandler getCommentHandler()
   {
     return this.commentHandler;
   }


  /**
   * Sets the entity handler.
   */       
  
  public void setEntityHandler( EntityHandler entityHandler )
  {
    this.entityHandler = entityHandler;
  }
  
  
  /**
   * Returns the entity flag handler.
   */
  
  public EntityHandler getEntityHandler()
  {
    return this.entityHandler;
  }
  
  
  /**
   * Sets the cdata handler.
   */
  
  public void setCDATAHandler( CDATAHandler cDATAHandler )
  {
    this.cDATAHandler = cDATAHandler;
  }
  
  
  /**
   * Returns the cdata handler.
   */
  
  public CDATAHandler getCDATAHandler()
  {
    return this.cDATAHandler;
  }
  

  /*------------------------- XMLReader methods ---------------------------*/
  
  
  /**
   * Return the current content handler. 
   */
   
  public ContentHandler getContentHandler()
  {
    return this.contentHandler;
  }
  

  /**
   * Return the current DTD handler. 
   */
   
  public DTDHandler getDTDHandler() 
  {
    return this.dtdHandler;
  }
  
 
  /**
   * Return the current entity resolver. 
   */
   
  public EntityResolver getEntityResolver() 
  {
    return this.entityResolver;
  }
  
 
  /**
   * Return the current error handler. 
   */
   
  public ErrorHandler getErrorHandler() 
  {
    return this.errorHandler;
  }
  
 
  /**
   * Look up the value of a feature. 
   */
   
  public boolean getFeature(String name) throws SAXNotRecognizedException, SAXNotSupportedException
  {
    return false;
  }
 
 
  /**
   * Look up the value of a property. 
   */
   
  public Object getProperty(String name) throws SAXNotRecognizedException, SAXNotSupportedException
  {
    return null;
  }
  
 
  /**
   * Parse an XML document. 
   */
   
  public void parse(InputSource input) throws IOException, SAXException
  {
    InputStream is = input.getByteStream();
    this.parse( is );
  }
  
 
  /**
   * Parse an XML document from a system identifier (URI). 
   */
   
  public void parse(String systemId) throws IOException, SAXException
  {
    InputSource inputSource = new InputSource( systemId );
    this.parse( inputSource );
  }
 
  /**
   * Allow an application to register a content event handler. 
   */
   
  public void setContentHandler(ContentHandler handler) 
  {
    this.contentHandler = handler;
  }
  
 
  /**
   * Allow an application to register a DTD event handler. 
   */
   
  public void setDTDHandler(DTDHandler handler) 
  {
    this.dtdHandler = handler;
  }
  
 
  /**
   * Allow an application to register an entity resolver. 
   */
   
  public void setEntityResolver(EntityResolver resolver) 
  {
    this.entityResolver = resolver;
  }
  
 
  /**
   * Allow an application to register an error event handler. 
   */
   
  public void setErrorHandler(ErrorHandler handler) 
  {
    this.errorHandler = handler;
  }
 
 
  /**
   * Set the state of a feature. 
   */
   
  public void setFeature(String name, boolean value) throws SAXNotRecognizedException, SAXNotSupportedException
  {
  }

 
  /**
   * Set the value of a property. 
   */
   
  public void setProperty(String name, Object value) 
  {
  }
 }
