package au.com.explodingsheep.diskDOM.dtdParser;


import java.util.*;
import java.io.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.w3c.dom.*;
import au.com.explodingsheep.diskDOM.*;
import au.com.explodingsheep.diskDOM.documentBuilder.*;
import au.com.explodingsheep.diskDOM.identifier.*;
import au.com.explodingsheep.diskDOM.nodeStore.*;


/**
 * Implements the DTDParser interface.
 */

public class DefaultDTDParser implements DTDParser, Serializable
{
  /* The string that starts the DTD declaration line. */    
  protected static String dtdDeclaration = "<!DOCTYPE";
  
    /* List of token strings. */
  protected String ELEMENT_START = "<!ELEMENT";
  protected String ATTRIBUTE_LIST_START = "<!ATTLIST";
  protected String ENTITY_START = "<!ENTITY";
  protected String NOTATION_START = "<!NOTATION";
  protected String NOTATION_TYPE_SYSTEM = "SYSTEM";
  protected String NOTATION_TYPE_PUBLIC = "PUBLIC";
  protected String END = ">";
  protected String OR = "|";
  protected String ONE_OR_MORE = "+";
  protected String ZERO_OR_ONE = "?";
  protected String ZERO_OR_MORE = "*";
  protected String LEFT_SQUARE_BRACKET = "[";
  protected String RIGHT_SQUARE_BRACKET = "[";
  protected String LEFT_ROUND_BRACKET = "(";
  protected String RIGHT_ROUND_BRACKET = ")";
  protected String COMMA = ",";
  protected String ELEMENT_DATATYPE_CDATA = "#CDATA";
  protected String ELEMENT_DATATYPE_PCDATA = "#PCDATA";
  protected String ATTRIBUTE_DATATYPE_CDATA = "CDATA";
  protected String ATTRIBUTE_DATATYPE_PCDATA = "PCDATA";
  protected String ATTRIBUTE_REQUIRED = "#REQUIRED";
  protected String ATTRIBUTE_IMPLIED = "#IMPLIED";
  protected String DTD_SYSTEM = "SYSTEM";
  protected String[] tokens = { ELEMENT_START, 
                                 ATTRIBUTE_LIST_START, 
                                 ENTITY_START,
                                  NOTATION_START,
                                 NOTATION_TYPE_SYSTEM,
                                 NOTATION_TYPE_PUBLIC,
                                 END, 
                                 OR,
                                 ONE_OR_MORE, 
                                 ZERO_OR_ONE, 
                                 ZERO_OR_MORE, 
//                                 LEFT_SQUARE_BRACKET, 
//                                 RIGHT_SQUARE_BRACKET, 
//                                 LEFT_ROUND_BRACKET, 
//                                 RIGHT_ROUND_BRACKET, 
                                 ELEMENT_DATATYPE_CDATA,
                                 ELEMENT_DATATYPE_PCDATA,
                                 COMMA,
                                 ATTRIBUTE_DATATYPE_CDATA,
                                 ATTRIBUTE_DATATYPE_PCDATA,
                                 ATTRIBUTE_REQUIRED,
                                 ATTRIBUTE_IMPLIED,
                                 DTD_SYSTEM };

  /* Whitespace characters separate everything else. They are always eaten and ignored. */
  protected char[] whiteSpaceChars = { ' ', '\n', '\r', '\t' };

  /* String terminating characters end strings. When we hit them in "getNextCharacter" */
  /* we put them back so they can be read with "readTerminatingCharacter".             */
  protected char[] terminatingChars = { ',', '?', '[', ']', '(', ')', '+', '|' };

  
  /* The list of DTDElements we parse. */
  protected transient List elements = new ArrayList();

  /* List of DTDAttributeList instances we parse. */
   protected transient List attributeLists = new ArrayList();

  /* Ungot characters. */
  protected transient String ungotChars = "";
  
  /* Document used to create entity nodes. */
  protected transient MyDocument document = null;
  
  /* Used if we need to create a DocumentType because the XML file */
  /* we are parsing contains a DTD declaration and we need to add  */
  /* a DocumentType to the Document.                               */
  protected transient DOMImplementation domImplementation = null;
  
  /* The name of the dtd document - the part in "[!DOCTYPE name...". */
  protected String dtdName = null;

  /* We load entity and notation nodes from here. */
  protected transient NodeStore nodeStore = null;

  /* What is returned in the "getNextString" method. */
  protected class NextString
  {
    /* What the next collected string is. */
    public static final int RESERVED_WORD = 1;
    public static final int STRING = 2;
    public static final int TERMINATOR = 3;
    
    public int stringType = RESERVED_WORD;
    public String value = null;
  }
  
    
  /**
   * Constructor. The passed Document is what we use to create Element and
   * Text nodes that go into a Entity.
   */
  
  public DefaultDTDParser( MyDocument document, DOMImplementation domImplementation )
  {
    this.document = document;
    this.domImplementation = domImplementation;
  }

  
  /**
   * Spins through the passed inputstream looking for the string "<!DOCTYPE "
   * If it is found the filename of the DTD declaration is found and
   * returned.
   */
  
  protected boolean findDTDLine( InputStream is ) throws DTDParserException, IOException
  {
    boolean result = false;
    String bfr = "";
  
    /* We are looking for the string "<!DOCTYPE " in the inputstream. */
    /* Only search the first 10000 characters for this string. If it  */
    /* not in there then return null.                                  */
    int byteSearchCount = 0;
    int i = is.read();
    while ( ( result          == false ) 
          && ( byteSearchCount <  10000 ) 
          && ( i               != -1    ) )
    {
      /* If we have more than "<!DOCTYPE " characters in bfr then */
      /* remove the first character and move the rest left.        */
      if ( bfr.length() >= DefaultDTDParser.dtdDeclaration.length() )
      {
        String s = bfr.substring( 1 );
        bfr = s;
      }
      
      bfr += ( char ) i;
      if ( bfr.equals( DefaultDTDParser.dtdDeclaration ) )
      {
        result = true;
      }
      else
      {
        byteSearchCount++;  
        i = is.read();
      }
    }

    return result;
  }

  
  /**
   * Returns an already parsed DefaultDTDElement whose name 
   * matches that passed or null if no match.
   */
  
  protected DefaultDTDElement getElement( String elementName )
  {
    DefaultDTDElement result = null;
    int loop = 0;
    int numElements = this.elements.size();
    while ( loop < numElements )
    {
      DefaultDTDElement element = ( DefaultDTDElement ) this.elements.get( loop );
      String s = element.getName();
      if ( s.equals( elementName ) )
      {
        result = element;
        loop = numElements;
      }
      else
      {
        loop++;
      }
    }
    return result;
  }


  /**
   * Gets the next character from the inputstream, or if "ungotChar" is
   * not -1 then return it instead.
   */
  
  protected int getChar( InputStream is ) throws IOException
  {
    int result = -1;
    if ( this.ungotChars.equals( "" ) == false )
    {
      result = this.ungotChars.charAt( 0 );
      if ( this.ungotChars.length() > 1 )
      {
        this.ungotChars = this.ungotChars.substring( 1 );
      }
      else
      {
        this.ungotChars = "";
      }
    }
    else
    {
      result = is.read();
    }
    return result;
  }
  
  
  /**
   * Makes the passed character the next to be returned in "getChar".
   */
  
  public void ungetChar( int i )
  {
    this.ungotChars += ( char ) i;
  }
  
  
  /**
   * Ungets the passed string.
   */
  
  public void ungetString( String s )
  {
    this.ungotChars += s;
  }
  

  /**
   * Returns true if the passed character is a whitespace character.
   */
  
  protected boolean isCharAWhiteSpace( char ch )
  {
    boolean result = false;
    int loop = 0;
    while ( loop < this.whiteSpaceChars.length )
    {
      if ( ch == this.whiteSpaceChars[loop] )
      {
        result = true;
        loop = this.whiteSpaceChars.length;
      }
      else
      {
        loop++;
      }
    }
    return result;
  }
  
  
  /**
   * Returns the matching token if the passed string is in 
   * the list of tokens.
   */
  
  protected String isStringAToken( String s )
  {
    String result = null;
    int loop = 0;
    while ( loop < this.tokens.length )
    {
      if ( s.equals( this.tokens[loop] ) == true )
      {
        result = this.tokens[loop];
        loop = this.tokens.length;
      }
      else
      {
        loop++;
      }
    }
    return result;
  }

  
  /**
   * Is the passed character one that ends a string
   * we are collecting.
   */
  
  protected boolean isCharATerminatingChar( char ch )
  {
    boolean result = false;
    int loop = 0;
    while ( loop < this.terminatingChars.length )
    {
      if ( ch == this.terminatingChars[loop] )
      {
        result = true;
        loop = this.terminatingChars.length;
      }
      else
      {
        loop++;
      }
    }
    return result;
  }

  
  /**
   * Is the passed character one that needs a matching 
   * ending character? Ie "<" and ">" or "(" and ")".
   */
  
  protected boolean isCharAStartingChar( char ch )
  {
    boolean result = false;
    if ( /*( ch == '(' ) ||*/ ( ch == '\"' ) || ( ch == '\'' ) )
      result = true;
    return result;
  }

  
  /**
   * Is the passed character one one that terminates a string?
   */
  
  protected char getMatchingEndingChar( char ch )
  {
    char result = 0;
//    if ( ch == '(' )
//      result = ')';
//    else
    if ( ch == '\"' )
      result = '\"';
    else
    if ( ch == '\'' )
      result = '\'';
    return result;
  }

  
  /**
   * Expects the next character to be a terminating character. 
   * If it is not (and it is also not a whitespace (we eat as many
   * of them as required to get to the next valid character)) 
   * then return throw an exception.
   */
  
  protected char readTerminatingChar( InputStream is ) throws DTDParserException
  {
    char result = 0;
    boolean go = true;
    do
    {
      try
      {
        char ch = ( char ) this.getChar( is );
      
        if ( this.isCharATerminatingChar( ch ) == true )
        {
          go = false;
          result = ch;
        }
        else
        if( this.isCharAWhiteSpace( ch ) == false )
        {
          throw new DTDParserException( "The next character was not a terminating character." );
        }  
      }
      catch ( IOException ioe )
      {
        throw new DTDParserException( ioe );
      }
    }
    while ( go == true );
    return result;
  }


  /**
   * Expects the next character to be a certain terminating character.
   */

  protected void readTerminatingChar( InputStream is, char expectedTerminatingChar ) throws DTDParserException
  {
    char gotTerminatingCharacter = this.readTerminatingChar( is );
    if ( gotTerminatingCharacter != expectedTerminatingChar )
    {
      throw new DTDParserException( "The next terminating character read ('" + gotTerminatingCharacter + "') was not the character we wanted ('" + expectedTerminatingChar + "')." );
    }
  }

    
  /**
   * Expects the next character to be one of a certain terminating character.
   */

  protected char readTerminatingChar( InputStream is, char[] expectedTerminatingChars ) throws DTDParserException
  {
    char result = 0;
    char gotTerminatingCharacter = this.readTerminatingChar( is );
    int loop = 0;
    while ( loop < expectedTerminatingChars.length )
    {
      if ( gotTerminatingCharacter == expectedTerminatingChars[loop] )
      {
        result = gotTerminatingCharacter;
        loop = expectedTerminatingChars.length;
      }
      else
      {
        loop++;
      }
    }
    if ( result == 0 )
    {
      throw new DTDParserException( "The next terminating character read ('" + gotTerminatingCharacter + "') was not in the list of expected characters." );
    }
    return result;
  }


  /**
   * Expects the next character to be one of a the passed list.
   * This is a quick and dirty hack to get "handleChildElementListDeclaration"
   * working correctly. This whole thing needs to be re-written 
   * so we don't have "terminating characters" versus "reserved words".
   */

  protected char readSpecialChar( InputStream is, char[] expectedSpecialChars ) throws DTDParserException
  {
    char result = 0;
    try
    {
      int i = is.read();
      while ( result == 0 )
      {
        char ch = ( char ) i;
        if ( this.isCharAWhiteSpace( ch ) == false )
        {
          int loop = 0;
          while ( loop < expectedSpecialChars.length )
          {
            if ( ch == expectedSpecialChars[loop] )
            {
              result = ch;
              loop = expectedSpecialChars.length;
            }
            else
            {
              loop++;
            }
          }
          if ( result == 0 )
          {
            throw new DTDParserException( "The next character read ('" + ch + "') was not in the list of expected characters." );
          }
        }
        else
        {
          i = is.read();
        }
      }
    }
    catch ( IOException ioe )
    {
      throw new DTDParserException( ioe );
    }
    return result;
  }


  /**
   * Returns the next string from the inputstream or null if we have
   * run out of data.
   * A string is either one of the reserved words in "tokens" or
   * a string terminated by a terminating character.
   */  
      
  protected NextString getNextString( InputStream is ) throws DTDParserException
  {
    NextString result = null;
    
    /* When we hold a character that needs a matching closing character */
    /* like '<' and '>' or two double quotes, the second character (the */
    /* '>') is added to this stack, so we don't stop collecting         */
    /* characters for the string until the final terminating character  */
    /* has been read.                                                    */
    List endingCharStack = new ArrayList();
    
    /* This is the next ending character we are looking for. */
    char endingChar = 0;
    
    /* Loop collecting characters until we either have a token or */
    /* a whitespace separated string.                              */
    try
    {
      String s = "";
      boolean go = true;
      int i = this.getChar( is );
      while ( ( go == true ) && ( i != -1 ) )
      {
        char ch = ( char ) i;
        
        if ( ch == endingChar )
        {
          s += ch;
          int stackSize = endingCharStack.size();
          if ( stackSize > 0 )
          {
            Character character = ( Character ) endingCharStack.get( stackSize - 1 );
            endingChar = character.charValue();
            endingCharStack.remove( stackSize - 1 );
          }
          else
          {
            /* We have hit the final ending character, and thus the */
            /* string we want to collect. Exit.                      */
            result = new NextString();
            result.value = s;
            result.stringType = NextString.STRING;
            go = false;
          }
        }
        else        
        if ( this.isCharAStartingChar( ch ) == true )
        {
          s += ch;
          if ( endingChar != 0 )
          {
            endingCharStack.add( new Character( endingChar ) );
          }
          endingChar = this.getMatchingEndingChar( ch );
        }
        else
        if ( this.isCharAWhiteSpace( ch ) == true )
        {
          /* If we have not collected any characters at all then */
          /* we don't want to exit this loop.                     */
          /* Also don't exit if we are still waiting to reading  */
          /* an ending character.                                 */
          if ( s.equals( "" ) == false ) 
          {
            if ( endingChar == 0 )
            {
              /* We have hit a whitespace character. Don't add that */
              /* character to the resulting string and exit.        */
              result = new NextString();
              result.value = s;
              result.stringType = NextString.STRING;
              go = false;
            }
            else
            {
              /* We are in the middle of a doublequote, or some other pair of */
              /* where we want to collect every last character between them,  */
              /* characters including whitespaces.                             */
              s += ch;
            }
          }
        }        
        else
        if ( this.isCharATerminatingChar( ch ) == true )
        {
          if ( endingChar == 0 )
          {
            if ( s.equals( "" ) == false )
            {
              /* We have hit a string terminating character. Don't add that */
              /* character to the resulting string and exit.                */
              result = new NextString();
              result.value = s;
              result.stringType = NextString.STRING;
              go = false;
              
              /* Also put back the terminating character. */
              this.ungetChar( i );
            }
            else
            {
              /* We have hit a string terminating character but have not */
              /* collected any other string yet. Return just the          */
              /* terminating character.                                   */
              s += ch;
              result = new NextString();
              result.value = s;
              result.stringType = NextString.TERMINATOR;
              go = false;
            }
          }
          else
          {
            /* This terminating character is actually between two */
            /* doublequotes so ignore it.                          */
            s += ch;
          }
        }
        else
        {
          /* It is just a normal character. Add it to */
          /* the string we are collecting and see if  */
          /* the string we have collected so far is a */
          /* reserved word.                            */
          s += ch;
              
          String token = this.isStringAToken( s );
          if ( token != null )
          {
            result = new NextString();
            result.value = token;
            result.stringType = NextString.RESERVED_WORD;
            go = false;
          }
        }
        
        if ( go == true )
        {    
          i = this.getChar( is );
        }
      }
    }
    catch ( IOException ioe )
    {
      throw new DTDParserException( ioe );
    }
    
    return result;
  }
  
  
//  protected NextString getNextString( InputStream is ) throws DTDParserException
//  {
//    NextString result = null;
//    boolean readDoubleQuote = false;
//    
//    /* Loop collecting characters until we either have a token or */
//    /* a whitespace separated string.                              */
//    try
//    {
//      String s = "";
//      boolean go = true;
//      int i = this.getChar( is );
//      while ( ( go == true ) && ( i != -1 ) )
//      {
//        /* If the character we have just read a whitespace character then */
//        /* return the string we have collected so far.                    */
//        if ( s.equals( "" ) == false )
//        {
//          /* Is the character we have just read a whitespace character? */
//          if ( ( this.isCharAWhiteSpace( ( char ) i ) == true ) && ( readDoubleQuote == false ) )
//          {
//            result = new NextString();
//            result.value = s;
//            result.stringType = NextString.STRING;
//            go = false;
//          }
//          
//          /* Is the character a string terminating character that is not      */
//          /* necessarily a whitespace (so we may want to keep the character)? */
//          if ( ( go == true ) 
//            && ( this.isCharAStringTerminatingChar( ( char ) i ) == true ) 
//            && ( readDoubleQuote == false ) )
//          {
//            this.ungetChar( i );
//            result = new NextString();
//            result.value = s;
//            result.stringType = NextString.STRING;
//            go = false;
//          }
//        }
//        
//        if ( go == true )
//        {
//          if ( i != '\"' )
//          {
//            /* Is the character we have just read a whitespace character? */
//            if ( ( this.isCharAWhiteSpace( ( char ) i ) == false ) 
//              || ( readDoubleQuote == true ) )
//            {
//              s += ( char ) i;
//              
//              /* Is the string we have collected a token? */
//              String token = this.isStringAToken( s );
//              if ( token != null )
//              {
//                result = new NextString();
//                result.value = token;
//                result.stringType = NextString.RESERVED_WORD;
//                go = false;
//              }
//            }
//          }
//          else
//          {
//            if ( readDoubleQuote == false )
//            {
//              readDoubleQuote = true;
//            }
//            else
//            {
//              result = new NextString();
//              result.value = s;
//              result.stringType = NextString.STRING;
//              go = false;
//            }
//          }
//        }
//        
//        if ( go == true )
//        {    
//          i = this.getChar( is );
//        }
//      }
//    }
//    catch ( IOException ioe )
//    {
//      throw new DTDParserException( ioe );
//    }
//    
//    return result;
//  }


  /**
   * Reads a string from the input stream and expects it to be
   * a string, not a reserved word. If it is not then throw an
   * exception.
   */
  
  public String readString( InputStream is ) throws DTDParserException
  {
    String result = null;
    NextString nextString = this.getNextString( is );
    if ( nextString != null )
    {
      if ( nextString.stringType == NextString.STRING )
      {
        result = nextString.value;
      }
      else
      {
        throw new DTDParserException( "Expected a string but found reserved word \"" + nextString.value + "\"." );
      }
    }
    return result;
  }
  
  
  /**
   * Expects the next string read from the inputstream to be
   * a reserved word in the passed list. If it is not then
   * throw an exception.
   */
  
  public String readReservedWord( InputStream is, String[] acceptableReservedWords ) throws DTDParserException
  {
    String result = null;
    NextString nextString = this.getNextString( is );
    if ( nextString != null )
    {
      if ( nextString.stringType == NextString.RESERVED_WORD )
      {
        int loop = 0;
        while ( loop < acceptableReservedWords.length )
        {
          if ( nextString.value.equals( acceptableReservedWords[loop] ) == true )
          {
            result = acceptableReservedWords[loop];
            loop = acceptableReservedWords.length;
          }
          else
          {
            loop++;
          }
        }
        
        if ( result == null )
        {
          throw new DTDParserException( "Reserved word \"" + nextString.value + "\" is not part of the list of acceptable reserved words." );
        }
      }
      else
      {
        throw new DTDParserException( "Expected reserved word but found string \"" + nextString.value + "\"." );
      }
    }
    return result;
  }


  /**
   * Expects the next string read from the inputstream to be
   * the passed reserved word. If it is not then
   * throw an exception.
   */
  
  public String readReservedWord( InputStream is, String acceptableReservedWord ) throws DTDParserException
  {
    String[] acceptableReservedWords = { acceptableReservedWord };
    String result = this.readReservedWord( is, acceptableReservedWords );
    return result;
  }


  /**
   * This method handles the list of child elements that are 
   * under an element declaration. 
   * Eg in the following element declaration 
   * "<!ELEMENT employee (employeeId, name, position, salary, gender, address)>"
   * this method will deal with the part in the round brackets, returning
   * a ChildDTDElement list.
   * 
   * @param firstChildElementName The first child element name in the list.
   * It was already collected before this method was called.
   */
  
  protected ChildDTDElementList handleChildElementListDeclaration( String firstChildElementName, InputStream is ) throws DTDParserException
  {
    /* Create the ChildDTDElementList we will return in this method */
    /* and add the name of the first child element that was passed  */
    /* as a parameter to this method.                                */
    ChildDTDElementList result = new DefaultChildDTDElementList();
    String childElementName = firstChildElementName;
    boolean go = true;
    while ( go == true )
    {  
      /* Given that we have already processed the first child      */
      /* element name, next there should be one of the following    */
      /* things:                                                   */
      /* -A comma, meaning another child element is following.     */
      /* -A "child element count" character.                         */
      /* -A closing right bracket, meaning no more child elements. */
      int minChildElements = 1;
      int maxChildElements = 1;
      boolean readComma = false;
//      String reservedWord = this.readReservedWord( is, new String[] { this.COMMA, this.ZERO_OR_MORE, this.ZERO_OR_ONE, this.ONE_OR_MORE, this.RIGHT_ROUND_BRACKET } );
      char ch = this.readTerminatingChar( is );
      String reservedWord = "";
      reservedWord += ch;
      if ( reservedWord.equals( this.ZERO_OR_MORE ) )
      {
        minChildElements = 0;
        maxChildElements = ChildDTDElement.NO_MAXIMUM_CHILDREN;
      }
      else
      if ( reservedWord.equals( this.ZERO_OR_ONE ) )
      {
        minChildElements = 0;
      }
      else
      if ( reservedWord.equals( this.ONE_OR_MORE ) )
      {
        maxChildElements = ChildDTDElement.NO_MAXIMUM_CHILDREN;
      }
      else
      if ( reservedWord.equals( this.RIGHT_ROUND_BRACKET ) )
      {
        /* There are no more child elements in the list. */
        go = false;
      }
      else
      if ( reservedWord.equals( this.COMMA ) )
      {
        readComma = true;
      }

      /* Create a new child DTD element given the information we */
      /* have collected and add it to the list.                   */
      ChildDTDElement childDTDElement = new DefaultChildDTDElement( childElementName, minChildElements, maxChildElements );
      result.add( childDTDElement );
      
      /* If go is still true then there must be either a comma or a */
      /* right round bracket next.                                  */
      if ( ( go == true ) && ( readComma == false ) )
      {
        ch = this.readTerminatingChar( is, new char[] { ',', ')' } );
        reservedWord = "" + ch;
//        reservedWord = this.readReservedWord( is, new String[] { this.COMMA, this.RIGHT_ROUND_BRACKET } );
        if ( reservedWord.equals( this.RIGHT_ROUND_BRACKET ) )
        {
          go = false;
        }
      }
      
      /* If go is still true then another child element name is following */
      /* so read it and loop.                                              */
      if ( go == true )
      {
        childElementName = this.readString( is );
      }
    }

    /* There may be an optional "count" character followed by */
    /* a final "greater than" sign.                           */
    int minChildElements = 1;
    int maxChildElements = 1;
    char ch = this.readSpecialChar( is, new char[] { '+', '*', '?', ')', '>' } );
    String reservedWord = "" + ch;
//    String reservedWord = this.readReservedWord( is, new String[] { this.ZERO_OR_MORE, this.ZERO_OR_ONE, this.ONE_OR_MORE, this.END } );
    if ( reservedWord.equals( this.ZERO_OR_MORE ) )
    {
      minChildElements = 0;
      maxChildElements = ChildDTDElement.NO_MAXIMUM_CHILDREN;
    }
    else
    if ( reservedWord.equals( this.ZERO_OR_ONE ) )
    {
      minChildElements = 0;
    }
    else
    if ( reservedWord.equals( this.ONE_OR_MORE ) )
    {
      maxChildElements = ChildDTDElement.NO_MAXIMUM_CHILDREN;
    }
    else
    if ( reservedWord.equals( this.END ) )
    {
      /* Unget the END string as we are expecting it later. */
      ungetString( this.END );
    }

    /* Set the minimum and maximum number of child element */
    /* group instances there can be.                       */
    result.setMinimumCount( minChildElements );
    result.setMaximumCount( maxChildElements );

    return result;
  }



  /**
   * Deals with all the following tokens that make up an
   * element declaration.
   */
  
  protected void processElement( InputStream is, MyDocumentType docType ) throws DTDParserException
  {
    /* The next string should be the element's name. */
    String elementName = this.readString( is );
      
    /* The next string is always a left bracket. */
    this.readTerminatingChar( is, '(' );

    int elementDataType = DTDDataTypes.NO_DATA_TYPE;
    ChildDTDElementList childElementList = null;
    
    /* If the next string is a token it must be an */
    /* element's datatype - CDATA or PCDATA.       */
    /* If it is a string then it must be the name  */
    /* of a child element.                         */
    NextString nextString = this.getNextString( is );
    if ( nextString.stringType == NextString.RESERVED_WORD )
    {
      if ( nextString.value.equals( this.ELEMENT_DATATYPE_CDATA  ) )
      {
        elementDataType = DTDDataTypes.CDATA;
      }
      else
      if ( nextString.value.equals( this.ELEMENT_DATATYPE_PCDATA ) )
      {
        elementDataType = DTDDataTypes.PCDATA;
      }
      else
      {
        throw new DTDParserException( "Expected a datatype but found \"" + nextString.value + "\"." );
      }

      /* There will now either be a right bracket or an "|" OR sign followed */
      /* by the name of another element type.                                 */
      char ch = this.readTerminatingChar( is, new char[] { ')', '|' } );
      String reservedWord = "" + ch;
//      String reservedWord = this.readReservedWord( is, new String[] { this.RIGHT_ROUND_BRACKET, this.OR } );
      if ( reservedWord.equals( this.OR ) )
      {
        /* Process the child element information. */
        String childElementName = this.readString( is );
        childElementList = this.handleChildElementListDeclaration( childElementName, is );
      }
    }  
    else
    {
      /* We have just read a child element name. Zero or more of */
      /* these will follow.                                       */
      /* Create a list of these child elements.                   */
      String childElementName = nextString.value;
      childElementList = this.handleChildElementListDeclaration( childElementName, is );
    }

    /* There should be a final greater than sign. */
    this.readReservedWord( is, this.END );

    /* We are happy. Create a DTD element and add it to the list. */
    DTDElement dtdElement = new DefaultDTDElement( elementName, elementDataType, childElementList );
    this.elements.add( dtdElement );

    if ( docType != null )
    {
      docType.dtdElementParsed( dtdElement );
    }
  }  


  /**
   * Processes an attribute list declaration.
   */
  
  protected void processAttributeList( InputStream is ) throws DTDParserException
  {
    /* The first string should be the name of the DTD element */
    /* these attributes go with.                              */
    String elementName = this.readString( is );
      
    /* Following should be one or more attribute declarations. */
    /* Try to read the first attribute's name.                 */
    String attributeName = this.readString( is );
    
    DTDAttributeList attributes = new DefaultDTDAttributeList( elementName );
      
    boolean go = true;
    while ( go == true )
    {
      /* Now there should be a datatype token. */
      int attributeDataType = DTDDataTypes.NO_DATA_TYPE;
      String reservedWord = this.readReservedWord( is, new String[] { this.ATTRIBUTE_DATATYPE_CDATA, this.ATTRIBUTE_DATATYPE_PCDATA } );
      if ( reservedWord.equals( this.ATTRIBUTE_DATATYPE_CDATA ) )
      {
        attributeDataType = DTDDataTypes.CDATA;
      }
      else
      if ( reservedWord.equals( this.ATTRIBUTE_DATATYPE_PCDATA ) )
      {
        attributeDataType = DTDDataTypes.PCDATA;
      }
  
      
      /* Next is a reserved that says whether the attribute */
      /* is required, or has a default value.                */
      boolean required = false;
      boolean implied = false;
      String defaultValue = null;
      NextString nextString = this.getNextString( is );
      if ( nextString.stringType == NextString.RESERVED_WORD )
      {
        if ( nextString.value.equals( this.ATTRIBUTE_REQUIRED ) )
        {
          required = true;
        }
        else
        if ( nextString.value.equals( this.ATTRIBUTE_IMPLIED ) )
        {
          implied = true;
        }
      }
      else
      {
        /* We have a default value for the attribute, surrounded by doublequotes. Remove them. */
        String s = nextString.value;
        defaultValue = s.substring( 1, s.length() -  1 );
      }
      
      /* Create the new DTDAttribute. */      
      DTDAttribute attribute = new DefaultDTDAttribute( attributeName, attributeDataType, defaultValue, required, implied );
      attributes.add( attribute );
        
      /* There should be either another attribute name following */
      /* or a greater than sign.                                 */
      nextString = this.getNextString( is );
      if ( ( nextString.stringType == NextString.RESERVED_WORD )
        && ( nextString.value.equals( this.END ) ) )
      {
        go = false;
      }
      else
      if ( nextString.stringType == NextString.STRING )
      {
        attributeName = nextString.value;
      }
      else
      {
        throw new DTDParserException( "Expected a greater than sign but found \"" + nextString.value + "\"." );
      }
    }  
      
    this.attributeLists.add( attributes );
  }


  /**
   * Processes an entity declaration.
   */
  
  public void processEntity( InputStream is, MyDocumentType docType ) throws DTDParserException
  {
    /* The first thing we should find is the name of the entity. */
    String entityName = this.readString( is );
    
    /* Next is either the key-word "PUBLIC" or a string in double quotes  */
    /* If it is a string, then tat string contains elements and text that */
    /* make up the data in the entity that we want to parse with a         */
    /* DocumentBuilder.                                                    */
    /* If it is "PUBLIC" then the public ID and the system ID follow, and */
    /* the name of a notation.                                            */
    NextString nextString = this.getNextString( is );
    if ( nextString.stringType == NextString.STRING )
    {
      /* We have read a string. Process the elements in it. */
//      nextString.value = "\"" + nextString.value + "\"";
    
      /* First extract the text between the doublequotes. */
      /* Find the first doublequote.                      */
      try
      {
        /* The first character must be a doublequote. If it is not then */
        /* we don't have a legal entity declaration.                    */
        int i = nextString.value.charAt( 0 );
        if ( i == '\"' )
        {
          ByteArrayOutputStream entityData = new ByteArrayOutputStream();
          
          /* If the first character is a '<' then we have nodes in the */
          /* data that will need parsing. If not then we just have      */
          /* text that doesn't need parsing.                           */
          boolean haveNodes = false;
          i = nextString.value.charAt( 1 );
          if ( i == '<' )
          {
            haveNodes = true;
          }
          
          try
          {
            boolean go = true;
            boolean foundGreaterThanSign = false;
            int index = 1;
            while ( go == true )
            {
              if ( i == '\"' ) 
              {
                if ( ( haveNodes == true ) && ( foundGreaterThanSign == true ) )
                {
                  go = false;
                }
                else
                if ( haveNodes == false )
                {
                  go = false;
                }
              }
              else
              if ( i == '>' )
              {
                foundGreaterThanSign = true;
              }
              else
              if ( i == -1 )
              {
                /* We have hit the end of the input stream without finding the end of the entity data. */
                /* Throw and exception.                                                                 */
                throw new DTDParserException( "Hit end of inputstream before we found the end of the entity data." );
              }
              else
              {
                foundGreaterThanSign = false;
              }
              
              if ( go == true )
              {
                entityData.write( i );
                index++;
                i = nextString.value.charAt( index );
              }
            }
            
            /* We have the data between the doublequotes. Do we need to parse it with */
            /* a DocumentBuilder?                                                      */
            if ( haveNodes == true )
            {
              byte[] entityDataAsBytes = entityData.toByteArray();
              InputStream entityDataInputStream = new ByteArrayInputStream( entityDataAsBytes );
              try
              {
                MyDocumentBuilderFactory dbf = ( MyDocumentBuilderFactory ) DocumentBuilderFactory.newInstance();
                try
                {
                  /* Tell the MyDocumentBuilderFactory that we want the */
                  /* DocumentBuilder to create Documents that write     */
                  /* their data to RAM, not disk, to speed things up    */
                  /* slightly, and also so there are no temporary files */
                  /* floating around that need cleaning up.              */
                  // !!! By default this is the case so we don't need   
                  // this line.
  //                dbf.setStorageFileName( null );
                  
                  /* Parse the entity data string into a tree of elements, */
                  /* attributes and text nodes.                            */
                  DefaultMyDocumentBuilder documentBuilder = ( DefaultMyDocumentBuilder) dbf.newDocumentBuilder();
                  try
                  {
                    InputSource inputSource = new InputSource( entityDataInputStream );
                    
                    MyEntity entity = ( MyEntity) this.document.createEntity( entityName, null, null, null );
    
                    /* Create an entity and make it the parent node of all the */
                    /* nodes we are about to parse. We then store that in a    */
                    /* list of all entities in the DTD.                         */
                    documentBuilder.parse( inputSource, document, entity );
                    
                    Identifier identifier = entity.getIdentifier();
                    
                    if ( docType != null )
                    {
                      docType.entityParsed( entity );
                    }
                  }
                  finally
                  {
                    documentBuilder.close();
                  }
                }
                catch ( ParserConfigurationException pce )
                {
                  throw new DTDParserException( pce );
                }
                catch ( MyDocumentBuilderException dbe )
                {
                  throw new DTDParserException( dbe );
                }
                catch ( SAXException se )
                {
                  throw new DTDParserException( se );
                }
              }
              finally
              {
                entityDataInputStream.close();
              }
            }
            else
            {
              String textData = entityData.toString();
              Text textNode = document.createTextNode( textData );
              MyEntity entity = ( MyEntity) document.createEntity( entityName, null, null, null );
              entity.appendChild( textNode );
              Identifier identifier = entity.getIdentifier();
              
              if ( docType != null )
              {
                docType.entityParsed( entity );
              }
            }
          }
          finally
          {
            entityData.close();
          }
        }
        else
        {
          throw new DTDParserException( "Bad entity declaration. Could not find a starting doublequote." );
        }
      }
      catch ( IOException ioe )
      {
        throw new DTDParserException( ioe );
      }
    }
    else
    {
      /* The entity looks like                                           */
      /* <!ENTITY ent5 PUBLIC "entityURI" "entityFile" NDATA notation1> */
      /* I thought the "entityURI" and "entityFile" where the public    */
      /* and system IDs. But it appears that the "notation1" is the     */
      /* public ID, and I am not sure what the system ID is.            */
      /* We are going to read each string following the "PUBLIC" string */
      /* and create an entity.                                          */

      String publicID = this.readString( is );
      String systemID = this.readString( is );
      this.readString( is );
      String notationName = this.readString( is );
      
      /* Remove the doublequotes surrounding all these strings. */
      publicID = publicID.substring( 1, publicID.length() - 1 );
      systemID = systemID.substring( 1, systemID.length() - 1 );
      notationName = notationName.substring( 0, notationName.length() - 1 );

      MyEntity entity = ( MyEntity) this.document.createEntity( entityName, publicID, systemID, notationName );
      if ( docType != null )
      {
        docType.entityParsed( entity );
      }
    }
  }
  
  
  /**
   * Processes a notation declaration.
   */
  
  protected void processNotation( InputStream is, MyDocumentType docType ) throws DTDParserException
  {
    /* First should be the name of the notation. */
    String notationName = this.readString( is );
    
    /* Next should be a string that says whether this is a public */
    /* or a system notation.                                      */
    NextString nextString = this.getNextString( is );
    if ( nextString != null )
    {
      if ( nextString.stringType == NextString.RESERVED_WORD )
      {
        String publicID = null;
        String systemID = null;
        if ( nextString.value.equals( this.NOTATION_TYPE_PUBLIC ) )
        {
          /* The next string is the public ID of the notation. */
          publicID = this.readString( is );
          
          /* Remove the doublequotes surrounding the string. */
          publicID = publicID.substring( 1, publicID.length() - 1 );
        }
        else
        if ( nextString.value.equals( this.NOTATION_TYPE_SYSTEM ) )
        {
          /* The next string is the system ID of the notation. */
          systemID = this.readString( is );
          
          /* Remove the doublequotes surrounding the string. */
          systemID = systemID.substring( 1, systemID.length() - 1 );
        }
        else
        {
          throw new DTDParserException( "Bad notation declaration. Expected reserved word SYSTEM or PUBLIC but got unexpected reserved word \"" + nextString.value + "\"." );
        }
        
        /* Create a new Notation and add it to the list. */
        MyNotation notation = ( MyNotation ) this.document.createNotation( notationName, publicID, systemID );
        Identifier identifier = notation.getIdentifier();
        
        /* There should be a final ">" sign. */
        nextString = this.getNextString( is );
        if ( ( nextString.stringType != NextString.RESERVED_WORD )
          || ( nextString.value.equals( this.END ) == false ) )
        {
          throw new DTDParserException( "Bad notation declaration. Expected final \">\" sign." );
        }  
        
        if ( docType != null )
        {
          docType.notationParsed( notation );
        }
      }
      else
      {
        throw new DTDParserException( "Bad notation declaration. Expected reserved word SYSTEM or PUBLIC but found \"" + nextString.value + "\"." );
      }  
    }
    else
    {
      throw new DTDParserException( "Bad notation declaration. Expected SYSTEM or PUBLIC but found nothing." );
    }        
  }

  
  /**
   * Parses the input DTD InputStream.
   * Builds a list of DTDElements and DTDAttributes.
   */
  
  public void parseElementsAttributesAndEntities( InputStream is, MyDocumentType docType ) throws DTDParserException
  {
    /* This method is pretty simple. We can be either collecting */
    /* an element an attribute or an attribute list. We hardwire */
    /* the processing of each situation below.                   */
    NextString nextString = this.getNextString( is );
    while ( nextString != null )
    {
      /* Is the returned next string a START_ELEMENT token? */
      if ( ( nextString.stringType == NextString.RESERVED_WORD )
        && ( nextString.value.equals( this.ELEMENT_START ) ) )
      {
        this.processElement( is, docType );
      }
      else
      if ( ( nextString.stringType == NextString.RESERVED_WORD )
        && ( nextString.value.equals( this.ATTRIBUTE_LIST_START ) ) )
      {
        this.processAttributeList( is );
      }
      else
      if ( ( nextString.stringType == NextString.RESERVED_WORD )
        && ( nextString.value.equals( this.ENTITY_START ) ) )
      {
        this.processEntity( is, docType );
      }
      else
      if ( ( nextString.stringType == NextString.RESERVED_WORD )
        && ( nextString.value.equals( this.NOTATION_START ) ) )
      {
        this.processNotation( is, docType );
      }
      
      nextString = this.getNextString( is );
    }
    
    /* Join all the attributelists to their elements. */
    int numAttributeLists = this.attributeLists.size();
    for ( int loop = 0; loop < numAttributeLists; loop++ )
    {
      DTDAttributeList attributeList = ( DTDAttributeList ) this.attributeLists.get( loop );
      String elementName = attributeList.getElementName();
      DefaultDTDElement element = ( DefaultDTDElement ) this.getElement( elementName );
      if ( element != null )
      {
        element.setAttributes( attributeList );
      }
      else
      {
        throw new DTDParserException( "Could not find element \"" + elementName + "\"." );
      }
    }        
  }
  
  
  /*----------------------- DTDParser methods -----------------------------------*/
  
  
  /**
   * Spins through the passed inputstream looking for
   * the "<!DOCTYPE " string. If found we then process all
   * the elements and attributes that follow it.
   */
  
  public boolean parse( InputStream is ) throws DTDParserException, java.io.IOException
  {
    boolean foundDOCTYPE = this.findDTDLine( is );
    if ( foundDOCTYPE == true )
    {
      /* The XML file has a DTD declaration. Create a DocumentType to */
      /* hold the DTD contents.                                        */
      MyDocumentType docType = ( MyDocumentType ) this.domImplementation.createDocumentType( null, null, null );
      this.document.setDocumentType( docType );
      
      /* Next should be a string that is the name of the DTD. */
      this.dtdName = this.readString( is );
      if ( docType != null )
      {
        docType.dtdNameParsed( this.dtdName );
      }
      
      /* Next should be the reserved word "SYSTEM". */
      this.readReservedWord( is, this.DTD_SYSTEM );
      
      /* Next is either a string that is the name of a separate */
      /* filename holding the DTD or a left square bracket.      */
      NextString nextString = this.getNextString( is );
      if ( nextString.stringType == NextString.STRING )
      {
        /* We have the filename of an external DTD file. Process it. */
        String externalDTDFileName = nextString.value;
        
        /* Remove the leading and trailing doublequotes. */
        externalDTDFileName = externalDTDFileName.substring( 1, externalDTDFileName.length() - 1 );

        
        // !!! For testing the dom test suite only !!!
        externalDTDFileName = "/Java/DOMTestSuite/tests/files/staff.dtd";


        try
        {
          InputStream is2 = new FileInputStream( externalDTDFileName );
          try
          {
            this.parseElementsAttributesAndEntities( is2, docType );
          }
          finally
          {
            try
            {
              is2.close();
            }
            catch ( IOException ioe )
            {
              throw new DTDParserException( ioe.getMessage() );
            }
          }
        }
        catch ( FileNotFoundException fnfe )
        {
          throw new DTDParserException( fnfe.getMessage() );
        }
        
        /* Next will be either a '[' or a '>'. */
//        String reservedWord = this.readReservedWord( is, new String[] { this.LEFT_SQUARE_BRACKET, this.END } );
        char ch = this.readSpecialChar( is, new char[] { '[', '>' } );
        String reservedWord = "" + ch;
        if ( reservedWord.equals( this.LEFT_SQUARE_BRACKET ) )
        {
          /* We found a '['. That means there are more DTD */
          /* declarations following. Process them.         */
          this.parseElementsAttributesAndEntities( is, docType );
        }
        else
        {
          this.ungetString( reservedWord );
        }
      }
      else
      {
        /* We found a '['. That means there are DTD */
        /* declarations following. Process them.     */
        this.parseElementsAttributesAndEntities( is, docType );
      }
      
      /* There should be a final '>' sign. */
      this.readReservedWord( is, this.END );
    }
    
    
    return foundDOCTYPE;
  }
  
  
  /**
   * Returns the name of the DTD as declared in 
   * "[!Doctype name ...".
   */
  
  public String getName()
  {
    return this.dtdName;
  }
}
