package au.com.explodingsheep.diskDOM.identifier;


import java.util.Arrays;
import au.com.explodingsheep.diskDOM.helpers.IntToBytes;
import au.com.explodingsheep.diskDOM.simpleFileSystem.*;


/**
 * An IdentifierListStore that stores lists of IntIdentifiers
 * to a file in a SimpleFileSystem.
 * 
 * We store the lists in a very similar way to how the SimpleFileSystem
 * itself works. One block has 128 pointers to second level blocks, 
 * which in turn have 128 pointers to the real data blocks. These have
 * the actual integer values that go in the identifiers.
 * On top of this we have an array of bytes that is broken up into
 * units of 12 bytes. These 12 bytes are in turn broken up into 3 * 4 bytes
 * with the following purpose:
 * -int firstBlockNum;
 * -int size;
 * -int capacity;
 * It is the offset into this array that is returned as the identifier for
 * each identifier list.
 */

public class SimpleFileSystemIntIdentifierListStore implements IdentifierListStore
{
  /* Means a pointer goes nowhere. */
  protected final static int INVALID_BLOCK_POINTER = -1;

  /* The number of bytes in one block. */
  protected final static int BYTES_PER_BLOCK = 16;
  
  /* The number of ints or pointers we can fit in one block. */
  protected final static int INTS_PER_BLOCK = BYTES_PER_BLOCK / 4;
  
  /* The number of ints pointed to by a data block. */
  protected final static int INTS_PER_DATABLOCK = INTS_PER_BLOCK;
  
  /* The number of ints pointed to by a second level block */
  protected final static int INTS_PER_SECOND_LEVEL_BLOCK = INTS_PER_DATABLOCK * INTS_PER_BLOCK;
  
  /* The number of ints pointed to by a first level block */
  protected final static int INTS_PER_FIRST_LEVEL_BLOCK = INTS_PER_SECOND_LEVEL_BLOCK * ( INTS_PER_BLOCK -  1 );
  
  /* Holds the map of allocated and free blocks. */
  protected BlockMapManager bmm = null;

  /* Where we store the Identifier lists to. */
  protected SimpleFileSystem sfs = null;

  /* Holds the list of top level pointers to the start of each list. */
  protected byte[] topLevelPointers = new byte[0];

  /* Top level pointer allocation blockmap. */
  protected BlockMapManager topLevelPointerBMM = null;

  /* Handle to the sfs file that holds all the pointers and int data. */
  protected SimpleFileSystemHandle handle = null;
  
  /* An attempt to speed up the code slightly. */
  protected byte[] fourBytes = new byte[4];

  /* The method "calculateBlockChainPositions" calculates all  */
  /* the values needed to set or get a data block.              */
  /* This structure holds all those calculated values.         */
  public class BlockChainPositions
  {
    /* The block number of the first pointer block that points to */
    /* the second level pointer blocks.                            */
    public int firstBlockNum = INVALID_BLOCK_POINTER;
    
    /* How many identifiers are in the list currently. */
    public int size = 0;
    
    /* Number of data blocks allocated to this list currently. */
    public int capacity = 0;

    /* This is the number of first level blocks we need to read */
    /* to get to the first level block we want.                  */
    int numFirstLevelBlocksToRead  = 0;
    
    /* This is the absolute block number of the first-level */
    /* pointer block we want.                                */
    int firstLevelPointerBlockNum = INVALID_BLOCK_POINTER;
    
    /* This is the actual first-level pointer block. */
    byte[] firstLevelPointerBlock = new byte[BYTES_PER_BLOCK];
    
    /* This is the relative pointer number in the first level */
    /* block we want that contains the number of the second   */
    /* level block we want.                                    */
    int offsetIntoFirstLevelPointerBlock = INVALID_BLOCK_POINTER;

    /* This is the absolute block number of the second-level */
    /* pointer block we want.                                  */
    int secondLevelPointerBlockNum = INVALID_BLOCK_POINTER;
    
    /* This is the actual second-level pointer block. */
    byte[] secondLevelPointerBlock = new byte[BYTES_PER_BLOCK];

    /* This is the relative pointer number in the second level */
    /* block that points at the data block.                     */
    int offsetIntoSecondLevelPointerBlock = INVALID_BLOCK_POINTER;

    /* This is the absolute block number of the data block. */
    int dataBlockNum = INVALID_BLOCK_POINTER;
    
    /* This is the offset into the data block for the int we want */
    /* to read or write.                                          */
    int offsetIntoDataBlock = INVALID_BLOCK_POINTER;
    
    /* This is the actual data block, assuming "dataBlockNum" */
    /* above is not INVALID_BLOCK_POINTER.                    */
    byte[] dataBlock = new byte[BYTES_PER_BLOCK];
  }
  
  /* Joins this store to a DOMImplementation. */
  protected Identifier domImplementationIdentifier = null;
  
  
  
  /**
   * Constructor.
   */
  
  public SimpleFileSystemIntIdentifierListStore( SimpleFileSystem sfs, Identifier domImplementationIdentifier ) throws SimpleFileSystemException, IdentifierListException, BlockMapManagerException
  {
    this.sfs = sfs;
    this.domImplementationIdentifier = domImplementationIdentifier;
    
    this.loadTopLevelPointers();

    /* Create the blockmap manager we use to record which blocks */
    /* are used or not.                                           */
    this.bmm = new SimpleFileSystemBlockMapManager( sfs, "SimpleFileSystemIntIdentifierListStore.blockmap" );
    
    /* Open the file that will hold the lists of ints. */
    this.handle = this.sfs.openFile( "SimpleFileSystemIntIdentifierListStore.ints" );
  }
  
  
  /**
   * Returns a block number given the passed block and index into it.
   * Ie the block contains 64 4 byte block numbers.
   */
  
  protected int getIntInBlock( byte[] block, int index ) throws IdentifierListException
  {
    int result = 0;
    if ( index * 4 < block.length )
    {
      this.fourBytes[0] = block[index * 4 + 0];
      this.fourBytes[1] = block[index * 4 + 1];
      this.fourBytes[2] = block[index * 4 + 2];
      this.fourBytes[3] = block[index * 4 + 3];
      result = IntToBytes.bytesToInt( this.fourBytes );
    }
    else
    {
      throw new IdentifierListException( "Index " + index + " is greater than size of passed block " + block.length + "." );
    }
    return result;
  }
  
  
  /**
   * Sets a block number given the passed block and index into it.
   * Ie the block contains 64 4 byte block numbers.
   */
  
  protected void setIntInBlock( byte[] block, int index, int blockNum )
  {
    IntToBytes.intToBytes( blockNum, this.fourBytes );
    block[index * 4 + 0] = this.fourBytes[0];
    block[index * 4 + 1] = this.fourBytes[1];
    block[index * 4 + 2] = this.fourBytes[2];
    block[index * 4 + 3] = this.fourBytes[3];
  }


  /**
   * Loads the top level pointers from the sfs.
   */
  
  protected void loadTopLevelPointers() throws IdentifierListException
  {
    try
    {
      SimpleFileSystemHandle handle = this.sfs.openFile( "SimpleFileSystemIntIdentifierListStore.toplevel" );
      try
      {
        int size = this.sfs.size( handle );
        this.topLevelPointers = new byte[size];
        this.sfs.read( handle, this.topLevelPointers );
      }
      finally
      {
        this.sfs.closeFile( handle );
      }
      
      /* Try to load the blockmap for the top level pointers. */
      this.topLevelPointerBMM = new SimpleFileSystemBlockMapManager( this.sfs, "SimpleFileSystemIntIdentifierListStore.TLBMM" );
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new IdentifierListException( sfse );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new IdentifierListException( bmme );
    }
  }


  /**
   * Saves the top level pointers from the sfs.
   */
  
  protected void saveTopLevelPointers() throws IdentifierListException
  {
    try
    {
      SimpleFileSystemHandle handle = this.sfs.openFile( "SimpleFileSystemIntIdentifierListStore.toplevel" );
      try
      {
        this.sfs.write( handle, this.topLevelPointers );
      }
      finally
      {
        this.sfs.closeFile( handle );
      }
      
      this.topLevelPointerBMM.close();
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new IdentifierListException( bmme );
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new IdentifierListException( sfse );
    }
  }
  
  
  /**
   * Creates a BlockChainPositions instance given the passed pointer
   * into the top level offset.
   */
  
  protected void readTopLevelPointer( BlockChainPositions bcp, int offset )
  {
    offset *= 12;

    this.fourBytes[0] = this.topLevelPointers[offset + 0];
    this.fourBytes[1] = this.topLevelPointers[offset + 1];  
    this.fourBytes[2] = this.topLevelPointers[offset + 2];        
    this.fourBytes[3] = this.topLevelPointers[offset + 3];
    bcp.firstBlockNum = IntToBytes.bytesToInt( this.fourBytes );

    this.fourBytes[0] = this.topLevelPointers[offset + 4];
    this.fourBytes[1] = this.topLevelPointers[offset + 5];  
    this.fourBytes[2] = this.topLevelPointers[offset + 6];        
    this.fourBytes[3] = this.topLevelPointers[offset + 7];
    bcp.size = IntToBytes.bytesToInt( this.fourBytes );
    
    this.fourBytes[0] = this.topLevelPointers[offset + 8];
    this.fourBytes[1] = this.topLevelPointers[offset + 9];  
    this.fourBytes[2] = this.topLevelPointers[offset + 10];        
    this.fourBytes[3] = this.topLevelPointers[offset + 11];
    bcp.capacity = IntToBytes.bytesToInt( this.fourBytes );
  }        


  /**
   * Creates the first blocknu, number of identifiers in the
   * lists, and number of blocks allocated back to the
   * toplevel block pointer byte array.
   */
  
  protected void updateTopLevelPointer( BlockChainPositions bcp, int offset ) throws IdentifierListException
  {
    if ( offset >= 0 )
    {
      offset *= 12;
      if ( offset < this.topLevelPointers.length )
      {
        IntToBytes.intToBytes( bcp.firstBlockNum, this.fourBytes );
        this.topLevelPointers[offset + 0] = this.fourBytes[0];
        this.topLevelPointers[offset + 1] = this.fourBytes[1];
        this.topLevelPointers[offset + 2] = this.fourBytes[2];
        this.topLevelPointers[offset + 3] = this.fourBytes[3];
    
        IntToBytes.intToBytes( bcp.size, this.fourBytes );
        this.topLevelPointers[offset + 4] = this.fourBytes[0];
        this.topLevelPointers[offset + 5] = this.fourBytes[1];
        this.topLevelPointers[offset + 6] = this.fourBytes[2];
        this.topLevelPointers[offset + 7] = this.fourBytes[3];
    
        IntToBytes.intToBytes( bcp.capacity, this.fourBytes );
        this.topLevelPointers[offset + 8] = this.fourBytes[0];
        this.topLevelPointers[offset + 9] = this.fourBytes[1];
        this.topLevelPointers[offset + 10] = this.fourBytes[2];
        this.topLevelPointers[offset + 11] = this.fourBytes[3];
      }
      else
      {
        throw new IdentifierListException( "Offset " + offset + " is larger than top level pointer size " + this.topLevelPointers.length );
      }
    }
    else
    {
      throw new IdentifierListException( "Offset " + offset + " must be zero or greater." );
    }  
  }  


  /**
   * Increase the size of the array that holds the top-level pointers.
   */
  
  protected void allocateMoreSpaceForTopLevelPointers() throws IdentifierListException
  {
    try
    {
      int currentNumBlocks = this.topLevelPointerBMM.getNumBlocks();
      int newNumBlocks = 8;
      if ( currentNumBlocks > 0 )
      {
        newNumBlocks = currentNumBlocks * 2;
      }
      
      int newNumBytes = newNumBlocks * 12;
      byte[] tmpTopLevelPointers = new byte[newNumBytes];
      java.util.Arrays.fill( tmpTopLevelPointers, 0, newNumBytes, ( byte ) -1 );
      for ( int loop = 0; loop < this.topLevelPointers.length; loop++ )
      {
        tmpTopLevelPointers[loop] = this.topLevelPointers[loop];
      }
      this.topLevelPointers = tmpTopLevelPointers;
      
      this.topLevelPointerBMM.setNumBlocks( newNumBlocks );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new IdentifierListException( bmme );
    }
  }
  

  /**
   * Allocate a new top-level pointer in the "topLevelPointers" byte array.
   * Returns the handle to the new top-level pointer.
   */
  
  protected int findFreeTopLevelPointer() throws IdentifierListException
  {
    int result = -1;
    try
    {
      result = this.topLevelPointerBMM.getFreeBlockNum();
      if ( result == -1 )
      {
        this.allocateMoreSpaceForTopLevelPointers();
        result = this.topLevelPointerBMM.getFreeBlockNum();
        if ( result == -1 )
        {
          throw new IdentifierListException( "Could not allocate more space." );
        }
      }
      this.topLevelPointerBMM.lockBlock( result );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new IdentifierListException( bmme );
    }
    return result;
  }
  

  /**
   * Allocates more blocks in the BlockIO so we have space to write 
   * things.
   */
  
  protected void allocateMoreSpace() throws SimpleFileSystemException, BlockMapManagerException
  {
    int newNumBlocks;
    int oldNumBlocks = this.bmm.getNumBlocks();
    if ( oldNumBlocks == 0 )
      newNumBlocks = 8;
    else
      newNumBlocks = oldNumBlocks * 2;
      
    this.bmm.setNumBlocks( newNumBlocks );
    
    /* Write empty blocks to the end of the file. */
    byte[] emptyBlock = new byte[BYTES_PER_BLOCK];
    this.sfs.seek( this.handle, SimpleFileSystem.SEEK_END, 0 );
    for ( int loop = oldNumBlocks; loop < newNumBlocks; loop++ )
    {
      this.sfs.write( this.handle, emptyBlock );
    }
  }

  
  /**
   * Writes a block to an abolute block position in the sfs file.
   */
  
  protected void writeBlock( int blockNum, byte[] blockData ) throws SimpleFileSystemException
  {
    this.sfs.seek( this.handle, blockNum * BYTES_PER_BLOCK );
    this.sfs.write( this.handle, blockData );
  }
  
  
  /**
   * Reads a block from an abolute block position in the sfs file.
   */
  
  protected void readBlock( int blockNum, byte[] blockData ) throws SimpleFileSystemException
  {
    this.sfs.seek( this.handle, blockNum * BYTES_PER_BLOCK );
    this.sfs.read( this.handle, blockData );
  }
  
  
  /**
   * Creates a first level pointer block, and returns
   * its absolute block number.
   * If "previousFirstLevelPointerBlockNum" is not -1 then point
   * it at the new first level pointer block we create.
   */
  
  protected int createFirstLevelBlockPointers( int previousFirstLevelPointerBlockNum ) throws SimpleFileSystemException, BlockMapManagerException
  {
    int newFirstLevelPointerBlockNum = this.bmm.getFreeBlockNum();
    if ( newFirstLevelPointerBlockNum == -1 )
    {
      this.allocateMoreSpace();
      newFirstLevelPointerBlockNum = this.bmm.getFreeBlockNum();
    }
    this.bmm.lockBlock( newFirstLevelPointerBlockNum );
  
    /* Create an empty first block and fill each pointer with -1. */
    byte[] emptyFirstLevelPointerBlock = new byte[BYTES_PER_BLOCK];
    Arrays.fill( emptyFirstLevelPointerBlock, 0, BYTES_PER_BLOCK, (byte ) -1 );

    /* Save the first and second level pointer blocks. */
    this.writeBlock( newFirstLevelPointerBlockNum, emptyFirstLevelPointerBlock );

    /* Point the previous first level block at the new one. */
    if ( previousFirstLevelPointerBlockNum != INVALID_BLOCK_POINTER )
    {
      byte[] previousFirstLevelPointerBlock = new byte[BYTES_PER_BLOCK];
      this.readBlock( previousFirstLevelPointerBlockNum, previousFirstLevelPointerBlock );
      this.setIntInBlock( previousFirstLevelPointerBlock, 0, newFirstLevelPointerBlockNum );
      this.writeBlock( previousFirstLevelPointerBlockNum, previousFirstLevelPointerBlock );
    }

    return  newFirstLevelPointerBlockNum;
  }

  
  /**
   * Finds a free block pointer space in the passed first level 
   * block and returns its index. If the block is full of pointers
   * then return -1.
   */
  
  protected int findFreeBlockPointerSpace( byte[] block, int startSpace ) throws IdentifierListException
  {
    int result = -1;
    int loop = startSpace;
    while ( loop < INTS_PER_BLOCK )
    {
      int blockNum = this.getIntInBlock( block, loop );
      if ( blockNum == INVALID_BLOCK_POINTER )
      {
        result = loop;
        loop = INTS_PER_BLOCK;
      }
      else
      {
        loop++;
      }
    }
    return result;
  }
    
  
  /**
   * Creates a second level pointer block and connects it to the
   * passed first level pointer block.
   */
  
  protected int createSecondLevelBlockPointer( int firstLevelPointerBlockNum ) throws IdentifierListException
  {
    int newSecondLevelPointerBlockNum = INVALID_BLOCK_POINTER;
    try
    {
      newSecondLevelPointerBlockNum = this.bmm.getFreeBlockNum();
      if ( newSecondLevelPointerBlockNum == -1 )
      {
        this.allocateMoreSpace();
        newSecondLevelPointerBlockNum = this.bmm.getFreeBlockNum();
      }
      this.bmm.lockBlock( newSecondLevelPointerBlockNum );
        
      /* Create an empty second level block and fill each pointer with -1. */
      byte[] emptySecondLevelPointerBlock = new byte[BYTES_PER_BLOCK];
      Arrays.fill( emptySecondLevelPointerBlock, 0, BYTES_PER_BLOCK, ( byte ) -1 );
  
      /* Load the first level pointer block and search it for a space      */
      /* so we can add the pointer to this new second level pointer block. */
      byte[] firstLevelPointerBlock = new byte[BYTES_PER_BLOCK];
      this.readBlock( firstLevelPointerBlockNum, firstLevelPointerBlock );
      int freeFirstLevelBlockPointerSpace = this.findFreeBlockPointerSpace( firstLevelPointerBlock, 1 );
  
      /* If the first level pointer block is full then create a new first */
      /* level block and try again.                                        */
      if ( freeFirstLevelBlockPointerSpace == -1 )
      {
        firstLevelPointerBlockNum = this.createFirstLevelBlockPointers( firstLevelPointerBlockNum );
        this.readBlock( firstLevelPointerBlockNum, firstLevelPointerBlock );
        freeFirstLevelBlockPointerSpace = this.findFreeBlockPointerSpace( firstLevelPointerBlock, 1 );
      }
        
      /* Point the first level block at the second level block. */
      this.setIntInBlock( firstLevelPointerBlock, freeFirstLevelBlockPointerSpace, newSecondLevelPointerBlockNum );
      
      /* Save the first and second level pointer blocks. */
      this.writeBlock( firstLevelPointerBlockNum, firstLevelPointerBlock );
      this.writeBlock( newSecondLevelPointerBlockNum, emptySecondLevelPointerBlock );
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new IdentifierListException( sfse );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new IdentifierListException( bmme );
    }
    
    return  newSecondLevelPointerBlockNum;
  }


  /**
   * Allocates a new data block and any pointer blocks as necessary.
   */
  
  protected void allocateAnotherDataBlock( BlockChainPositions bcp ) throws IdentifierListException
  {
    try
    {
      /* First allocate a new empty data block. */
      int dataBlockNum = this.bmm.getFreeBlockNum();
      if ( dataBlockNum == -1 )
      {
        this.allocateMoreSpace();
        dataBlockNum = this.bmm.getFreeBlockNum();
      }
      this.bmm.lockBlock( dataBlockNum );
      
      /* If absolutely no other data blocks are allocated, then this is an empty file, */
      /* so deal with it specially.                                                     */
      if ( bcp.firstBlockNum != INVALID_BLOCK_POINTER )
      {
        /* Next goto the last block number we have currently allocated, then */
        /* look at where we are. Do we need to allocate another second level */
        /* pointer block, or maybe a first level and a second level, or is    */
        /* there space to straight allocate a new data block for the file?   */
        this.calculateBlockChainPositions( bcp, bcp.size - 1 );
    
        /* Is there space in the current second level pointer block for */
        /* the new data block?                                          */
        if ( ( bcp.offsetIntoSecondLevelPointerBlock >= INTS_PER_BLOCK - 1 )
          && ( bcp.offsetIntoFirstLevelPointerBlock  >= INTS_PER_BLOCK - 1 ) )
        {
          /* Both the first and second level pointer blocks are full. */
          /* We need to create new instances of each, then recurse    */
          /* and try again.                                            */
          int newFirstLevelPointerBlockNum = this.createFirstLevelBlockPointers( bcp.firstLevelPointerBlockNum );
          bcp.secondLevelPointerBlockNum = this.createSecondLevelBlockPointer( newFirstLevelPointerBlockNum );
  
          /* We now have space in the new second level pointer block for */
          /* another data block pointer. Add it and save the changes.    */
          this.setIntInBlock( bcp.secondLevelPointerBlock, 0, dataBlockNum );
          this.writeBlock( bcp.secondLevelPointerBlockNum, bcp.secondLevelPointerBlock );
    
          /* We have added a block, increment the total block count in the */
          /* DirectoryEntry.                                               */
          bcp.capacity += INTS_PER_BLOCK;
        }
        else
        if ( bcp.offsetIntoSecondLevelPointerBlock >= INTS_PER_BLOCK - 1 )
        {
          /* The current second level pointer block is full but the first level */
          /* pointer block has space. Create a new second level pointer block   */
          /* then recurse.                                                      */
          bcp.secondLevelPointerBlockNum = this.createSecondLevelBlockPointer( bcp.firstLevelPointerBlockNum );
  
          /* We now have space in the new second level pointer block for */
          /* another data block pointer. Add it and save the changes.    */
          this.readBlock( bcp.secondLevelPointerBlockNum, bcp.secondLevelPointerBlock );
          this.setIntInBlock( bcp.secondLevelPointerBlock, 0, dataBlockNum );
          this.writeBlock( bcp.secondLevelPointerBlockNum,  bcp.secondLevelPointerBlock );
    
          /* We have added a block, increment the total block count in the */
          /* DirectoryEntry.                                               */
          bcp.capacity += INTS_PER_BLOCK;
        }
        else
        {
          /* We have space in the current second level pointer block for */
          /* another data block pointer. Add it and save the changes.    */
          this.setIntInBlock( bcp.secondLevelPointerBlock, bcp.offsetIntoSecondLevelPointerBlock + 1, dataBlockNum );
          this.writeBlock( bcp.secondLevelPointerBlockNum, bcp.secondLevelPointerBlock );
    
          /* We have added a block, increment the total block count in the */
          /* DirectoryEntry.                                               */
          bcp.capacity += INTS_PER_BLOCK;
        }
      }
      else
      {
        /* We need to assign first and second level pointer blocks, then */
        /* append a data block off of them.                               */
        bcp.firstBlockNum = this.createFirstLevelBlockPointers( INVALID_BLOCK_POINTER );
        bcp.secondLevelPointerBlockNum = this.createSecondLevelBlockPointer( bcp.firstBlockNum );
  
        /* We now have space in the new second level pointer block for */
        /* another data block pointer. Add it and save the changes.    */
        this.readBlock( bcp.secondLevelPointerBlockNum, bcp.secondLevelPointerBlock );
        this.setIntInBlock( bcp.secondLevelPointerBlock, 0, dataBlockNum );
        this.writeBlock( bcp.secondLevelPointerBlockNum, bcp.secondLevelPointerBlock );
        
        /* We have added a block, increment the total block count in the */
        /* DirectoryEntry.                                               */
        bcp.capacity += INTS_PER_BLOCK;
      }
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new IdentifierListException( sfse );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new IdentifierListException( bmme );
    }
  }


  /**
   * Calculates the first, second and data block numbers and offsets -
   * all the things needed to set or get an int.
   */
  
  protected void calculateBlockChainPositions( BlockChainPositions blockChainPositions, int index ) throws IdentifierListException
  {
    try
    {
      if ( index < blockChainPositions.capacity )
      {    
        /* We need to find the appropriate second level pointer block. */
        /* Remember that a first level pointer block points to         */
        /* "this.INT_PER_BLOCK - 1" second level pointer blocks,       */
        /* and each of them points to "INTS_PER_BLOCK" data blocks.    */
  
        /* Reset all the values in "blockChainPositions". */
        blockChainPositions.dataBlockNum = INVALID_BLOCK_POINTER;
        blockChainPositions.numFirstLevelBlocksToRead = 0;
        blockChainPositions.offsetIntoFirstLevelPointerBlock = 1;
        blockChainPositions.offsetIntoSecondLevelPointerBlock = 0;
        blockChainPositions.firstLevelPointerBlockNum = blockChainPositions.firstBlockNum;
        blockChainPositions.secondLevelPointerBlockNum = INVALID_BLOCK_POINTER;
  
        /* Loop to figure out the first level block number. */
        this.readBlock( blockChainPositions.firstLevelPointerBlockNum, blockChainPositions.firstLevelPointerBlock );
        while ( index >= INTS_PER_FIRST_LEVEL_BLOCK )
        {
          blockChainPositions.numFirstLevelBlocksToRead++;
          index -= INTS_PER_FIRST_LEVEL_BLOCK;
          blockChainPositions.firstLevelPointerBlockNum = this.getIntInBlock( blockChainPositions.firstLevelPointerBlock, 0 );
          this.readBlock( blockChainPositions.firstLevelPointerBlockNum, blockChainPositions.firstLevelPointerBlock );
        }
  
        /* Now figure out the offset into the first level block      */
        /* that contains the block number of the second level block. */
        while ( index >= INTS_PER_SECOND_LEVEL_BLOCK )
        {
          index -= INTS_PER_SECOND_LEVEL_BLOCK;
          blockChainPositions.offsetIntoFirstLevelPointerBlock++;
        }
        blockChainPositions.secondLevelPointerBlockNum = this.getIntInBlock( blockChainPositions.firstLevelPointerBlock, blockChainPositions.offsetIntoFirstLevelPointerBlock );
        this.readBlock( blockChainPositions.secondLevelPointerBlockNum, blockChainPositions.secondLevelPointerBlock );
  
        /* Figure out the offset into the second level block that */
        /* contains the pointer to the data block.                */
        while ( index >= INTS_PER_DATABLOCK )
        {
          index -= INTS_PER_BLOCK;
          blockChainPositions.offsetIntoSecondLevelPointerBlock++;
        }
        blockChainPositions.dataBlockNum = this.getIntInBlock( blockChainPositions.secondLevelPointerBlock, blockChainPositions.offsetIntoSecondLevelPointerBlock );
        
        
        /* Calculate the offset into the datablock for the integer */
        /* we want to read or write.                               */
        blockChainPositions.offsetIntoDataBlock = index;
        
        /* Read the data block. */
        if ( blockChainPositions.dataBlockNum != INVALID_BLOCK_POINTER )
        {
          this.readBlock( blockChainPositions.dataBlockNum, blockChainPositions.dataBlock );
        }
      }
      else
      {
        throw new IdentifierListException( "Relative row number too big." );
      }
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new IdentifierListException( sfse );
    }
  }


  
  /*---------------- IdentifierListStore methods -------------------------*/
  
  
  /**
   * Creates a new IdentifierList and returns it. The list is
   * initially empty.
   */
  
  public IdentifierList createIdentifierList() throws IdentifierListException
  {
    BlockChainPositions blockChainPositions = new BlockChainPositions();

    int topLevelPointerPosition = this.findFreeTopLevelPointer();
    
    /* Prime the top-level pointer with -1 as the block number */
    /* of the 1sr level pointer block.                         */
    blockChainPositions.firstBlockNum = INVALID_BLOCK_POINTER;
    blockChainPositions.size = 0;
    blockChainPositions.capacity = 0;
    this.updateTopLevelPointer( blockChainPositions, topLevelPointerPosition );

    /* Create an identifier that uniquely identifies the */
    /* IdentifierList we are about to create.             */    
    IntIdentifier intIdentifierListIdentifier = new IntIdentifier( topLevelPointerPosition );
    
    /* Create the IdentifierList we are about to return. */
    IntIdentifierList result = new IntIdentifierList( this, this.domImplementationIdentifier, intIdentifierListIdentifier );
    return result;
  }

  
  /**
   * Returns the value of the block whose index is passed.
   * Returns IntListManager if index is out of range.
   */
  
  public Identifier get( IdentifierList il, int index ) throws IdentifierListException
  {
    BlockChainPositions blockChainPositions = new BlockChainPositions();

    /* Get the block number of the first first level pointer block. */
    IntIdentifier ii = ( IntIdentifier ) il.getIdentifier();
    int offset = ii.getValue();
    this.readTopLevelPointer( blockChainPositions, offset );
    
    /* Calculate the position of the data block containing the */
    /* int we want to read.                                     */
    this.calculateBlockChainPositions( blockChainPositions, index );

    int i = this.getIntInBlock( blockChainPositions.dataBlock, blockChainPositions.offsetIntoDataBlock );

    Identifier result = new IntIdentifier( i );
    return result;
  }
    

  /**
   * Sets the the value of the block whose index is passed.
   * Returns IntListManager if index is out of range.
   */
  
  public void set( IdentifierList il, int index, Identifier i ) throws IdentifierListException
  {
    BlockChainPositions blockChainPositions = new BlockChainPositions();

    /* Get the block number of the first first level pointer block. */
    IntIdentifier ii = ( IntIdentifier ) il.getIdentifier();
    int offset = ii.getValue();
    this.readTopLevelPointer( blockChainPositions, offset );
    
    /* Calculate the position of the data block containing the */
    /* int we want to read.                                     */
    this.calculateBlockChainPositions( blockChainPositions, index );
    
    /* Get the int value of the passed Identifier (that we are assuming */
    /* is actually an IntIdentifier).                                    */
    ii = ( IntIdentifier ) i;
    int value = ii.getValue();

    this.setIntInBlock( blockChainPositions.dataBlock, blockChainPositions.offsetIntoDataBlock, value );
    try
    {
      this.writeBlock( blockChainPositions.dataBlockNum, blockChainPositions.dataBlock );
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new IdentifierListException( sfse );
    }
  }
 
  
  /**
   * Deletes the current value in the IntList.
   * The current value becomes the next value in the list.
   */
  
  public void remove( IdentifierList il, int index ) throws IdentifierListException
  {
    if ( index >= 0 )
    {
      BlockChainPositions blockChainPositions = new BlockChainPositions();
  
      /* Get the block number of the first first level pointer block. */
      IntIdentifier ii = ( IntIdentifier ) il.getIdentifier();
      int offset = ii.getValue();
      this.readTopLevelPointer( blockChainPositions, offset );
      
      /* !!! For this first version we will slowly move each successive */
      /* int one spot to the left. At some point in the future we could */
      /* move whole block left 4 bytes, which would be much quicker.    */
      int size = this.size( il );
      if ( ( index >= 0 ) && ( index < size ) )
      {
        for ( int loop = index; loop < size - 1; loop++ )
        {
          Identifier i = this.get( il, loop + 1 );
          this.set( il, loop, i );
        }
        
        blockChainPositions.size--;
        this.updateTopLevelPointer( blockChainPositions, offset );
      }
    }
    else
    {
      throw new IdentifierListException( "Index positions must be non-negative." );
    }
  }
  
  
  /**
   * Deletes the current value in the IntList.
   * The current value becomes the next value in the list.
   */
  
  public void remove( IdentifierList il, Identifier identifier ) throws IdentifierListException
  {
    int index = this.indexOf( il, identifier );
    if ( index >= 0 )
    {
      this.remove( il, index );
    }
  }  

  
  /**
   * Appends a value to the end of the IntList.
   */
  
  public void add( IdentifierList il, Identifier identifier ) throws IdentifierListException
  {
    BlockChainPositions blockChainPositions = new BlockChainPositions();

    /* Get the block number of the first first level pointer block. */
    IntIdentifier ii = ( IntIdentifier ) il.getIdentifier();
    int offset = ii.getValue();
    this.readTopLevelPointer( blockChainPositions, offset );
    
    /* If there is space in the current data blocks to fit another */
    /* int then do so, otherwise first add a new data block.       */
    if ( blockChainPositions.size < blockChainPositions.capacity )
    {
      this.set( il, blockChainPositions.size, identifier );
      blockChainPositions.size++;
      this.updateTopLevelPointer( blockChainPositions, offset );
    }
    else
    {
      /* We must allocate another new data block to hold ints in. */
      this.allocateAnotherDataBlock( blockChainPositions );
      this.updateTopLevelPointer( blockChainPositions, offset );

      this.set( il, blockChainPositions.size, identifier );
      blockChainPositions.size++;
      this.updateTopLevelPointer( blockChainPositions, offset );
    }
  }    
      
  
  /**
   * Inserts a value before the current item in the IntList.
   */
  
  public void insert( IdentifierList il, int index, Identifier identifier ) throws IdentifierListException
  {
    /* First we must move all the current values one to the right. */
    /* To do this we first append a new value, then move            */
    /* everything right.                                           */
    
    /* First add the current last identifer to the end of the list. */
    int size = this.size( il );
    if ( size > 0 )
    {
      if ( index >= 0 ) 
      {
        if( index < size )
        {
          Identifier lastIdentifierInList = this.get( il, size - 1 );
          this.add( il, lastIdentifierInList );
          size++;
          
          /* Now move the other identifiers one position right. */
          for ( int loop = size - 1; loop > index; loop-- )
          {
            Identifier i = this.get( il, loop - 1 );
            this.set( il, loop, i );
          }
  
          /* Now set the Identifier we want to insert into the list. */
          this.set( il, index, identifier );
        }
        else
        {
          this.add( il, identifier );
        }
      }
      else
      {
        throw new IdentifierListException( "Index positions must be non-negative." );
      }
    }
    else
    {
      this.add( il, identifier );
    }
  }


  /**
   * Returns the number of items in the IntList.
   */
  
  public int size( IdentifierList il ) throws IdentifierListException
  {
    BlockChainPositions blockChainPositions = new BlockChainPositions();

    /* Get the block number of the first first level pointer block. */
    IntIdentifier ii = ( IntIdentifier ) il.getIdentifier();
    int offset = ii.getValue();
    this.readTopLevelPointer( blockChainPositions, offset );
    return blockChainPositions.size;
  }  
  
    
  /**
   * Returns the index position of the block with the passed
   * integer, or -1 if it is not in the list.
   */
  
  public int indexOf( IdentifierList il, Identifier identifier ) throws IdentifierListException
  {
    int result = -1;
    BlockChainPositions blockChainPositions = new BlockChainPositions();
    
    /* Get the block number of the first first level pointer block. */
    IntIdentifier ii = ( IntIdentifier ) il.getIdentifier();
    int offset = ii.getValue();
    this.readTopLevelPointer( blockChainPositions, offset );

    int size = this.size( il );
    int loop = 0;
    while ( loop < size )
    {
      Identifier i = this.get( il, loop );
      if ( i.equals( identifier ) )
      {
        result = loop;
        loop = size;
      }
      else
      {
        loop++;
      }
    }
    
    return result;
  }
  
  
  /**
   * Closes this object.
   */
  
  public void close() throws IdentifierListException
  {
    try
    {
      this.sfs.closeFile( this.handle );
      this.saveTopLevelPointers();
      this.bmm.close();
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new IdentifierListException( bmme );
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new IdentifierListException( sfse );
    }
  }
}
