package au.com.explodingsheep.diskDOM.identifier;


import java.io.*;
import au.com.explodingsheep.diskDOM.simpleFileSystem.*;


/**
 * This is a file based linked list of integers. It actually allows
 * multiple linked lists to exist in the one file.
 * 
 * You create a new list by calling "createIdentifier()". The returned
 * object is then used to manipulate the list. 
 * 
 * Each entry in a list is made up of 12 bytes:
 *   byte[4] PreviousBlockNumber.
 *   byte[4] NextBlockNumber.
 *   byte[4] IntData.
 * 
 * Each item is referred to by its block number.
 * 
 * When a new list is created a special block is written. This is the
 * handle block. The block number of this block is what is passed to
 * FileIntList as its handle. Its "NextBlockNumber" points to the first
 * block of the list. It "PreviousBlockNumber" is NO_BLOCK. Its data is
 * the number of the last block in the list (this is an attempt to speed
 * up appending data to a list).
 * We do this so if the first block is deleted and
 * thus possibly changes to another value any other IntList instances
 * will not be out of sync.
 * 
 * 
 * !!! This class is not used. It has been replaced by SimpleFileSystemIntIdentifierListStore.
 */

public class SlowIntIdentifierListStore implements IdentifierListStore
{
  /* Size of each block. */
  protected static final int BLOCK_SIZE = 12;
  
  /* How many blocks to write ahead when creating new space in the file. */
  /* Must be a multiple of 8.                                            */
  protected static final int GROW_SIZE = 128;
  
  /* This value means "no block". It is used in the list to  */
  /* indicate when we are at the beginning or end of a list. */
  public static final int NO_BLOCK = -1;
  
  /* Means the value in a block is empty. */
  public static final int NO_VALUE = -1;

  /* Identifiers a DOMInstance we use when we create IdentifierList instances. */
  protected Identifier domImplementationIdentifier = null;
  
  /* Array of bitmap of used blocks. */
  protected byte[] blockMap = new byte[0];
  
  /* Access to the block storage file. */
  protected SimpleFileSystemHandle handle = null;
  
  /* Empty block. Used to write new empty blocks. */
  protected byte[] emptyData = new byte[GROW_SIZE * BLOCK_SIZE];

  /* Cache used when reading and writing integers to disk. */
  protected byte[] fourBytes = new byte[4];
  
  /* Where we read and wirte the IdentifierLists to. */
  protected SimpleFileSystem sfs = null;  


  /* Inner class used to read and write blocks. */
  protected class Block
  {
    int previousBlockNum = SlowIntIdentifierListStore.NO_BLOCK;
    int currentBlockNum = SlowIntIdentifierListStore.NO_BLOCK;
    int nextBlockNum = SlowIntIdentifierListStore.NO_BLOCK;
    int value = SlowIntIdentifierListStore.NO_VALUE;
  }
  
  
  /**
   * Constructor.
   */
  
  public SlowIntIdentifierListStore( Identifier domImplementationIdentifier, SimpleFileSystem sfs ) throws IdentifierListException
  {
    this.domImplementationIdentifier = domImplementationIdentifier;
    this.sfs = sfs;
    
    Exception e = null;
    try
    {
      this.handle = this.sfs.openFile( "SlowIntIdentifierListStore.blocks" );
    }
    catch ( SimpleFileSystemException sfse )
    {
      e = sfse;
    }
    
    if ( e != null )
    {
      throw new IdentifierListException( e.getMessage() );
    }
  }


  /**
   * Closes this object.
   */
  
  public void close() throws IdentifierListException
  {
    /* Write back the block map. */
    this.saveBlockMap();
    
    /* Close the block file. */
    try
    {
      this.sfs.closeFile( this.handle );
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new IdentifierListException( sfse.getMessage() );
    }
  }
  
  
  /**
   * Loads the blockmap from a special file.
   */
  
  protected void loadBlockMap() throws IdentifierListException
  {
    try
    {
      SimpleFileSystemHandle blockMapHandle = this.sfs.openFile( "SlowIntIdentifierListStore.blockmap" );
      try
      {
        byte[] bfr = new byte[1024];
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        int bytesRead = this.sfs.read( blockMapHandle, bfr );
        while ( bytesRead > 0 )
        {
          baos.write( bfr, 0, bytesRead );
          bytesRead = this.sfs.read( blockMapHandle, bfr );
        }
        this.blockMap = baos.toByteArray();
      }
      finally
      {
        this.sfs.closeFile( blockMapHandle );
      }
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new IdentifierListException( sfse.getMessage() );
    }
  }


  /**
   * Saves the blockmap to a file.
   */
  
  protected void saveBlockMap()  throws IdentifierListException
  {
    try
    {
      SimpleFileSystemHandle blockMapHandle = this.sfs.openFile( "SlowIntIdentifierListStore.blockmap" );
      try
      {
        this.sfs.write( blockMapHandle, this.blockMap );
      }
      finally
      {
        this.sfs.closeFile( blockMapHandle );
      }
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new IdentifierListException( sfse.getMessage() );
    }
  }


  /**
   * Turns the passed int into 4 bytes.
   */
  
  protected void intToBytes( int i, byte[] b )
  {
    b[0] = ( byte ) ( i & 255 );
    b[1] = ( byte ) ( ( i >> 8 ) & 255 );
    b[2] = ( byte ) ( ( i >> 16 ) & 255 );
    b[3] = ( byte ) ( ( i >> 24 ) & 255 );
  }
  

  /**
   * Turns the passed 4 bytes into an integer.
   */
  
  protected int bytesToInt( byte[] bytes )
  {
    int result = bytes[0] & 255;
    result |= ( bytes[1] & 255 ) << 8;
    result |= ( bytes[2] & 255 ) << 16;
    result |= ( bytes[3] & 255 ) << 24;
    return result;
  }
  
  
  /**
   * Reads a block. block.currentBlockNum must be set to the number
   * of the block to be read. If successful the other values of 
   * block are filled in.
   */
  
  protected void readBlock( Block block ) throws IdentifierListException
  {
    if ( block != null )
    {
      if ( block.currentBlockNum >= 0 )
      {
        if ( this.isBlockLocked( block.currentBlockNum ) == true )
        {
          try
          {
            this.sfs.seek( this.handle, block.currentBlockNum * SlowIntIdentifierListStore.BLOCK_SIZE );
            
            this.sfs.read( this.handle, this.fourBytes );
            block.previousBlockNum = this.bytesToInt( this.fourBytes );
            
            this.sfs.read( this.handle, this.fourBytes );
            block.nextBlockNum = this.bytesToInt( this.fourBytes );

            this.sfs.read( this.handle, this.fourBytes );
            block.value = this.bytesToInt( this.fourBytes );
          }
          catch ( SimpleFileSystemException sfse )
          {
            throw new IdentifierListException( sfse.getMessage() );
          }
        }
        else
        {
          throw new IdentifierListException( "Block " + block.currentBlockNum + " is not locked." );
        }
      }
      else
      {
        throw new IdentifierListException( "Block number must be zero or greater - not " + block.currentBlockNum );
      }
    }
    else
    {
      throw new IdentifierListException( "Passed Block is null." );
    }
  }


  /**
   * Writes data to the passed block. 
   */
  
  public void writeBlock( Block block ) throws IdentifierListException
  {
    if ( block != null )
    {
      this.lockBlock( block.currentBlockNum );
      
      try
      {
        this.sfs.seek( this.handle, block.currentBlockNum * SlowIntIdentifierListStore.BLOCK_SIZE );
        
        this.intToBytes( block.previousBlockNum, this.fourBytes );
        this.sfs.write( this.handle, this.fourBytes );
  
        this.intToBytes( block.nextBlockNum, this.fourBytes );
        this.sfs.write( this.handle, this.fourBytes );
  
        this.intToBytes( block.value, this.fourBytes );
        this.sfs.write( this.handle, this.fourBytes );
      }
      catch ( SimpleFileSystemException sfse )
      {
        throw new IdentifierListException( sfse.getMessage() );
      }
    }
    else
    {
      throw new IdentifierListException( "Passed Block is null." );
    }
  }


  /**
   * Write the handle block with the passed fist block.
   */
  
  protected void writeHandleBlock( int handleBlockNum, int nextBlockNum ) throws IdentifierListException
  {
    Block block = new Block();
    block.previousBlockNum = SlowIntIdentifierListStore.NO_BLOCK;
    block.currentBlockNum = handleBlockNum;
    block.nextBlockNum = nextBlockNum;
    block.value = SlowIntIdentifierListStore.NO_VALUE;
    this.writeBlock( block );
  }
  
  
  /**
   * Returns the block number of the first block for the passed
   * IntList.
   */

  protected int getFirstBlockNum( int handle ) throws IdentifierListException
  {
    int result = SlowIntIdentifierListStore.NO_BLOCK;
    if ( handle >= 0 )
    {
      Block block = new Block();
      block.currentBlockNum = handle;

      this.readBlock( block );
      result = block.nextBlockNum;
    }
    else
    {
      throw new IdentifierListException( "Passed handle is invalid." );
    }
    return result;
  }


  /**
   * Given the passed handle, return the block number of the last block
   * in the list.
   */
  
  public int getLastBlockNum( int handle ) throws IdentifierListException
  {
    int result = SlowIntIdentifierListStore.NO_BLOCK;
    if ( handle >= 0 )
    {
      Block block = new Block();
      int nextBlockNum = this.getFirstBlockNum( handle );
      while ( nextBlockNum != SlowIntIdentifierListStore.NO_BLOCK )
      {
        result = nextBlockNum;
        block.currentBlockNum = nextBlockNum;
        this.readBlock( block );
        nextBlockNum = block.nextBlockNum;
      }
    }    
    else
    {
      throw new IdentifierListException( "Passed handle is invalid." );
    }

    return result;
  }


  /**
   * Reads the block whose index is passed. Fills in "block" if
   * the index is valid, otherwise returns false.
   */
  
  public Block readBlock( int handle, int index ) throws IdentifierListException
  {
    Block result = null;
    Block block = new Block();
    if ( handle >= 0 )
    {
      if ( index >= 0 )
      {
        int firstBlockNum = this.getFirstBlockNum( handle );
        if ( firstBlockNum != SlowIntIdentifierListStore.NO_BLOCK )
        {
          int count = 0;
          block.currentBlockNum = firstBlockNum;
          this.readBlock( block );
          while ( ( count < index ) && ( block.nextBlockNum != SlowIntIdentifierListStore.NO_BLOCK ) )
          {
            count++;
            block.currentBlockNum = block.nextBlockNum;
            this.readBlock( block );
          }
          if ( count == index )
          {
            result = block;
          }
        }
      }
      else
      {
        throw new IdentifierListException( "Passed index must be >= 0." );
      }
    }      
    else
    {
      throw new IdentifierListException( "Passed handle is invalid." );
    }
    return result;
  }


  /**
   * Appends a block to the end of the list.
   */
  
  protected void appendBlock( int handle, int value ) throws IdentifierListException
  {
    Block newBlock = new Block();
    newBlock.currentBlockNum = this.getNextFreeBlockNum();
    newBlock.value = value;

    /* First get the last block number. */
    int lastBlockNum = this.getLastBlockNum( handle );
    
    /* If there is at least one block in the list then load it   */
    /* and point the new block we are about to write back to it. */
    if ( lastBlockNum != SlowIntIdentifierListStore.NO_BLOCK )
    {
      Block lastBlock = new Block();
      lastBlock.currentBlockNum = lastBlockNum;
      readBlock( lastBlock );
      lastBlock.nextBlockNum = newBlock.currentBlockNum;
      
      newBlock.previousBlockNum = lastBlockNum;
      
      this.writeBlock( newBlock );
      this.writeBlock( lastBlock );
    }
    else
    {
      /* This is the very first block. Update the handle to point to it. */
      newBlock.previousBlockNum = handle;
      this.writeBlock( newBlock );
      
      this.writeHandleBlock( handle, newBlock.currentBlockNum );
    }
  }


  /**
   * Inserts a block before the current block.
   */
  
  protected void insertBlock( int handle, int index, int value ) throws IdentifierListException
  {
    if ( index >= 0 )
    {
      Block newBlock = new Block();
      newBlock.currentBlockNum = this.getNextFreeBlockNum();
      newBlock.value = value;
  
      /* Try to load the block we want to insert the new block before. */
      Block nextBlock = this.readBlock( handle, index );
      if ( nextBlock != null )
      {
        /* Is there a previous block as well? */
        if ( nextBlock.previousBlockNum != SlowIntIdentifierListStore.NO_BLOCK )
        {
          Block previousBlock = new Block();
          previousBlock.currentBlockNum = nextBlock.previousBlockNum;
          this.readBlock( previousBlock );
          
          /* Write the new block and join the next and previous */
          /* blocks to it.                                      */
          newBlock.nextBlockNum = nextBlock.currentBlockNum;
          newBlock.previousBlockNum = previousBlock.currentBlockNum;
          nextBlock.previousBlockNum = newBlock.currentBlockNum;
          previousBlock.nextBlockNum = newBlock.currentBlockNum;
          
          this.unLockBlock( previousBlock.currentBlockNum );
          this.unLockBlock( nextBlock.currentBlockNum );
          this.writeBlock( newBlock );
          this.writeBlock( previousBlock );
          this.writeBlock( nextBlock );
        }
      }
      else
      {
        /* There is no next block, write this block to the end */
        /* of the list.                                         */
        int lastBlockNum = this.getLastBlockNum( handle );
        if ( lastBlockNum != SlowIntIdentifierListStore.NO_BLOCK )
        {
          Block lastBlock = new Block();
          lastBlock.currentBlockNum = lastBlockNum;
          lastBlock.nextBlockNum = newBlock.currentBlockNum;
          newBlock.previousBlockNum = lastBlockNum;
          
          this.unLockBlock( lastBlock.currentBlockNum );
          this.writeBlock( newBlock );
          this.writeBlock( lastBlock );
        }
        else
        {
          /* The list is empty. This will be the first block */
          /* in the list.                                     */
          newBlock.previousBlockNum = handle;
          this.writeBlock( newBlock );
    
          this.writeHandleBlock( handle, newBlock.currentBlockNum );
        }
      }
    }
    else
    {
      throw new IdentifierListException( "Passed index must be >= 0, not " + index + "." );    
    }
  }


  /**
   * Removes a block from the list.
   */
  
  protected void removeBlock( int handle, int index ) throws IdentifierListException
  {
    if ( index >= 0 )
    {
      if ( handle >= 0 )
      {
        Block blockToDelete = this.readBlock( handle, index );
        if ( blockToDelete != null )
        {
          this.unLockBlock( blockToDelete.currentBlockNum );
          
          /* Connect the previous and next blocks together. */
          if ( ( blockToDelete.previousBlockNum != SlowIntIdentifierListStore.NO_BLOCK )
            && ( blockToDelete.nextBlockNum != SlowIntIdentifierListStore.NO_BLOCK ) )
          {
            Block nextBlock = new Block();
            Block previousBlock = new Block();
            nextBlock.currentBlockNum = blockToDelete.nextBlockNum;
            previousBlock.currentBlockNum = blockToDelete.previousBlockNum;
            
            this.readBlock( nextBlock );
            this.readBlock( previousBlock );
            previousBlock.nextBlockNum = nextBlock.currentBlockNum;
            nextBlock.previousBlockNum = previousBlock.currentBlockNum;
            this.unLockBlock( previousBlock.currentBlockNum );
            this.unLockBlock( nextBlock.currentBlockNum );
            this.writeBlock( previousBlock );
            this.writeBlock( nextBlock );
          }
          else
          if ( blockToDelete.previousBlockNum != SlowIntIdentifierListStore.NO_BLOCK )
          {
            Block previousBlock = new Block();
            previousBlock.currentBlockNum = blockToDelete.previousBlockNum;
            this.readBlock( previousBlock );
            previousBlock.nextBlockNum = SlowIntIdentifierListStore.NO_BLOCK;
            this.unLockBlock( previousBlock.currentBlockNum );
            this.writeBlock( previousBlock );
          }
        }
      }
      else
      {
        throw new IdentifierListException( "Passed handle is invalid." );
      }
    }
    else
    {
      throw new IdentifierListException( "Passed index must be >= 0, not " + index + "." );    
    }
  }


  /**
   * Locks the block so no-one else can use it.  
   */
  
  protected void lockBlock( int blockNum ) throws IdentifierListException
  {
    int numBlocksInFile = this.blockMap.length * 8;
    if ( blockNum < numBlocksInFile )
    {
      int byteNum = blockNum / 8;
      int bitNum = blockNum % 8;
      byte b = this.blockMap[byteNum];
      byte b2 = ( byte ) ( 1 << bitNum );
      b |= b2;
      this.blockMap[byteNum] = b;
    }
    else
    {
      throw new IdentifierListException( "Block " + blockNum + " is out of range." );
    }
  }


  /**
   * Un-locks the block so it can be re-used.
   */
  
  protected void unLockBlock( int blockNum )
  {
    int numBlocksInFile = this.blockMap.length * 8;
    if ( blockNum < numBlocksInFile )
    {
      int byteNum = blockNum / 8;
      int bitNum = blockNum % 8;
      byte b = this.blockMap[byteNum];
      byte b2 = ( byte ) ( 1 << bitNum );
      b -= b2;
      this.blockMap[byteNum] = b;
    }
  }


  /**
   * Returns if a block is currently locked or not.    
   */
  
  protected boolean isBlockLocked( int blockNum )
  {
    boolean result = false;
    int numBlocksInFile = this.blockMap.length * 8;
    if ( blockNum < numBlocksInFile )
    {
      int byteNum = blockNum / 8;
      int bitNum = blockNum % 8;
      byte b = this.blockMap[byteNum];
      byte b2 = ( byte ) ( 1 << bitNum );
      result = ( ( b & b2 ) != 0 );
    }
    return result;
  }


  /**
   * Allocates more space in the file for more blocks of data.
   */
  
  protected void allocateMoreSpace() throws IdentifierListException
  {
    try
    {
      /* Write empty space to the end of the file. */
      int fileSize = this.sfs.size( this.handle );
      this.sfs.seek( this.handle, fileSize );
      this.sfs.write( this.handle, this.emptyData );
      
      /* Re-size the block bitmap. */
      int numBytesToAddToBlockMap = SlowIntIdentifierListStore.GROW_SIZE / 8;
      byte[] newBlockMap = new byte[ this.blockMap.length + numBytesToAddToBlockMap];
      
      for ( int loop = 0; loop < this.blockMap.length; loop++ )
      {
        newBlockMap[loop] = this.blockMap[loop];
      }
      this.blockMap = newBlockMap;
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new IdentifierListException( sfse.getMessage() );
    }
  }
  
  
  /**
   * Returns the number of the next available block.
   * If no block is available then allocate more space in the file
   * for more blocks.
   * 
   * @param numContinuousBlocks The number of continuous blocks
   * we want to find.
   * 
   */
  
  public int getNextFreeBlockNum() throws IdentifierListException
  {
    int result = SlowIntIdentifierListStore.NO_BLOCK;
    int loop = 0;
    while ( loop < this.blockMap.length )
    {
      if ( this.blockMap[loop] != -1 )
      {
        result = loop * 8;
          
        byte b = this.blockMap[loop];
        if ( ( b & 1 ) == 0 )
          result += 0;
        else
        if ( ( b & 2 ) == 0 )
          result += 1;
        else        
        if ( ( b & 4 ) == 0 )
          result += 2;
        else        
        if ( ( b & 8 ) == 0 )
          result += 3;
        else        
        if ( ( b & 16 ) == 0 )
          result += 4;
        else        
        if ( ( b & 32 ) == 0 )
          result += 5;
        else        
        if ( ( b & 64 ) == 0 )
          result += 6;
        else        
        if ( ( b & 128 ) == 0 )
          result += 7;
          
        loop = this.blockMap.length;
      }
      else
      {
        loop++;
      }
    }
    
    /* If no blocks are free then allocate more space and try again. */
    if ( result == SlowIntIdentifierListStore.NO_BLOCK )
    {
      this.allocateMoreSpace();
      result = this.getNextFreeBlockNum();
    }

    return result;
  }
  
  
  /*--------------------- IdentifierListManager methods. --------------------------*/
  
  
  /**
   * Creates a new IdentifierList and returns it. The list is
   * initially empty.
   */
  
  public IdentifierList createIdentifierList() throws IdentifierListException
  {
    /* Write the header block. */
    int handle = this.getNextFreeBlockNum();
    this.writeHandleBlock( handle, SlowIntIdentifierListStore.NO_BLOCK );
    
    /* Create an Identifier that identifies the IdentifierList we */
    /* are about to create.                                        */
    Identifier identifier = new IntIdentifier( handle );
    
    IntIdentifierList result = new IntIdentifierList( this, this.domImplementationIdentifier, identifier );
//    result.init();
    return result;
  }

  
  /**
   * Returns the value of the block whose index is passed.
   * Returns IntListManager if index is out of range.
   */
  
  public Identifier get( IdentifierList il, int index ) throws IdentifierListException
  {
    Identifier result = null;
    
    IntIdentifier identifierListIdentifier = ( IntIdentifier ) il.getIdentifier();
    int handle = identifierListIdentifier.getValue();
    Block block = this.readBlock( handle, index );
    if ( block != null )
    {
      result = new IntIdentifier( block.value );
    }
    return result;
  }
  
  
  /**
   * Deletes the current value in the IntList.
   * The current value becomes the next value in the list.
   */
  
  public void remove( IdentifierList il, int index ) throws IdentifierListException
  {
    IntIdentifier identifier = ( IntIdentifier ) il.getIdentifier();
    int handle = identifier.getValue();
    this.removeBlock( handle, index );
  }

  
  /**
   * Deletes the current value in the IntList.
   * The current value becomes the next value in the list.
   */
  
  public void remove( IdentifierList il, Identifier identifier ) throws IdentifierListException
  {
    int index = this.indexOf( il, identifier );
    if ( index >= 0 )
    {
      this.remove( il, index );
    }
  }
  
  
  /**
   * Appends a value to the end of the IntList.
   */
  
  public void add( IdentifierList il, Identifier identifier ) throws IdentifierListException
  {
    IntIdentifier identifierListIdentifier = ( IntIdentifier ) il.getIdentifier();
    int handle = identifierListIdentifier.getValue();
    int value = ( ( IntIdentifier ) identifier ).getValue();
    this.appendBlock( handle, value );
  }

  
  /**
   * Inserts a value before the current item in the IntList.
   */
  
  public void insert( IdentifierList il, int index, Identifier identifier ) throws IdentifierListException
  {
    IntIdentifier identifierListIdentifier = ( IntIdentifier ) il.getIdentifier();
    int handle = identifierListIdentifier.getValue();
    int value = ( ( IntIdentifier ) identifier ).getValue();
    this.insertBlock( handle, index, value );
  }    


  /**
   * Sets the the value of the block whose index is passed.
   * Returns IntListManager if index is out of range.
   */
  
  public void set( IdentifierList il, int index, Identifier identifier ) throws IdentifierListException
  {
    IntIdentifier identifierListIdentifier = ( IntIdentifier ) il.getIdentifier();
    int handle = identifierListIdentifier.getValue();
    int value = ( ( IntIdentifier ) identifier ).getValue();
    
    /* If there is already a value at the passed index then */
    /* delete it and insert the new value.                  */
    int oldItemIndex = this.indexOf( il, identifier );
    if ( oldItemIndex >= 0 )
    {
      this.remove( il, index );
      this.insert( il, index + 1, identifier );
    }
  }
      


  /**
   * Returns the number of items in the IntList.
   */
  
  public int size( IdentifierList il ) throws IdentifierListException
  {
    IntIdentifier identifierListIdentifier = ( IntIdentifier ) il.getIdentifier();
    int handle = identifierListIdentifier.getValue();
    int result = 0;
    Block block = new Block();
    block.currentBlockNum = this.getFirstBlockNum( handle );
    while ( block.currentBlockNum != SlowIntIdentifierListStore.NO_BLOCK )
    {
      this.readBlock( block );
      block.currentBlockNum = block.nextBlockNum;
      result++;
    }
    return result;
  }
  
    
  /**
   * Returns the index position of the block with the passed
   * integer, or -1 if it is not in the list.
   */
  
  public int indexOf( IdentifierList il, Identifier identifier ) throws IdentifierListException
  {
    IntIdentifier identifierListIdentifier = ( IntIdentifier ) il.getIdentifier();
    int handle = identifierListIdentifier.getValue();
    int value = ( ( IntIdentifier ) identifier ).getValue();
    int result = -1;
    int index = 0;
    Block block = new Block();
    block.currentBlockNum = this.getFirstBlockNum( handle );
    if ( block.currentBlockNum != -1 )
    {
      this.readBlock( block );
      while ( ( block.currentBlockNum != SlowIntIdentifierListStore.NO_BLOCK ) && ( block.value != value ) )
      {
        this.readBlock( block );
        block.currentBlockNum = block.nextBlockNum;
        if ( block.value != value )
          index++;
      }
      if ( block.value == value )
      {
        result = index;
      }
    }
    else
    {
      /* The list is empty, it has no items. This is ok. */
    }
    return result;
  }
}
