package au.com.explodingsheep.diskDOM.nodeStore;


/**
 * This NodeStore assumes that the Nodes being stored are DefaultMyNodes,
 * and all their Identifiers are IntIdentifiers. We are storing only
 * those identifiers and the node's name and type.
 * 
 * We write the serialized node data into one or more blocks of 64 bytes.
 * The first four bytes hold the number of bytes used in this block.
 * The next four bytes point to the next block.
 * The remaining bytes hold as much data as we can fit in the block.
 * When updating a node we don't delete the first node as this is what
 * the Node identifier points at.
 */


import java.io.*;
import org.w3c.dom.*;
import au.com.explodingsheep.diskDOM.*;
import au.com.explodingsheep.diskDOM.identifier.*;
import au.com.explodingsheep.diskDOM.simpleFileSystem.*;


public class ExperimentalNodeStore implements NodeStore
{
  /* The size of one block that we think a serialized node will */
  /* fit into.                                                  */
  protected static final int BLOCK_SIZE = 64;

  /* Means a block does not point to another block. */
  protected static final int NO_NEXT_BLOCK = -1;

  /* Where we store the int identifiers to. */
  protected transient SimpleFileSystem sfs = null;

  /* Handle to the SimpleFile that stores the int identifiers. */
  protected transient SimpleFileSystemHandle handle = null;

  /* Tracks which 64 byte blocks are in use. */
  protected BlockMapManager bmm = null;

  /* Cache used when reading and writing integers to disk. */
  protected byte[] fourBytes = new byte[4];
  
  protected int numNodesWritten = 0;
  protected int numBytesWritten = 0;
  protected int numElements = 0;
  protected int numAttrs = 0;
  protected int numText = 0;
  protected int numDocument = 0;
  protected int numDocType = 0;
  


  /**
   * Constructor.
   */
  
  public ExperimentalNodeStore( SimpleFileSystem sfs ) throws NodeStoreException
  {
    this.sfs = sfs;
    try
    {
      this.handle = this.sfs.openFile( "ExperimentalNodeStore" );
      
      /* Try to open the blockmap manager. This holds what blocks */
      /* have already been used or not.                            */
      this.bmm = new SimpleFileSystemBlockMapManager( this.sfs, "ExperimentalNodeStore.blockmap" );
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new NodeStoreException( sfse );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new NodeStoreException( bmme );
    }
  }


  /**
   * Turns the passed int into 4 bytes.
   */
  
  protected void intToBytes( int i, byte[] b )
  {
    b[0] = ( byte ) ( i & 255 );
    b[1] = ( byte ) ( ( i >> 8 ) & 255 );
    b[2] = ( byte ) ( ( i >> 16 ) & 255 );
    b[3] = ( byte ) ( ( i >> 24 ) & 255 );
  }
  

  /**
   * Turns the passed 4 bytes into an integer.
   */
  
  protected int bytesToInt( byte[] bytes )
  {
    int result = bytes[0] & 255;
    result |= ( bytes[1] & 255 ) << 8;
    result |= ( bytes[2] & 255 ) << 16;
    result |= ( bytes[3] & 255 ) << 24;
    return result;
  }


  /**
   * Turns the passed DefaultMyNode into an array of bytes.
   * Overrides the equivilent method in BlockIONodeStore.
   */
  
  protected byte[] serializeObject( Object o ) throws NodeStoreException
  {
    byte[] result = null;
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    try
    {
      ObjectOutputStream oos = new ObjectOutputStream( baos );
      try
      {
        DefaultMyNode node = ( DefaultMyNode ) o;
        
        short nodeType = node.getNodeType();
        oos.writeShort( nodeType );

        if ( nodeType != Node.DOCUMENT_TYPE_NODE )
        {
          IntIdentifier identifier = ( IntIdentifier ) node.getDOMImplementationIdentifier();
          int value = identifier.getValue();
          oos.writeInt( value );
    
          identifier = ( IntIdentifier ) node.getIdentifier();
          if ( identifier == null )
            identifier = new IntIdentifier( IntIdentifier.INVALID_VALUE );
          value = identifier.getValue();
          oos.writeInt( value );
    
          IntIdentifierIdentifier parent = ( IntIdentifierIdentifier ) node.getParentIdentifierIdentifier();
          if ( parent == null )
            parent = new IntIdentifierIdentifier( IntIdentifierIdentifier.INVALID_VALUE );
          value = parent.getValue();
          oos.writeInt( value );
    
          int intIdentifierHandle = IntIdentifier.INVALID_VALUE;
          IntIdentifier ii = null;
          IntIdentifierList children = ( IntIdentifierList) node.getChildrenIdentifiers();
          if ( children != null )
          {
            ii = ( IntIdentifier ) children.getIdentifier();
            intIdentifierHandle = ii.getValue();
          }
          oos.writeInt( intIdentifierHandle );
  
          if ( nodeType == Node.ELEMENT_NODE )
          {
            DefaultMyElement myElement = ( DefaultMyElement) node;
            IntIdentifierList attributes = ( IntIdentifierList) myElement.getAttributeIdentifiers();
            ii = ( IntIdentifier ) attributes.getIdentifier();
            intIdentifierHandle = ii.getValue();
            oos.writeInt( intIdentifierHandle );
          }
  
          if ( nodeType == Node.ATTRIBUTE_NODE )
          {
            Attr attr = ( Attr ) node;
            boolean specified = attr.getSpecified();
            oos.writeBoolean( specified );
          }
    
          String prefix = node.getPrefix();
          oos.writeObject( prefix );
    
          String namespaceURI = node.getNamespaceURI();
          oos.writeObject( namespaceURI );

          if ( ( nodeType == Node.ATTRIBUTE_NODE              )
            || ( nodeType == Node.ELEMENT_NODE                )
            || ( nodeType == Node.ENTITY_NODE                 )
            || ( nodeType == Node.ENTITY_REFERENCE_NODE       )
            || ( nodeType == Node.NOTATION_NODE               )
            || ( nodeType == Node.PROCESSING_INSTRUCTION_NODE ) )
          {
            String name = node.getNodeName();
            oos.writeObject( name );
          }
          
          if ( ( nodeType == Node.TEXT_NODE                   )
            || ( nodeType == Node.COMMENT_NODE                )
            || ( nodeType == Node.PROCESSING_INSTRUCTION_NODE )
            || ( nodeType == Node.CDATA_SECTION_NODE          ) )
          {
            HasValueIdentifier hvi = ( HasValueIdentifier ) node;
            identifier = ( IntIdentifier ) hvi.getValueIdentifier();
            if ( identifier != null )
              value = identifier.getValue();
            else
              value = IntIdentifier.INVALID_VALUE;
            oos.writeInt( value );
          }
          
          identifier = ( IntIdentifier ) node.getDocumentIdentifier();
          if ( identifier == null )
            identifier = new IntIdentifier( IntIdentifier.INVALID_VALUE );
          value = identifier.getValue();
          oos.writeInt( value );
          
          boolean readOnly = node.isReadOnly();
          oos.writeBoolean( readOnly );
          
          if ( nodeType == Node.NOTATION_NODE )
          {
            Notation notation = ( Notation ) node;
            String publicID = notation.getPublicId();
            oos.writeObject( publicID );
            String systemID = notation.getSystemId();
            oos.writeObject( systemID );
          }
    
          if ( nodeType == Node.DOCUMENT_NODE )
          {
            MyDocument document = ( MyDocument ) node;
            identifier = ( IntIdentifier ) document.getDocumentTypeIdentifier();
            if ( identifier != null )
              value = identifier.getValue();
            else
              value = IntIdentifier.INVALID_VALUE;
            oos.writeInt( value );
            
            String documentName = document.getDocumentName();
            oos.writeObject( documentName );
          }
          
          if ( nodeType == Node.DOCUMENT_TYPE_NODE )
          {
            MyDocumentType docType = ( MyDocumentType ) node;
  
            IdentifierList entityList = docType.getEntityIdentifierList();
            identifier = ( IntIdentifier ) entityList.getIdentifier();
            value = identifier.getValue();
            oos.writeInt( value );
  
            IdentifierList notationList = docType.getNotationIdentifierList();
            identifier = ( IntIdentifier ) notationList.getIdentifier();
            value = identifier.getValue();
            oos.writeInt( value );
          }
          
          if ( nodeType == Node.ENTITY_NODE )
          {
            Entity entity = ( Entity ) node;
            String publicID = entity.getPublicId();
            String systemID = entity.getSystemId();
            String notationName = entity.getNotationName();
            oos.writeObject( publicID );
            oos.writeObject( systemID );
            oos.writeObject( notationName );
          }
        }
        else
        {          
          /* If the node is a DocumentType then handle it specially so all */
          /* its DTDElement and DTDAttrs are serialized as well.           */
          oos.writeObject( node );
        }

         switch ( nodeType )
         {
           case Node.ELEMENT_NODE: this.numElements++; break;
           case Node.ATTRIBUTE_NODE: this.numAttrs++; break;
           case Node.DOCUMENT_NODE: this.numDocument++; break;
           case Node.DOCUMENT_TYPE_NODE: this.numDocType++; break;
           case Node.TEXT_NODE: this.numText++; break;
         }

      }
      finally
      {
        oos.close();
      }

       result = baos.toByteArray();
       
       this.numNodesWritten++;
       this.numBytesWritten += result.length;
         
    }
    catch ( Exception e )
    {
      throw new NodeStoreException( e );
    }    
    return result;
  }


  /**
   * Turns the passed byte array into a DefaultMyNode.
   * Overrides the equivilent method in BlockIONodeStore.
   */
  
  protected Object deSerializeObject( byte[] data ) throws NodeStoreException
  {
    DefaultMyNode result = null;
    if ( data != null )
    {
      ByteArrayInputStream bais = new ByteArrayInputStream( data );
      try
      {
        ObjectInputStream ois = new ObjectInputStream( bais );
        
        short nodeType = ois.readShort();
        
        if ( nodeType != Node.DOCUMENT_TYPE_NODE )
        {
          int value = ois.readInt();
          IntIdentifier instanceID = new IntIdentifier( value );
    
          value = ois.readInt();
          IntIdentifier identifier = new IntIdentifier( value );
    
          IntIdentifierIdentifier parent = null;
          value = ois.readInt();
          Identifier parentIdentifier = null;
          if ( value != IntIdentifierIdentifier.INVALID_VALUE )
            parentIdentifier = new IntIdentifier( value );
          else
            parentIdentifier = null;
           parent = new IntIdentifierIdentifier( value );
          
          int intIdentifierHandle = ois.readInt();
          IntIdentifierList children = null;
          IntIdentifier ii = null;
          if ( intIdentifierHandle != IntIdentifier.INVALID_VALUE )
          {
            ii = new IntIdentifier( intIdentifierHandle );
            children = new IntIdentifierList( instanceID, ii );
          }
  
          IntIdentifierList attributes = null;
          if ( nodeType == Node.ELEMENT_NODE )
          {  
            intIdentifierHandle = ois.readInt();
            ii = new IntIdentifier( intIdentifierHandle );
            attributes = new IntIdentifierList( instanceID, ii );
          }
          
          boolean specified = false;
          if ( nodeType == Node.ATTRIBUTE_NODE )
          {
            specified = ois.readBoolean();
          }
    
          String prefix = ( String ) ois.readObject();
          String namespaceURI = ( String ) ois.readObject();
          
          String name = null;
          if ( ( nodeType == Node.ATTRIBUTE_NODE              )
            || ( nodeType == Node.ELEMENT_NODE                )
            || ( nodeType == Node.ENTITY_NODE                 )
            || ( nodeType == Node.ENTITY_REFERENCE_NODE       )
            || ( nodeType == Node.NOTATION_NODE               )
            || ( nodeType == Node.PROCESSING_INSTRUCTION_NODE ) )
          {
            name = ( String ) ois.readObject();
          }
  
          IntIdentifier valueIdentifier = null;
          if ( ( nodeType == Node.TEXT_NODE                   )
            || ( nodeType == Node.COMMENT_NODE                )
            || ( nodeType == Node.PROCESSING_INSTRUCTION_NODE )
            || ( nodeType == Node.CDATA_SECTION_NODE          ) )
          {
            value = ois.readInt();
            valueIdentifier = new IntIdentifier( value );
          }
    
          value = ois.readInt();
          IntIdentifier documentIdentifier = new IntIdentifier( value );
  
          boolean readOnly = ois.readBoolean();
          
          String publicID = null;
          String systemID = null;
          if ( nodeType == Node.NOTATION_NODE )
          {
            publicID = ( String ) ois.readObject();
            systemID = ( String ) ois.readObject();
          }
  
          Identifier docTypeIdentifier = null;
          String documentName = null;
          if ( nodeType == Node.DOCUMENT_NODE )
          {
            value = ois.readInt();
            docTypeIdentifier = new IntIdentifier( value );
            documentName = ( String ) ois.readObject();
          }
  
          IdentifierList entityList = null;
          IdentifierList notationList = null;
          if ( nodeType == Node.DOCUMENT_TYPE_NODE )
          {
            value = ois.readInt();
            identifier = new IntIdentifier( value );
            entityList = new IntIdentifierList( instanceID, identifier );
  
            value = ois.readInt();
            identifier = new IntIdentifier( value );
            notationList = new IntIdentifierList( instanceID, identifier );
          }
          
          String notationName = null;
          if ( nodeType == Node.ENTITY_NODE )
          {
            publicID = ( String ) ois.readObject();
            systemID = ( String ) ois.readObject();
            notationName = ( String ) ois.readObject();
          }
    
            
          switch ( nodeType )
          {
            case Node.DOCUMENT_NODE:
              result = new DefaultMyDocument( instanceID,
                                              identifier,
                                              children,
                                              attributes,
                                              docTypeIdentifier,
                                              documentName );
              break;
            
            case Node.ELEMENT_NODE:
              result = new DefaultMyElement( instanceID,
                                      identifier,
                                      parent, 
                                      children,
                                      attributes,
                                      prefix,
                                      namespaceURI,
                                      name,
                                      Node.ELEMENT_NODE,
  //                                    valueIdentifier,
                                      documentIdentifier,
                                      readOnly );
              break;
              
            case Node.ATTRIBUTE_NODE:
              result = new DefaultMyAttr( instanceID,
                                    identifier,
                                    parent, 
                                    children,
  //                                  attributes,
                                    prefix,
                                    namespaceURI,
                                    name,
                                    Node.ATTRIBUTE_NODE,
  //                                  valueIdentifier,
                                    documentIdentifier,
                                    specified,
                                    readOnly );
              break;
              
            case Node.TEXT_NODE:
              result = new DefaultMyText( instanceID,
                                    identifier,
                                    parent, 
//                                    children,
  //                                  attributes,
                                    prefix,
                                    namespaceURI,
                                    name,
                                    Node.TEXT_NODE,
                                    valueIdentifier,
                                    documentIdentifier,
                                    readOnly );
              break;
              
            case Node.CDATA_SECTION_NODE:
              result = new DefaultMyCDATASection( instanceID,
                                    identifier,
                                    parent, 
//                                    children,
  //                                  attributes,
                                    prefix,
                                    namespaceURI,
                                    name,
                                    Node.CDATA_SECTION_NODE,
                                    valueIdentifier,
                                    documentIdentifier );
              break;
                          
            case Node.COMMENT_NODE:
              result = new DefaultMyComment( instanceID,
                                    identifier,
                                    parent, 
//                                    children,
  //                                  attributes,
                                    prefix,
                                    namespaceURI,
                                    name,
                                    Node.COMMENT_NODE,
                                    valueIdentifier,
                                    documentIdentifier );
              break;
                          
            case Node.DOCUMENT_FRAGMENT_NODE:
              result = new DefaultMyDocumentFragment( instanceID,
                                    identifier,
                                    parent, 
                                    children,
  //                                  attributes,
                                    prefix,
                                    namespaceURI,
                                    name,
                                    Node.DOCUMENT_FRAGMENT_NODE,
  //                                  valueIdentifier,
                                    documentIdentifier );
              break;
                          
            case Node.ENTITY_REFERENCE_NODE:
              result = new DefaultMyEntityReference( instanceID,
                                    identifier,
                                    parent, 
                                    children,
  //                                  attributes,
                                    prefix,
                                    namespaceURI,
                                    name,
                                    Node.ENTITY_REFERENCE_NODE,
  //                                  valueIdentifier,
                                    documentIdentifier,
                                    readOnly );
              break;
  
            case Node.ENTITY_NODE:
                result = new DefaultMyEntity( instanceID,
                              identifier,       // Identifier
                              children, 
                              null,       // prefix,
                              null,       // namespaceURI,
                              name,
                              documentIdentifier,
                              publicID,
                              systemID,
                              notationName );
            break;
  
                          
            case Node.PROCESSING_INSTRUCTION_NODE:
              result = new DefaultMyProcessingInstruction( instanceID,
                                    identifier,
                                    parent, 
//                                    children,
  //                                  attributes,
                                    prefix,
                                    namespaceURI,
                                    name,
                                    Node.PROCESSING_INSTRUCTION_NODE,
                                    valueIdentifier,
                                    documentIdentifier,
                                    readOnly );
              break;
              
            case Node.NOTATION_NODE:
              result = new DefaultMyNotation( instanceID,
                                        identifier, 
                                        parent, 
//                                        children,
                                         prefix,
                                        namespaceURI,
                                        name,
                                        publicID,
                                        systemID,
                                        documentIdentifier );
            break;
          }
        }
        else
        {
          /* Specially de-serialize the DocumentType node. */
          result = ( DefaultMyNode ) ois.readObject();
        }
          
        
        if ( result != null )
        {
          result.init();
        }

      }
      catch ( Exception e )
      {
        throw new NodeStoreException( e );
      }
    }
    return result;
  }


  /**
   * Allocates more space for us to allocate and store nodes into.
   */
  
  protected void allocateMoreSpace() throws NodeStoreException
  {
    try
    {
      /* Allocate space in the blockmap manager. */
      int newNumBlocks = 8;
      int currentNumBlocks = this.bmm.getNumBlocks();
      if ( currentNumBlocks >= 8 )
      {
        newNumBlocks = currentNumBlocks * 2;
      }
      this.bmm.setNumBlocks( newNumBlocks );
      
      /* Also allocate space in the file. */
      this.sfs.seek( this.handle, SimpleFileSystem.SEEK_END, 0 );
      byte[] emptyBlock = new byte[BLOCK_SIZE];
      while ( currentNumBlocks < newNumBlocks )
      {
        this.sfs.write( this.handle, emptyBlock );
        currentNumBlocks++;
      }
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new NodeStoreException( sfse );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new NodeStoreException( bmme );
    }
  }


  /**
   * Writes a node's serialized data to one or more blocks.
   * This is called by the public methods "appendNode" and "updateNode".
   */
  
  protected void writeNodeData( byte[] data, int firstBlockNum ) throws NodeStoreException
  {
    try
    {
      int blockNum = firstBlockNum;
      byte[] block = new byte[BLOCK_SIZE];
      int numBytesWritten = 0;
      while ( numBytesWritten < data.length )
      {
        /* Figure out how many bytes we can fit in one block. */
        int numBytesToWriteInBlock = BLOCK_SIZE - 8;
        if ( data.length - numBytesWritten < BLOCK_SIZE - 8 )
        {
          numBytesToWriteInBlock = data.length - numBytesWritten;
        }
        
        /* Write the number of bytes in this block. */
        this.intToBytes( numBytesToWriteInBlock, block );
        
        /* Write the pointer to the next block. */
        int nextBlockNum = NO_NEXT_BLOCK;
        if ( data.length - numBytesWritten > BLOCK_SIZE - 8 )
        {
          nextBlockNum = this.bmm.getFreeBlockNum();
          if ( nextBlockNum == -1 )
          {
            this.allocateMoreSpace();
            nextBlockNum = this.bmm.getFreeBlockNum();
          }
          this.bmm.lockBlock( nextBlockNum );
        }
        this.intToBytes( nextBlockNum, this.fourBytes );
        block[4] = this.fourBytes[0];
        block[5] = this.fourBytes[1];
        block[6] = this.fourBytes[2];
        block[7] = this.fourBytes[3];
        
        /* Copy some or all of the block's data into the current block. */
        for ( int loop = 0; loop < numBytesToWriteInBlock; loop++ )
        {
          block[8 + loop] = data[numBytesWritten + loop];
        }
        
        /* Write the current block's data. */
        this.sfs.seek( this.handle, blockNum * BLOCK_SIZE );
         this.sfs.write( this.handle, block );
         numBytesWritten += numBytesToWriteInBlock;
         
         blockNum = nextBlockNum;
      }
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new NodeStoreException( sfse );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new NodeStoreException( bmme );
    }
  }
  
  
  /**
   * Unlocks the block whose number is passed, and all blocks
   * that this block points to, ie all the other blocks in
   * the block chain that store a nodes data. If "unlockFirstBlock"
   * is false the don't unlock the first block pointed to by
   * "blockNum", only unlock the blocks that follow it.
   */
  
  protected void unlockBlocksInChain( int blockNum, boolean unlockFirstBlock ) throws NodeStoreException
  {
    /* Loop loading the blocks to get their next block number and */
    /* unlocking them.                                            */
    try
    {
      byte[] block = new byte[BLOCK_SIZE];
      while ( blockNum != NO_NEXT_BLOCK )
      {
        /* Read a block of data containing some of the serialized node */
        /* from the disk.                                               */
        this.sfs.seek( this.handle, blockNum * BLOCK_SIZE );
        this.sfs.read( this.handle, block );

        if ( unlockFirstBlock == true )      
          this.bmm.unlockBlock( blockNum );
        else
          unlockFirstBlock = true;

        /* Get the pointer to the next node. */
        this.fourBytes[0] = block[4];
        this.fourBytes[1] = block[5];
        this.fourBytes[2] = block[6];
        this.fourBytes[3] = block[7];
        blockNum = this.bytesToInt( this.fourBytes );
      }
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new NodeStoreException( bmme );
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new NodeStoreException( sfse );
    }
  }
  
  
  /*------------------------- NodeStore methods ------------------------------*/
  
  
  /**
   * Writes a MyNode, assigning it an Identifier in the process.
   */
  
  public Identifier appendNode( MyNode node ) throws NodeStoreException
  {
    Identifier result = null;

    try
    {
      /* Allocate the first block. This is what the Node Identifier we   */
      /* return will point to.                                           */
      /* If there are no free blocks in the blockmap then allocate more. */
      int blockNum = this.bmm.getFreeBlockNum();
      if ( blockNum == -1 )
      {
        this.allocateMoreSpace();
        blockNum = this.bmm.getFreeBlockNum();
      }
      this.bmm.lockBlock( blockNum );
      
      result = new IntIdentifier( blockNum );
      node.setIdentifer( result );
      
      byte[] data = this.serializeObject( node );
      
      this.writeNodeData( data, blockNum );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new NodeStoreException( bmme );
    }

    return result;
  }
  

  /**
   * Load a node from where ever and returns it.
   * After loading the MyNode it's "init" method is called.
   */
  
  public MyNode loadNode( Identifier identifier ) throws NodeStoreException
  {
    MyNode result = null;
    if ( identifier != null )
    {
      IntIdentifier intIdentifier = ( IntIdentifier ) identifier;
      int blockNum = intIdentifier.getValue();
      if ( blockNum != IntIdentifier.INVALID_VALUE )
      {
        try
        {
          ByteArrayOutputStream baos = new ByteArrayOutputStream();
          
          byte[] block = new byte[BLOCK_SIZE];
          
          boolean go = true;
          while ( go == true )
          {
            /* Read a block of data containing some of the serialized node */
            /* from the disk.                                               */
            this.sfs.seek( this.handle, blockNum * BLOCK_SIZE );
            this.sfs.read( this.handle, block );
          
            /* Get the number of bytes in the node.  */
            int numBytesInBlock = this.bytesToInt( block );
            
            /* Get the pointer to the next node. */
            this.fourBytes[0] = block[4];
            this.fourBytes[1] = block[5];
            this.fourBytes[2] = block[6];
            this.fourBytes[3] = block[7];
            int nextBlockNum = this.bytesToInt( this.fourBytes );
            
            /* Copy the data from the block into the temporary output stream. */
            for ( int loop = 0; loop < numBytesInBlock; loop++ )
            {
              baos.write( ( int ) block[8 + loop] );
            }
            
            if ( nextBlockNum != NO_NEXT_BLOCK )
              blockNum = nextBlockNum;
            else
              go = false;
          }
          
          byte[] data = baos.toByteArray();
            
          result = ( MyNode ) this.deSerializeObject( data );
          result.init();
        }
        catch ( SimpleFileSystemException sfse )
        {
          throw new NodeStoreException( sfse );
        }
        catch ( MyNodeException mne )
        {
          throw new NodeStoreException( mne );
        }
      }
    }
    return result;
  }
  
  
  /**
   * Writes back the passed MyNode, overwriting the previous
   * MyNode that was stored in the NodeStore somewhere. The Identifier
   * shouldn't change as a result of this update.
   */
  
  public void updateNode( MyNode node ) throws NodeStoreException
  {
    /* Unlock any blocks following the first block. */
    IntIdentifier intIdentifier = ( IntIdentifier ) node.getIdentifier();
    int blockNum = intIdentifier.getValue();
    this.unlockBlocksInChain( blockNum, false );
    
    
    /* Write the node's data to block. */
    byte[] data = this.serializeObject( node );
    this.writeNodeData( data, blockNum );
  }
  
  
  /**
   * Deletes the passed node from the NodeStore.
   */
  
  public void deleteNode( Identifier identifier ) throws NodeStoreException
  {
    IntIdentifier intIdentifier = ( IntIdentifier ) identifier;
    int blockNum = intIdentifier.getValue();
    this.unlockBlocksInChain( blockNum, true );
  }    
  
  
  /**
   * Called when we want to close down a NodeStore.
   */
  
  public void close() throws NodeStoreException
  {
    try
    {
      this.sfs.closeFile( this.handle );
      this.bmm.close();
  
//      if ( this.numNodesWritten > 0 )
//      {    
//        int i = this.numBytesWritten / this.numNodesWritten;
//        System.out.println( "Nodes = " + this.numNodesWritten );
//        System.out.println( "Bytes = " + this.numBytesWritten );
//        System.out.println( "Average node size = " + i );
//        
//        System.out.println( "element=" + this.numElements );
//        System.out.println( "attr=" + this.numAttrs );
//        System.out.println( "text=" + this.numText );
//        System.out.println( "document=" + this.numDocument );
//        System.out.println( "doctype=" + this.numDocType );
//      }
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new NodeStoreException( sfse );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new NodeStoreException( bmme );
    }
  }
}
