package au.com.explodingsheep.diskDOM.simpleFileSystem;


import java.util.*;


/**
 * Sits on top of a BlockIO to cache the blocks. Written blocks
 * are only written back to the underlying BlockIO when this class
 * is closed.
 * 
 * !!! This still has a bug somewhere.
 */

public class CachedBlockIO implements BlockIO
{
  /* The number of blocks we can cache. This needs to be a */
  /* configurable value.                                   */
  protected int numBlockToCache = 0;
  
  /* The real block IO that we read and write blocks to. */
  protected BlockIO blockIO = null;

  /* What we hold the cached blocks in. Note that we are using a SortedMap */
  /* with the key being the block number.                                   */
  protected SortedMap cachedBlocks = new TreeMap();
  
  /* The last cached block we added to the above list. */
  protected CachedBlock lastCachedBlockAdded = null;
  
  /* The last cached block in the list of those pointed to by */
  /* CachedBlock.nextCachedBlock.                              */
  protected CachedBlock lastCachedBlockInList = null;
  
  /* The number of cached blocks currently in this list. */
  /* This is updated by "addCachedBlock" and             */
  /* "writeLeastUsedBlockToDisk".                         */
  int numCachedBlocks = 0;
    
  /* Holds a cached block. */
  public class CachedBlock implements Comparable
  {
    public byte[] block = null;
    public int blockNum = -1;
    public int useCount = 0;
    public boolean mustWriteBlock = false;
    
    /* This points to another CachedBlock. The one it points to */
    /* is the one use "useCount" is greater than this one.       */
    /* These get dynamically swapped around in the               */
    /* "incrmentUeeCount" method. The end of this chain is the  */
    /* CachedBlock instance that is used least.                  */
    public CachedBlock higherCachedBlock = null;
    public CachedBlock lowerCachedBlock = null;
    
    
    /* Used by a SortedMap to make sure "cachedBlocks" remains sorted. */
    public int compareTo( Object o )
    {
      CachedBlock cachedBlock = ( CachedBlock ) o;
      int result = this.blockNum - cachedBlock.blockNum;
      return result;
    }
  }


  /**
   * Constructor.
   */
  
  public CachedBlockIO( BlockIO blockIO, int numBlockToCache )
  {
    this.blockIO = blockIO;
    this.numBlockToCache = numBlockToCache;
  }

  
  /**
   * Returns a cached block given the passed block number,
   * or null if it is not full.
   */
  
  protected CachedBlock getCachedBlock( int blockNum )
  {
    CachedBlock result = ( CachedBlock ) this.cachedBlocks.get( new Integer( blockNum ) );
    return result;
  }


  /**
   * If the cached block list is full then write the
   * lease used one to the real blockIO.
   */
  
  protected void writeLeastUsedBlockToDisk() throws BlockIOException
  {
    if ( this.numCachedBlocks > this.numBlockToCache )
    {
      /* To do this we need only remove the CachedBlock instance */
      /* pointed to by "lastCachedBlockInList.                    */
      CachedBlock cachedBlockToRemove = this.lastCachedBlockInList;
      if ( cachedBlockToRemove != null )
      {
        this.lastCachedBlockInList = cachedBlockToRemove.higherCachedBlock;
        this.lastCachedBlockInList.lowerCachedBlock = null;
        
        this.cachedBlocks.remove( new Integer( cachedBlockToRemove.blockNum ) );
    
        if ( cachedBlockToRemove.mustWriteBlock == true )
        {
          this.blockIO.writeBlock( cachedBlockToRemove.blockNum, cachedBlockToRemove.block );
        }
        this.numCachedBlocks--;
      }
    }
  }
  

  /**
   * Increments the use count of the passed CachedBlock, and 
   * repositions it in the useRankingList. If if as a result of
   * the increment this CachedBlock's count is greater than the
   * one above it then swap the.
   */
  
  protected void incrementuseCount( CachedBlock cachedBlock )
  {
    cachedBlock.useCount++;
    if ( cachedBlock.higherCachedBlock != null )
    {
      CachedBlock higherCachedBlock = cachedBlock.higherCachedBlock;
      if ( cachedBlock.useCount > higherCachedBlock.useCount )
      {
        CachedBlock tmpCachedBlock = cachedBlock.lowerCachedBlock;

        cachedBlock.higherCachedBlock = higherCachedBlock.higherCachedBlock;
        cachedBlock.lowerCachedBlock = higherCachedBlock;
        higherCachedBlock.higherCachedBlock = cachedBlock;
        higherCachedBlock.lowerCachedBlock = tmpCachedBlock;
        
        if ( higherCachedBlock.lowerCachedBlock == null )
        {
          this.lastCachedBlockInList = higherCachedBlock;
        }
      }
    }
  }


  /**
   * Adds the passed CachedBlock to the "cachedBlocks" list, and
   * also updates the "higherCachedBlock" and "lowerCachedBlock"
   * pointer on it so it gets added at the right point into
   * that list.
   */
  
  protected void addCachedBlock( CachedBlock cachedBlock )
  {
    if ( cachedBlock != null )
    {
      this.cachedBlocks.put( new Integer( cachedBlock.blockNum ), cachedBlock );
      this.numCachedBlocks++;
      
      if ( this.lastCachedBlockAdded != null )
      {
        CachedBlock previousCachedBlock = this.lastCachedBlockAdded.lowerCachedBlock;
        CachedBlock nextCachedBlock = this.lastCachedBlockAdded.higherCachedBlock;

        if ( cachedBlock.useCount >= this.lastCachedBlockAdded.useCount )
        {
          /* Insert the new block BEFORE the last added block. */
          cachedBlock.higherCachedBlock = this.lastCachedBlockAdded.higherCachedBlock;
          cachedBlock.lowerCachedBlock = this.lastCachedBlockAdded;
          this.lastCachedBlockAdded.higherCachedBlock = cachedBlock;
          
          if ( nextCachedBlock != null )
            nextCachedBlock.lowerCachedBlock = cachedBlock;
          if ( previousCachedBlock != null )
            previousCachedBlock.higherCachedBlock = this.lastCachedBlockAdded;
          else
            this.lastCachedBlockInList = this.lastCachedBlockAdded;
        }
        else
        {
          /* Insert the new block AFTER the last added block. */
          cachedBlock.higherCachedBlock = this.lastCachedBlockAdded;
          cachedBlock.lowerCachedBlock = this.lastCachedBlockAdded.lowerCachedBlock;
          this.lastCachedBlockAdded.lowerCachedBlock = cachedBlock;

          if ( nextCachedBlock != null )
            nextCachedBlock.lowerCachedBlock = this.lastCachedBlockAdded;
          if ( previousCachedBlock != null )
            previousCachedBlock.higherCachedBlock = cachedBlock;
          else
            this.lastCachedBlockInList = cachedBlock;
        }
      }
      else
      {
        this.lastCachedBlockInList = cachedBlock;
      }
  
      this.lastCachedBlockAdded = cachedBlock;
    }
  }
    

  /*--------------------------- BlockIO methods --------------------------------*/
  
  
  /**
   * Returns the size of one block.
   */
  
  public int getBlockSize()
  {
    int result = this.blockIO.getBlockSize();
    return result;
  }
  

  /**
   * Reads data from the passed block. Look for the block in the
   * cache. If it is not there then load it from the real BlockIO.
   * If the cache is getting full then we must write back the 
   * least used block back to the real BlockIO.
   */
  
  public void readBlock( int blockNum, byte[] data ) throws BlockIOException
  {
    int numBlocks = this.getNumBlocks();
    if ( ( blockNum >= 0 ) && ( blockNum < numBlocks ) )
    {
      CachedBlock cachedBlock = this.getCachedBlock( blockNum );
      if ( cachedBlock != null )
      {
        System.arraycopy( cachedBlock.block, 0, data, 0, data.length );
        cachedBlock.useCount++;
      }
      else
      {
        /* The block is not in the cache. Add it. */
        this.blockIO.readBlock( blockNum, data );
        CachedBlock newCachedBlock = new CachedBlock();
        newCachedBlock.block = new byte[data.length];
        System.arraycopy( data, 0, newCachedBlock.block, 0, data.length );
        newCachedBlock.blockNum = blockNum;
        newCachedBlock.useCount = 1;
        this.addCachedBlock( newCachedBlock );
      }
      
      this.writeLeastUsedBlockToDisk();
    }
    else
    {
      throw new BlockIOException( "Block " + blockNum + " is out of range. Maximum is " + numBlocks + "." );
    }
  }
  
  
  /**
   * Writes data to the passed block. We don't actually write the
   * block, we just mark its cached block as needing to be written.
   */
  
  public void writeBlock( int blockNum, byte[] data ) throws BlockIOException
  {
    int numBlocks = this.getNumBlocks();
    if ( ( blockNum >= 0 ) && ( blockNum < numBlocks ) )
    {
      CachedBlock cachedBlock = this.getCachedBlock( blockNum );
      if ( cachedBlock == null )
      {
         /* The block is not in the cache. Add it. */
        cachedBlock = new CachedBlock();
        cachedBlock.block = new byte[data.length];
        cachedBlock.blockNum = blockNum;
        cachedBlock.useCount = 1;
        cachedBlock.mustWriteBlock = true;
        this.addCachedBlock( cachedBlock );
      }
      else
      {
        cachedBlock.useCount++;
      }
   
      System.arraycopy( data, 0, cachedBlock.block, 0, data.length );
       cachedBlock.mustWriteBlock = true;

   
      this.writeLeastUsedBlockToDisk();
    }
    else
    {
      throw new BlockIOException( "Block " + blockNum + " is out of range. Maximum is " + numBlocks + "." );
    }
  }
  
  
  /**
   * Flushes any cached blocks to disk.
   */
  
  public void flush() throws BlockIOException
  {
    Collection values = this.cachedBlocks.values();
    CachedBlock[] cachedBlockArray = ( CachedBlock[] ) values.toArray( new CachedBlock[0] );
    
    int numBlocksInCache = cachedBlockArray.length;
    for ( int loop = 0; loop < numBlocksInCache; loop++ )
    {
      CachedBlock cachedBlock = cachedBlockArray[loop];
      if ( cachedBlock.mustWriteBlock == true )
      {
        this.blockIO.writeBlock( cachedBlock.blockNum, cachedBlock.block );
      }
    }
  }
  
  
  /**
   * Closes this object. Note that we are not closing the real
   * underlying BlockIO.
   */
  
  public void close() throws BlockIOException
  {
    this.flush();
  }
  
  
  /**
   * Returns the number of blocks currently available.
   */
  
  public int getNumBlocks()
  {
    int result = this.blockIO.getNumBlocks();
    return result;
  }
  
  
  /**
   * Set the number of blocks available.
   */
  
  public void setNumBlocks( int numBlocks ) throws BlockIOException
  {
    this.blockIO.setNumBlocks( numBlocks );
  }
}
