package au.com.explodingsheep.diskDOM.simpleFileSystem;


/**
 * Enables all the functions of BlockMapManager minus a method
 * to load and save the block data. Sub-instances should override
 * the constructor and the "close" methods to load and save the
 * data.
 */

public class DefaultBlockMapManager implements BlockMapManager
{
  /* The list of blockmaps. */
  protected byte[] blockMap = new byte[0];

  /* The offset into the above blockmap that we locked or */
  /* un-locked a block. We use this to speed up block      */
  /* allocation.                                          */
  int offsetIntoBlockMapOfLastTouchedByte = 0;


  /**
   * Constructor.
   */

  public DefaultBlockMapManager() throws BlockMapManagerException
  {
  }


  /**
   * Closes this object, saving the blockmap information.
   * Note that we are writing the blockmap as a file to the SimpleFileSystem.
   * This in turn allocates blocks in this blockmap. To overcome this
   * we first write a block of "this.blockMap.length" zeros, then we
   * seek back to the start of the file and write the blockmap data.
   * We first write a block of zeros
   */

  public void close() throws BlockMapManagerException
  {
  }


  /**
   * Locks the block so no-one else can use it.
   */

  public void lockBlock( int blockNum ) throws BlockMapManagerException
  {
    int numBlocksInFile = this.getNumBlocks();
    if ( blockNum < numBlocksInFile )
    {
      int byteNum = blockNum / 8;
      int bitNum = blockNum % 8;
      int b = this.blockMap[byteNum];
      int b2 = 1 << bitNum;
      b |= b2;
      this.blockMap[byteNum] = ( byte ) b;
      this.offsetIntoBlockMapOfLastTouchedByte = byteNum;
    }
    else
    {
      throw new BlockMapManagerException( "Block " + blockNum + " is out of range." );
    }
  }


  /**
   * Un-locks the block so it can be re-used.
   */

  public void unlockBlock( int blockNum ) throws BlockMapManagerException
  {
    int numBlocksInFile = this.blockMap.length * 8;
    if ( blockNum < numBlocksInFile )
    {
      int byteNum = blockNum / 8;
      int bitNum = blockNum % 8;
      int b = this.blockMap[byteNum];
      int b2 = 1 << bitNum;
      b2 ^= 255;
      b = ( byte ) ( b & b2 );
      this.blockMap[byteNum] = ( byte ) b;
      this.offsetIntoBlockMapOfLastTouchedByte = byteNum;
    }
  }


  /**
   * Returns if a block is currently locked or not.
   */

  public boolean isBlockLocked( int blockNum ) throws BlockMapManagerException
  {
    boolean result = false;
    int numBlocksInFile = this.blockMap.length * 8;
    if ( blockNum < numBlocksInFile )
    {
      int byteNum = blockNum / 8;
      int bitNum = blockNum % 8;
      int b = this.blockMap[byteNum];
      int b2 = 1 << bitNum;
      result = ( ( b & b2 ) != 0 );
    }
    return result;
  }


  /**
   * Finds the next free block in the map or -1 if no block is free.
   */

  public int getFreeBlockNum() throws BlockMapManagerException
  {
    int result = -1;
    int loop = this.offsetIntoBlockMapOfLastTouchedByte;
    while ( loop < this.blockMap.length )
    {
      if ( this.blockMap[loop] != -1 )
      {
        result = loop * 8;
        byte b = this.blockMap[loop];
        if ( ( b & 1 ) == 0 )
          result += 0;
        else
        if ( ( b & 2 ) == 0 )
          result += 1;
        else
        if ( ( b & 4 ) == 0 )
          result += 2;
        else
        if ( ( b & 8 ) == 0 )
          result += 3;
        else
        if ( ( b & 16 ) == 0 )
          result += 4;
        else
        if ( ( b & 32 ) == 0 )
          result += 5;
        else
        if ( ( b & 64 ) == 0 )
          result += 6;
        else
        if ( ( b & 128 ) == 0 )
          result += 7;

        loop = this.blockMap.length;
      }
      else
      {
        loop++;
      }
    }

    return result;
  }
  

  /**
   * Returns the number of blocks in the blockmap.
   */

  public int getNumBlocks() throws BlockMapManagerException
  {
    int result = 0;
    if ( this.blockMap != null )
    {
      result = this.blockMap.length * 8;
    }
    return result;
  }


  /**
   * Sets the number of available blocks in the map.
   * If the number of blocks is increased then the old
   * block information is retained. It is assumed that the
   * number of blocks being allocated is a multiple of eight.
   */

  public void setNumBlocks( int numBlocks ) throws BlockMapManagerException
  {
    if ( numBlocks >= 0 )
    {
      int numBytesToAllocate = numBlocks / 8;

      byte[] newBlockMap = new byte[numBytesToAllocate];
      if ( this.blockMap != null )
      {
        int count = newBlockMap.length;
        if ( count > this.blockMap.length )
        {
           count = this.blockMap.length;
        }
        
        System.arraycopy( this.blockMap, 0, newBlockMap, 0, count );
      }
      this.blockMap = newBlockMap;
    }
    else
    {
      throw new BlockMapManagerException( "Number of blocks must not be negative." );
    }
    
    this.offsetIntoBlockMapOfLastTouchedByte = 0;
  }
}
