package au.com.explodingsheep.diskDOM.simpleFileSystem;


/**
 * An implementation of the SimpleFileSystem interface that stores all
 * its data into one file.
 * 
 * *** How it works ***
 * We open a file and read and write blocks of 512 bytes in it.
 * The first (MAX_FILES * 64 / 512) blocks hold the MAX_FILES 
 * directory entries. 
 * Each directory entry points to the contents of each file. 
 * It is made up of 64 bytes in the following structure:
 *   -"Entry used" char[1]
 *   -"File name" char[56]
 *   -"File size" int
 *   -"First block number" int
 * where ints are stored as char[4].
 * 
 * The block pointed to by "First block number" holds int pointers
 * to other blocks. Those blocks in turn point to the real data
 * blocks.
 * 
 * A bitmap of those blocks used is held in a special file
 * called "blockmap".
 */


import java.util.*;

import au.com.explodingsheep.diskDOM.helpers.IntToBytes;


public class DefaultSimpleFileSystem implements SimpleFileSystem
{
  /* The maximum number of files we can create. */
  public static int MAX_FILES = 128;

  /* The maximum length of a filename. */
  public static int MAX_FILENAME_SIZE = 32;
  
  /* This value is used to to indicate that a block pointer */
  /* is invalid or points nowhere.                          */
  protected static int INVALID_BLOCK_POINTER = -1;
  
  /* Holds the map of allocated and free blocks. */
  protected BlockMapManager bmm = null;

  /* Holds a list of directory entries. */
  protected List directoryEntries = new ArrayList();
  
  /* We cache file pointer blocks here. */
  protected BlockIO filePointerBlockIO = null;
  
  /* We cache other normal blocks here. */
  protected BlockIO fileDataBlockIO = null;
  
  /* The number of bytes per block in the blockIO. */
  protected int bytesPerBlock = 0;
  
  /* The number of block pointers we can fit in a block. */
  protected int pointersPerBlock = 0;

  /* An attempt to speed up the code slightly. */
  protected byte[] fourBytes = new byte[4];
  
  /* Used as a common block used when reading and writing data. */
  protected byte[] readBlock = null;
  protected byte[] writeBlock = null;
  protected byte[] emptyBlock = null;

  /* Used by "calculateBlockChainPositions", holds the results */
  /* of the calculations.                                       */
  public class BlockChainPositions
  {
    /* This is the number of first level blocks we need to read */
    /* to get to the first level block we want.                  */
    int numFirstLevelBlocksToRead  = 0;
    
    /* This is the absolute block number of the first-level */
    /* pointer block we want.                                */
    int firstLevelPointerBlockNum = INVALID_BLOCK_POINTER;
    
    /* This is the actual first-level pointer block. */
    byte[] firstLevelPointerBlock = null;
    
    /* This is the relative pointer number in the first level */
    /* block we want that contains the number of the second   */
    /* level block we want.                                    */
    int offsetIntoFirstLevelPointerBlock = INVALID_BLOCK_POINTER;

    /* This is the absolute block number of the second-level */
    /* pointer block we want.                                  */
    int secondLevelPointerBlockNum = INVALID_BLOCK_POINTER;
    
    /* This is the actual second-level pointer block. */
    byte[] secondLevelPointerBlock = null;

    /* This is the relative pointer number in the second level */
    /* block that points at the data block.                     */
    int offsetIntoSecondLevelPointerBlock = INVALID_BLOCK_POINTER;

    /* This is the absolute block number of the data block. */
    int dataBlockNum = INVALID_BLOCK_POINTER;
    
    /* This is the actual data block, assuming "dataBlockNum" */
    /* above is not INVALID_BLOCK_POINTER.                    */
    byte[] dataBlock = null;
  }
  
  /* Shared instance of above to slightly speed up things. */
  protected BlockChainPositions blockChainPositions = null;
  
    
  /**
   * Constructor.
   * 
   * @param fileName The name of the file we read and write the 
   * files to.
   */
  
  public DefaultSimpleFileSystem( BlockIO blockIO ) throws SimpleFileSystemException, BlockMapManagerException, BlockIOException
  {
//    this.blockIO = blockIO;
    
    /* Notice that we have two block caches running against the one */
    /* blockIO. This could be a bad thing, however we are making    */
    /* sure that we read and write only each block-type to the      */
    /* appropriate block cache.                                      */
//    this.fileDataBlockIO = new CachedBlockIO( blockIO, 1024 );
    this.filePointerBlockIO = new CachedBlockIO( blockIO, 1024 );
    this.fileDataBlockIO = blockIO;
//    this.filePointerBlockIO = blockIO;
    
    this.bytesPerBlock = this.filePointerBlockIO.getBlockSize();
    this.pointersPerBlock = this.bytesPerBlock / 4;
    
    this.blockChainPositions = new BlockChainPositions();
    this.blockChainPositions.firstLevelPointerBlock = new byte[this.bytesPerBlock];
    this.blockChainPositions.secondLevelPointerBlock = new byte[this.bytesPerBlock];
    this.blockChainPositions.dataBlock = new byte[this.bytesPerBlock];
    
    this.readBlock = new byte[this.bytesPerBlock];
    this.writeBlock = new byte[this.bytesPerBlock];

    /* Create an empty first block and fill each pointer with -1. */
    this.emptyBlock = new byte[this.bytesPerBlock];
    for ( int loop = 0; loop < this.pointersPerBlock; loop++ )
    {
      this.setBlockNum( this.emptyBlock, loop, INVALID_BLOCK_POINTER );
    }


    /* Read the directory entries. */
    this.readDirectory();
    
    /* Create the blockmap manager we use to record which blocks */
    /* are used or not.                                           */
    this.bmm = new SimpleFileSystemBlockMapManager( this, "blockmap" );
    
    /* If the BlockIO object has zero blocks in it then it is brand new, so we want to reserve at */
    /* least "numBlocksRequiredByDirectory" blocks.                                                */
    int numBlocksRequiredByDirectory = MAX_FILES / ( this.bytesPerBlock / 64 );
    int numBlocks = this.filePointerBlockIO.getNumBlocks();
    if ( numBlocks == 0 )
    {
      this.filePointerBlockIO.setNumBlocks( numBlocksRequiredByDirectory );
      this.bmm.setNumBlocks( numBlocksRequiredByDirectory );
    }
    
    /* Make sure the first NUM_DIRECTORY_BLOCKS blocks that hold the */
    /* directory structure are locked.                               */
    for ( int loop = 0; loop < numBlocksRequiredByDirectory; loop++ )
    {
      this.bmm.lockBlock( loop );
    }
  }
  
  
  /**
   * Reads the directory entries from the file.
   * Because we have said that we can hold a maximum
   * of MAX_FILES files and each file (otherwise known
   * as a directory entry) will fit into 64 bytes we can
   * fit all the directory entries into MAX_FILES/64 blocks.
   * Read those blocks and their directory entries.
   */
  
  protected void readDirectory() throws BlockIOException
  {
    /* Only attempt to read the directory if the blockIO contains */
    /* at least as many blocks as we want.                        */
    int numBlocksRequired = MAX_FILES / ( this.bytesPerBlock / 64 );
    int numBlocksInBlockIO = this.filePointerBlockIO.getNumBlocks();
    if ( numBlocksInBlockIO >= numBlocksRequired )
    {
      byte[] directoryBlocks = new byte[numBlocksRequired * this.bytesPerBlock];
  
      /* Read the blocks. */
      byte[] oneBlock = new byte[this.bytesPerBlock];
      for ( int loop = 0; loop < numBlocksRequired; loop++ )
      {
        this.filePointerBlockIO.readBlock( loop, oneBlock );
        
        for ( int loop2 = 0; loop2 < this.bytesPerBlock; loop2++ )
        {
          directoryBlocks[loop * this.bytesPerBlock + loop2] = oneBlock[loop2];
        }
      }
  
      /* Now create a directory entry for each one in the directory */
      /* blocks we just read. Remember that each directory entry is */
      /* 64 bytes, made up of:                                      */
      /*   -"directory entry in use" char                            */
      /*   -"file name"              char[32]                        */
      /*   -"file size"              int                            */
      /*   -"first block num"        int                            */
      /*   -"num allocated blocks"   int                            */
      
      for ( int loop = 0; loop < DefaultSimpleFileSystem.MAX_FILES; loop++ )
      {
        char directoryEntryUsed = ( char ) directoryBlocks[loop * 64];
        if ( directoryEntryUsed == 'Y' )
        {
          DefaultSimpleFileSystemDirectoryEntry directoryEntry = new DefaultSimpleFileSystemDirectoryEntry();
          
          /* Get the filename. */
          directoryEntry.fileName = "";
          int loop2 = 1;
          while ( directoryBlocks[loop * 64 + loop2] != 0 )
          {
            directoryEntry.fileName += ( char ) directoryBlocks[loop * 64 + loop2];
            loop2++;
          }
          loop2++;
          
          /* Get the file size. */
          this.fourBytes[0] = directoryBlocks[loop * 64 + loop2 + 0];
          this.fourBytes[1] = directoryBlocks[loop * 64 + loop2 + 1];
          this.fourBytes[2] = directoryBlocks[loop * 64 + loop2 + 2];
          this.fourBytes[3] = directoryBlocks[loop * 64 + loop2 + 3];
          directoryEntry.size = IntToBytes.bytesToInt( this.fourBytes );
  
          /* Get the first pointer block number. */
          this.fourBytes[0] = directoryBlocks[loop * 64 + loop2 + 4];
          this.fourBytes[1] = directoryBlocks[loop * 64 + loop2 + 5];
          this.fourBytes[2] = directoryBlocks[loop * 64 + loop2 + 6];
          this.fourBytes[3] = directoryBlocks[loop * 64 + loop2 + 7];
          directoryEntry.firstBlock = IntToBytes.bytesToInt( this.fourBytes );
  
          /* Get the number of allocated blocks. */
          this.fourBytes[0] = directoryBlocks[loop * 64 + loop2 + 8];
          this.fourBytes[1] = directoryBlocks[loop * 64 + loop2 + 9];
          this.fourBytes[2] = directoryBlocks[loop * 64 + loop2 + 10];
          this.fourBytes[3] = directoryBlocks[loop * 64 + loop2 + 11];
          directoryEntry.numBlocks = IntToBytes.bytesToInt( this.fourBytes );
  
          this.directoryEntries.add( directoryEntry );
        }
      }
    }
  }
  
  
  /**
   * Write the directory entries to disk.
   */
    
  public void writeDirectory() throws BlockMapManagerException, BlockIOException
  {
    int numBlocksRequired = MAX_FILES / ( this.bytesPerBlock / 64 );
    byte[] directoryBlocks = new byte[numBlocksRequired * this.bytesPerBlock];

    /* Now copy the directory entries into the above block. */
    int numDirectoryEntries = this.directoryEntries.size();
    for ( int loop = 0; loop < numDirectoryEntries; loop++ )
    {
      directoryBlocks[loop * 64] = 'Y';
      
      DefaultSimpleFileSystemDirectoryEntry de = ( DefaultSimpleFileSystemDirectoryEntry ) this.directoryEntries.get( loop );

      /* Write the filename. */
      int offset = 1;
      for ( int loop2 = 0; loop2 < de.fileName.length(); loop2++ )
      {
        char ch = de.fileName.charAt( loop2 );
        directoryBlocks[loop * 64 + offset] = ( byte ) ch;
        offset++;
      }
      directoryBlocks[loop * 64 + offset] = 0;
      offset++;
      
      /* Write the filesize. */
      IntToBytes.intToBytes( de.size, this.fourBytes );
      directoryBlocks[loop * 64 + offset + 0] = this.fourBytes[0];
      directoryBlocks[loop * 64 + offset + 1] = this.fourBytes[1];
      directoryBlocks[loop * 64 + offset + 2] = this.fourBytes[2];
      directoryBlocks[loop * 64 + offset + 3] = this.fourBytes[3];

      /* Write the first pointer block number. */
      IntToBytes.intToBytes( de.firstBlock, this.fourBytes );
      directoryBlocks[loop * 64 + offset + 4] = this.fourBytes[0];
      directoryBlocks[loop * 64 + offset + 5] = this.fourBytes[1];
      directoryBlocks[loop * 64 + offset + 6] = this.fourBytes[2];
      directoryBlocks[loop * 64 + offset + 7] = this.fourBytes[3];

      /* Write the number of allocated blocks. */
      IntToBytes.intToBytes( de.numBlocks, this.fourBytes );
      directoryBlocks[loop * 64 + offset + 8] = this.fourBytes[0];
      directoryBlocks[loop * 64 + offset + 9] = this.fourBytes[1];
      directoryBlocks[loop * 64 + offset + 10] = this.fourBytes[2];
      directoryBlocks[loop * 64 + offset + 11] = this.fourBytes[3];
    }
      
    /* Write the directory block data to the file. */
    byte[] oneBlock = new byte[this.bytesPerBlock];
    for ( int loop2 = 0; loop2 < numBlocksRequired; loop2++ )
    {
      for ( int loop3 = 0; loop3 < this.bytesPerBlock; loop3++ )
      {
        oneBlock[loop3] = directoryBlocks[loop2 * this.bytesPerBlock + loop3];
      }
      this.filePointerBlockIO.writeBlock( loop2, oneBlock );
    }
  }


  /**
   * Finds and returns a directory entry given the passed filename,
   * or null if no filename matches.
   */
  
  protected DirectoryEntry getDirectoryEntry( String fileName )
  {
    DirectoryEntry result = null;
    int loop = 0;
    int numDirectoryEntries = this.directoryEntries.size();
    while ( loop < numDirectoryEntries )
    {
      DefaultSimpleFileSystemDirectoryEntry de = ( DefaultSimpleFileSystemDirectoryEntry ) this.directoryEntries.get( loop );
      if ( de.fileName.equals( fileName ) )
      {
        result = de;
        loop = numDirectoryEntries;
      }
      else
      {
        loop++;
      }
    }
    return result;
  }
  
  
  /**
   * Returns a block number given the passed block and index into it.
   * Ie the block contains 64 4 byte block numbers.
   */
  
  protected int getBlockNum( byte[] block, int index )
  {
    this.fourBytes[0] = block[index * 4 + 0];
    this.fourBytes[1] = block[index * 4 + 1];
    this.fourBytes[2] = block[index * 4 + 2];
    this.fourBytes[3] = block[index * 4 + 3];
    int result = IntToBytes.bytesToInt( this.fourBytes );
    return result;
  }
  
  
  /**
   * Sets a block number given the passed block and index into it.
   * Ie the block contains 64 4 byte block numbers.
   */
  
  protected void setBlockNum( byte[] block, int index, int blockNum )
  {
    IntToBytes.intToBytes( blockNum, this.fourBytes );
    block[index * 4 + 0] = this.fourBytes[0];
    block[index * 4 + 1] = this.fourBytes[1];
    block[index * 4 + 2] = this.fourBytes[2];
    block[index * 4 + 3] = this.fourBytes[3];
  }
  
  
  /**
   * Allocates more blocks in the BlockIO so we have space to write 
   * things.
   */
  
  protected void allocateMoreSpace() throws BlockIOException, BlockMapManagerException
  {
    int numBlocks = this.filePointerBlockIO.getNumBlocks();
    if ( numBlocks == 0 )
    {
      numBlocks = 64;
    }
    else
    {
      /* Increase the number of blocks by 25%, making sure the */
      /* final figure is a multiple of 8.                       */
      numBlocks *= 1.25;
      numBlocks /= 8;
      numBlocks = ( numBlocks + 1 ) * 8;
    }
      
    this.filePointerBlockIO.setNumBlocks( numBlocks );
    this.bmm.setNumBlocks( numBlocks );
  }


  /**
   * Creates a first level pointer block, and returns
   * its absolute block number.
   * If "previousFirstLevelPointerBlockNum" is not -1 then point
   * it at the new first level pointer block we create.
   */
  
  protected int createFirstLevelBlockPointers( int previousFirstLevelPointerBlockNum ) throws BlockIOException, BlockMapManagerException
  {
    int newFirstLevelPointerBlockNum = this.bmm.getFreeBlockNum();
    if ( newFirstLevelPointerBlockNum == -1 )
    {
      this.allocateMoreSpace();
      newFirstLevelPointerBlockNum = this.bmm.getFreeBlockNum();
    }
    this.bmm.lockBlock( newFirstLevelPointerBlockNum );
  
    /* Save the first and second level pointer blocks. */
    this.filePointerBlockIO.writeBlock( newFirstLevelPointerBlockNum, this.emptyBlock );
    
    /* Point the previous first level block at the new one. */
    if ( previousFirstLevelPointerBlockNum != INVALID_BLOCK_POINTER )
    {
      byte[] previousFirstLevelPointerBlock = new byte[this.bytesPerBlock];
      this.filePointerBlockIO.readBlock( previousFirstLevelPointerBlockNum, previousFirstLevelPointerBlock );
      this.setBlockNum( previousFirstLevelPointerBlock, 0, newFirstLevelPointerBlockNum );
      this.filePointerBlockIO.writeBlock( previousFirstLevelPointerBlockNum, previousFirstLevelPointerBlock );
    }

    return  newFirstLevelPointerBlockNum;
  }

  
  /**
   * Finds a free block pointer space in the passed first level 
   * block and returns its index. If the block is full of pointers
   * then return -1.
   */
  
  protected int findFreeBlockPointerSpace( byte[] block, int startSpace )
  {
    int result = -1;
    int loop = startSpace;
    while ( loop < this.pointersPerBlock )
    {
      int blockNum = this.getBlockNum( block, loop );
      if ( blockNum == INVALID_BLOCK_POINTER )
      {
        result = loop;
        loop = this.pointersPerBlock;
      }
      else
      {
        loop++;
      }
    }
    return result;
  }
    
  
  /**
   * Creates a second level pointer block and connects it to the
   * passed first level pointer block.
   */
  
  protected int createSecondLevelBlockPointer( int firstLevelPointerBlockNum ) throws BlockIOException, BlockMapManagerException
  {
    int newSecondLevelPointerBlockNum = this.bmm.getFreeBlockNum();
    if ( newSecondLevelPointerBlockNum == -1 )
    {
      this.allocateMoreSpace();
      newSecondLevelPointerBlockNum = this.bmm.getFreeBlockNum();
    }
    this.bmm.lockBlock( newSecondLevelPointerBlockNum );
      
    /* Create an empty second level block and fill each pointer with -1. */
    byte[] emptySecondLevelPointerBlock = new byte[this.bytesPerBlock];
    Arrays.fill( emptySecondLevelPointerBlock, 0, this.bytesPerBlock, ( byte ) -1 );

    /* Load the first level pointer block and search it for a space      */
    /* so we can add the pointer to this new second level pointer block. */
    byte[] firstLevelPointerBlock = new byte[this.bytesPerBlock];
    this.filePointerBlockIO.readBlock( firstLevelPointerBlockNum, firstLevelPointerBlock );
    int freeFirstLevelBlockPointerSpace = this.findFreeBlockPointerSpace( firstLevelPointerBlock, 1 );

    /* If the first level pointer block is full then create a new first */
    /* level block and try again.                                        */
    if ( freeFirstLevelBlockPointerSpace == -1 )
    {
      firstLevelPointerBlockNum = this.createFirstLevelBlockPointers( firstLevelPointerBlockNum );
      this.filePointerBlockIO.readBlock( firstLevelPointerBlockNum, firstLevelPointerBlock );
      freeFirstLevelBlockPointerSpace = this.findFreeBlockPointerSpace( firstLevelPointerBlock, 1 );
    }
      
    /* Point the first level block at the second level block. */
    this.setBlockNum( firstLevelPointerBlock, freeFirstLevelBlockPointerSpace, newSecondLevelPointerBlockNum );
    
    /* Save the first and second level pointer blocks. */
    this.filePointerBlockIO.writeBlock( firstLevelPointerBlockNum, firstLevelPointerBlock );
    this.filePointerBlockIO.writeBlock( newSecondLevelPointerBlockNum, emptySecondLevelPointerBlock );

    return  newSecondLevelPointerBlockNum;
  }


  /**
   * Calculates the Sets a relative block number given the passed handle.
   */
  
  protected void calculateBlockChainPositions( DefaultSimpleFileSystemDirectoryEntry directoryEntry, int relativeBlockNum, BlockChainPositions blockChainPositions, boolean readDataBlock ) throws SimpleFileSystemException, BlockIOException
  {
    if ( relativeBlockNum < directoryEntry.numBlocks )
    {    
      /* We need to find the appropriate second level pointer block. */
      /* Remember that a first level pointer block points to         */
      /* "this.pointersPerBlock - 1" second level pointer blocks,    */
      /* and each of them points to "this.pointersPerBlock" blocks.   */
      /* This means each first level pointer block points to a total */
      /* of "(this.pointersPerBlock - 1) * this.pointersPerBlock"     */
      /* blocks.                                                     */
      
      int numBlocksPointedToByFirstLevelPointerBlock = ( this.pointersPerBlock - 1 ) * this.pointersPerBlock;
      int numBlocksPointedToBySecondLevelPointerBlock =  this.pointersPerBlock;

      /* Reset all the values in "blockChainPositions". */
      blockChainPositions.dataBlockNum = INVALID_BLOCK_POINTER;
      blockChainPositions.firstLevelPointerBlockNum = directoryEntry.firstBlock;
      blockChainPositions.numFirstLevelBlocksToRead = 0;
      blockChainPositions.offsetIntoFirstLevelPointerBlock = 1;
      blockChainPositions.offsetIntoSecondLevelPointerBlock = 0;
      blockChainPositions.secondLevelPointerBlockNum = INVALID_BLOCK_POINTER;

      /* Loop to figure out the first level block number. */
      this.filePointerBlockIO.readBlock( blockChainPositions.firstLevelPointerBlockNum, blockChainPositions.firstLevelPointerBlock );
      while ( relativeBlockNum >= numBlocksPointedToByFirstLevelPointerBlock )
      {
        blockChainPositions.numFirstLevelBlocksToRead++;
        relativeBlockNum -= numBlocksPointedToByFirstLevelPointerBlock;
        blockChainPositions.firstLevelPointerBlockNum = this.getBlockNum( blockChainPositions.firstLevelPointerBlock, 0 );
        this.filePointerBlockIO.readBlock( blockChainPositions.firstLevelPointerBlockNum, blockChainPositions.firstLevelPointerBlock );
      }

      /* Now figure out the offset into the first level block      */
      /* that contains the block number of the second level block. */
      while ( relativeBlockNum >= numBlocksPointedToBySecondLevelPointerBlock )
      {
        relativeBlockNum -= numBlocksPointedToBySecondLevelPointerBlock;
        blockChainPositions.offsetIntoFirstLevelPointerBlock++;
      }
      blockChainPositions.secondLevelPointerBlockNum = this.getBlockNum( blockChainPositions.firstLevelPointerBlock, blockChainPositions.offsetIntoFirstLevelPointerBlock );
      this.filePointerBlockIO.readBlock( blockChainPositions.secondLevelPointerBlockNum, blockChainPositions.secondLevelPointerBlock );

      /* Figure out the offset into the second level block that */
      /* contains the pointer to the data block.                */
      blockChainPositions.offsetIntoSecondLevelPointerBlock = relativeBlockNum;
      blockChainPositions.dataBlockNum = this.getBlockNum( blockChainPositions.secondLevelPointerBlock, blockChainPositions.offsetIntoSecondLevelPointerBlock );
      
      /* Read the data block. */
      if ( ( blockChainPositions.dataBlockNum != INVALID_BLOCK_POINTER )
        && ( readDataBlock == true ) )
      {
        this.filePointerBlockIO.readBlock( blockChainPositions.dataBlockNum, blockChainPositions.dataBlock );
      }
    }
    else
    {
      throw new SimpleFileSystemException( "Relative row number too big." );
    }
  }


  /**
   * Allocates a new data block and any pointer blocks as necessary.
   */
  
  protected void allocateAnotherDataBlock( DefaultSimpleFileSystemHandle handle ) throws SimpleFileSystemException, BlockIOException, BlockMapManagerException
  {
    /* First allocate a new empty data block. */
    int dataBlockNum = this.bmm.getFreeBlockNum();
    if ( dataBlockNum == -1 )
    {
      this.allocateMoreSpace();
      dataBlockNum = this.bmm.getFreeBlockNum();
    }
    this.bmm.lockBlock( dataBlockNum );
    
    /* If absolutely no other data blocks are allocated, then this is an empty file, */
    /* so deal with it specially.                                                     */
    if ( handle.directoryEntry.numBlocks > 0 )
    {
      /* Next goto the last block number we have currently allocated, then */
      /* look at where we are. Do we need to allocate another second level */
      /* pointer block, or maybe a first level and a second level, or is    */
      /* there space to straight allocate a new data block for the file?   */
      this.calculateBlockChainPositions( handle.directoryEntry, handle.directoryEntry.numBlocks - 1, this.blockChainPositions, false );
  
      /* Is there space in the current second level pointer block for */
      /* the new data block?                                          */
      if ( ( this.blockChainPositions.offsetIntoSecondLevelPointerBlock >= this.pointersPerBlock - 1 )
        && ( this.blockChainPositions.offsetIntoFirstLevelPointerBlock  >= this.pointersPerBlock - 1 ) )
      {
        /* Both the first and second level pointer blocks are full. */
        /* We need to create new instances of each, then recurse    */
        /* and try again.                                            */
        int newFirstLevelPointerBlockNum = this.createFirstLevelBlockPointers( this.blockChainPositions.firstLevelPointerBlockNum );
        this.blockChainPositions.secondLevelPointerBlockNum = this.createSecondLevelBlockPointer( newFirstLevelPointerBlockNum );

        /* We now have space in the new second level pointer block for */
        /* another data block pointer. Add it and save the changes.    */
        this.setBlockNum(this.blockChainPositions.secondLevelPointerBlock, 0, dataBlockNum );
        this.filePointerBlockIO.writeBlock( this.blockChainPositions.secondLevelPointerBlockNum, this.blockChainPositions.secondLevelPointerBlock );
  
        /* We have added a block, increment the total block count in the */
        /* DirectoryEntry.                                               */
        handle.directoryEntry.numBlocks++;
      }
      else
      if ( this.blockChainPositions.offsetIntoSecondLevelPointerBlock >= this.pointersPerBlock - 1 )
      {
        /* The current second level pointer block is full but the first level */
        /* pointer block has space. Create a new second level pointer block   */
        /* then recurse.                                                      */
        this.blockChainPositions.secondLevelPointerBlockNum = this.createSecondLevelBlockPointer( this.blockChainPositions.firstLevelPointerBlockNum );

        /* We now have space in the new second level pointer block for */
        /* another data block pointer. Add it and save the changes.    */
        this.filePointerBlockIO.readBlock( this.blockChainPositions.secondLevelPointerBlockNum, this.blockChainPositions.secondLevelPointerBlock );
        this.setBlockNum(this.blockChainPositions.secondLevelPointerBlock, 0, dataBlockNum );
        this.filePointerBlockIO.writeBlock( this.blockChainPositions.secondLevelPointerBlockNum, this.blockChainPositions.secondLevelPointerBlock );
  
        /* We have added a block, increment the total block count in the */
        /* DirectoryEntry.                                               */
        handle.directoryEntry.numBlocks++;
      }
      else
      {
        /* We have space in the current second level pointer block for */
        /* another data block pointer. Add it and save the changes.    */
        this.setBlockNum(this.blockChainPositions.secondLevelPointerBlock, this.blockChainPositions.offsetIntoSecondLevelPointerBlock + 1, dataBlockNum );
        this.filePointerBlockIO.writeBlock( this.blockChainPositions.secondLevelPointerBlockNum, this.blockChainPositions.secondLevelPointerBlock );
  
        /* We have added a block, increment the total block count in the */
        /* DirectoryEntry.                                               */
        handle.directoryEntry.numBlocks++;
      }
    }
    else
    {
      /* We have added a block, increment the total block count in the */
      /* DirectoryEntry.                                               */
      handle.directoryEntry.numBlocks++;

      /* Assign the first data block to the first space in the second level pointer block. */
      this.calculateBlockChainPositions( handle.directoryEntry, 0, this.blockChainPositions, false );
      this.setBlockNum(this.blockChainPositions.secondLevelPointerBlock, 0, dataBlockNum );
      this.filePointerBlockIO.writeBlock( this.blockChainPositions.secondLevelPointerBlockNum, this.blockChainPositions.secondLevelPointerBlock );
    }
  }
  
  
  /**
   * Writes a block to the file whose handle is passed.
   */
  
  protected void writeBlock( DefaultSimpleFileSystemHandle handle, int relativeBlockNum, byte[] data ) throws SimpleFileSystemException, BlockIOException
  {
    this.calculateBlockChainPositions( handle.directoryEntry, relativeBlockNum, this.blockChainPositions, false );
    this.fileDataBlockIO.writeBlock( this.blockChainPositions.dataBlockNum, data );
  }
  
  
  /**
   * Reads a block to the file whose handle is passed.
   */
  
  protected void readBlock( DefaultSimpleFileSystemHandle handle, int relativeBlockNum, byte[] data ) throws SimpleFileSystemException, BlockIOException
  {
    this.calculateBlockChainPositions( handle.directoryEntry, relativeBlockNum, this.blockChainPositions, false );
    if ( this.blockChainPositions.dataBlockNum != INVALID_BLOCK_POINTER )
    {
      this.fileDataBlockIO.readBlock( this.blockChainPositions.dataBlockNum, data );
    }
    else
    {
      throw new SimpleFileSystemException( "The absolute data block number pointed to by relative block number " + relativeBlockNum + " is -1." );
    }
  }
  

  /*------------------------ SimpleFileSystem methods -------------------------*/
    
  
  /**
   * Closes this SimpleFileSystem.
   */
  
  public void close() throws SimpleFileSystemException
  {
    try
    {
      this.bmm.close();
      
      /* Write the directory entries to the file. */
      this.writeDirectory();
      
      /* Close the block caches. */
      this.fileDataBlockIO.flush();
      this.filePointerBlockIO.flush();
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new SimpleFileSystemException( bmme );
    }
    catch ( BlockIOException be )
    {
      throw new SimpleFileSystemException( be );
    }
  }


  /**
   * Opens a file. If it doesn't exist then create it.
   */

  public SimpleFileSystemHandle openFile( String fileName ) throws SimpleFileSystemException
  {
    DefaultSimpleFileSystemHandle result = new DefaultSimpleFileSystemHandle();
    
    /* If the file already exists then open it, otherwise create it. */
    DefaultSimpleFileSystemDirectoryEntry directoryEntry = ( DefaultSimpleFileSystemDirectoryEntry ) this.getDirectoryEntry( fileName );
    if ( directoryEntry == null )
    {
      /* Create a new file. */
      directoryEntry = new DefaultSimpleFileSystemDirectoryEntry();
      directoryEntry.fileName = fileName;
      directoryEntry.size = 0;
      directoryEntry.firstBlock = INVALID_BLOCK_POINTER;      

      this.directoryEntries.add( directoryEntry );
    }

    result.directoryEntry = directoryEntry;
    directoryEntry.position = 0;
    
    
    return result;
  }


  /**
   * Closes a file. We don't need to do anything here.
   */

  public void closeFile( SimpleFileSystemHandle handle ) throws SimpleFileSystemException
  {
  }
  

  /**
   * Deletes a file. If the file doesn't exist then do nothing.
   */

  public void deleteFile( String fileName ) throws SimpleFileSystemException
  {
    DefaultSimpleFileSystemDirectoryEntry de = ( DefaultSimpleFileSystemDirectoryEntry ) this.getDirectoryEntry( fileName );
    if ( de != null )
    {
      this.directoryEntries.remove( de );
      
      /* First free up the data and directory blocks in the file. */
      for ( int loop = 0; loop < de.numBlocks; loop++ )
      {
        try
        {
          this.calculateBlockChainPositions( de, loop, this.blockChainPositions, false );
          this.bmm.unlockBlock( this.blockChainPositions.firstLevelPointerBlockNum );
          this.bmm.unlockBlock( this.blockChainPositions.secondLevelPointerBlockNum );
          this.bmm.unlockBlock( this.blockChainPositions.dataBlockNum );
        }
        catch ( BlockMapManagerException bmme )
        {
          throw new SimpleFileSystemException( bmme );
        }
        catch ( BlockIOException be )
        {
          throw new SimpleFileSystemException( be );
        }
      }
    }
  }

  
  /**
   * Returns the size of the file.
   */
  
  public int size( SimpleFileSystemHandle handle ) throws SimpleFileSystemException
  {
    DefaultSimpleFileSystemHandle handle2 = ( DefaultSimpleFileSystemHandle ) handle;
    return handle2.directoryEntry.size;
  }
  
  
  /**
   * Moves the current file position to that passed.
   * 
   * Note that we don't do any real seeking of the underlying
   * file here, only update the position in the handle.
   * 
   * @param startPosition one of SEEK_START, SEEK_RELATIVE or SEEK_END.
   * @param distance the distance to seek relative to startPosition.
   */
  
  public void seek( SimpleFileSystemHandle handle, int startPosition, int distance ) throws SimpleFileSystemException
  {
    DefaultSimpleFileSystemHandle handle2 = ( DefaultSimpleFileSystemHandle ) handle;

    switch ( startPosition )
    {
      case SimpleFileSystem.SEEK_START:
        handle2.directoryEntry.position = distance;
        break;
    
      case SimpleFileSystem.SEEK_END:
        handle2.directoryEntry.position = handle2.directoryEntry.size - distance;
        break;
    
      case SimpleFileSystem.SEEK_RELATIVE:
        handle2.directoryEntry.position += distance;
        break;
        
      default:
        throw new SimpleFileSystemException( "Unknown start position " + startPosition );
    }
  }      


  /**
   * Moves the current file position to that passed.
   * 
   * Note that we don't do any real seeking of the underlying
   * file here, only update the position in the handle.
   * 
   * @param distance the distance to seek relative to startPosition.
   */
  
  public void seek( SimpleFileSystemHandle handle, int distance ) throws SimpleFileSystemException
  {
    this.seek( handle, SimpleFileSystem.SEEK_START, distance );
  }      


  /**
   * Reads data from the current position in the file.
   */
  
  public int read( SimpleFileSystemHandle handle, byte[] data ) throws SimpleFileSystemException
  {
    int numBytesRead = 0;

    if ( ( data != null ) && ( data.length > 0 ) )
    {
      DefaultSimpleFileSystemHandle handle2 = ( DefaultSimpleFileSystemHandle ) handle;
      int currentPosition = handle2.directoryEntry.position;
      int fileSize = handle2.directoryEntry.size;
      int numBlocks = handle2.directoryEntry.numBlocks;
  
      int totalBytesToRead = data.length;
      if ( totalBytesToRead > fileSize - currentPosition )
      {
        totalBytesToRead = fileSize - currentPosition;
      }
      
      /* Load the nearest block to the current position */
      /* then loop reading the data.                    */
      int blockToRead = currentPosition / this.bytesPerBlock;
      int offsetIntoBlock = currentPosition % this.bytesPerBlock;
      int numBytesToReadInBlock = this.bytesPerBlock - offsetIntoBlock;
      if ( numBytesToReadInBlock > totalBytesToRead )
      {
        numBytesToReadInBlock = totalBytesToRead;
      }

      while ( numBytesRead < totalBytesToRead )
      {
        try
        {
          this.readBlock( handle2, blockToRead, readBlock );
          
          System.arraycopy( readBlock, offsetIntoBlock, data, numBytesRead, numBytesToReadInBlock );
          numBytesRead += numBytesToReadInBlock;
          handle2.directoryEntry.position += numBytesToReadInBlock;
          
          blockToRead++;
          offsetIntoBlock = 0;
          numBytesToReadInBlock = totalBytesToRead - numBytesRead;
          if ( numBytesToReadInBlock > this.bytesPerBlock )
          {
            numBytesToReadInBlock = this.bytesPerBlock;
          }
        }
        catch ( BlockIOException be )
        {
          throw new SimpleFileSystemException( be );
        }
      }
    }
    
    return numBytesRead;
  }      


  /**
   * Write data to the current position in the file. If necessary grow
   * the size of the file to fit all the data.
   */
  
  public void write( SimpleFileSystemHandle handle, byte[] data ) throws SimpleFileSystemException
  {
    DefaultSimpleFileSystemHandle handle2 = ( DefaultSimpleFileSystemHandle ) handle;
    int currentPosition = handle2.directoryEntry.position;
    int fileSize = handle2.directoryEntry.size;
    int numBlocks = handle2.directoryEntry.numBlocks;
    
    try
    {
      /* If no first of second level pointer blocks yet exist (because */
      /* this is a brand new file) then create them now.               */
      if ( handle2.directoryEntry.firstBlock == INVALID_BLOCK_POINTER )
      {
        try
        {
          handle2.directoryEntry.firstBlock = this.createFirstLevelBlockPointers( INVALID_BLOCK_POINTER );
          this.createSecondLevelBlockPointer( handle2.directoryEntry.firstBlock );
        }
        catch ( BlockIOException be )
        {
          throw new SimpleFileSystemException( be );
        }
        catch ( BlockMapManagerException bmme )
        {
          throw new SimpleFileSystemException( bmme );
        }
      }
      
      while ( currentPosition + data.length > numBlocks * this.bytesPerBlock )
      {
        this.allocateAnotherDataBlock( handle2 );
        numBlocks = handle2.directoryEntry.numBlocks;
      }
      
      /* Load the nearest block to the current position */
      /* then loop writing the data.                    */
      int blockToRead = currentPosition / this.bytesPerBlock;
      int offsetIntoBlock = currentPosition % this.bytesPerBlock;
      int numBytesWritten = 0;
      int numBytesToWriteInBlock = this.bytesPerBlock - offsetIntoBlock;
      if ( numBytesToWriteInBlock > data.length )
      {
        numBytesToWriteInBlock = data.length;
      }

      while ( numBytesWritten < data.length )
      {
        /* If we want to write an amount smaller than one block then */
        /* we need to first read in the current block and append the */
        /* new data to it.                                           */
        if ( numBytesToWriteInBlock < this.bytesPerBlock )
        {
          this.readBlock( handle2, blockToRead, writeBlock );
        }
        
        System.arraycopy( data, numBytesWritten, writeBlock, offsetIntoBlock, numBytesToWriteInBlock );
        numBytesWritten += numBytesToWriteInBlock;
        handle2.directoryEntry.position += numBytesToWriteInBlock;
        if ( handle2.directoryEntry.position > handle2.directoryEntry.size )
        {
          handle2.directoryEntry.size = handle2.directoryEntry.position;
        }
        
        this.writeBlock( handle2, blockToRead, writeBlock );
  
        blockToRead++;
        offsetIntoBlock = 0;
        if ( data.length - numBytesWritten > this.bytesPerBlock )
          numBytesToWriteInBlock = this.bytesPerBlock;
        else
          numBytesToWriteInBlock = ( data.length - numBytesWritten );
      }
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new SimpleFileSystemException( bmme );
    }
    catch ( BlockIOException be )
    {
      throw new SimpleFileSystemException( be );
    }
  }      


  /**
   * Does a file exist?
   */
  
  public boolean doesFileExist( String fileName ) throws SimpleFileSystemException
  {
    DirectoryEntry directoryEntry = this.getDirectoryEntry( fileName );
    boolean result = ( directoryEntry != null );
    return result;
  }
}
