package au.com.explodingsheep.diskDOM.simpleFileSystem;


import java.io.*;


/**
 * File instance of the BlockIO interface. Reads and writes
 * data to a file. 
 */

public class FileBlockIO implements BlockIO
{
  /* Size of each block. */
  protected static final int BLOCK_SIZE = 512;

  /* Access to the block storage file. */
  protected RandomAccessFile file = null;

  /* Number of blocks currently in this file. */
  protected int numBlocksInFile = 0;
  
  
  /**
   * Constructor.
   */
  
  public FileBlockIO( String fileName ) throws BlockIOException
  {
    Exception e = null;
    try
    {
      this.file = new RandomAccessFile( fileName, "rw" );
      
      int fileLength = ( int ) this.file.length();
      this.numBlocksInFile = fileLength / BLOCK_SIZE;
    }
    catch ( FileNotFoundException fnfe )
    {
      e = fnfe;
    }
    catch ( IOException ioe )
    {
      e = ioe;
    }
    catch ( IllegalArgumentException iae )
    {
      e = iae;
    }
    catch ( SecurityException se )
    {
      e = se;
    }
    
    if ( e != null )
    {
      throw new BlockIOException( e.getMessage() );
    }
  }


  /*----------------------- BlockIO methods ---------------------------*/
  

  /**
   * Closes this object.
   */
  
  public void close() throws BlockIOException
  {
    /* Close the block file. */
    try
    {
      this.file.close();
    }
    catch ( IOException ioe )
    {
      throw new BlockIOException( ioe.getMessage() );
    }
  }

  
  
  /**
   * Reads data from the passed block. 
   */
  
  public void readBlock( int blockNum, byte[] data ) throws BlockIOException
  {
    try
    {
      this.file.seek( blockNum * BLOCK_SIZE );
      this.file.read( data );
    }
    catch ( IOException ioe )
    {
      ioe.printStackTrace();
      throw new BlockIOException( ioe.getMessage() );
    }
  }
  
  
  /**
   * Writes data to the passed block. If the data is larger than the
   * block size then the data will span multiple blocks.
   */
  
  public void writeBlock( int blockNum, byte[] data ) throws BlockIOException
  {
    try
    {
      this.file.seek( blockNum * BLOCK_SIZE );
      this.file.write( data );
    }
    catch ( IOException ioe )
    {
      throw new BlockIOException( ioe.getMessage() );
    }
  }


  /**
   * Returns the size of one block.
   */
  
  public int getBlockSize()
  {
    return BLOCK_SIZE;
  }


  /**
   * Returns the number of blocks currently available.
   */
  
  public int getNumBlocks()
  {
    return this.numBlocksInFile;
  }

  
  /**
   * Set the number of blocks available.
   */
  
  public void setNumBlocks( int numBlocks ) throws BlockIOException
  {
    byte[] emptyBlock = new byte[BLOCK_SIZE];
    try
    {
      this.file.seek( this.file.length() );
      while ( numBlocks > this.numBlocksInFile )
      {
        this.file.write( emptyBlock );
        this.numBlocksInFile++;
      }
    }
    catch ( IOException ioe )
    {
      throw new BlockIOException( ioe.getMessage() );
    }
  }

  
  /**
   * Flushes any cached blocks to disk.
   */
  
  public void flush() throws BlockIOException
  {
  }
}
