package au.com.explodingsheep.diskDOM.simpleFileSystem;


import java.util.*;


/**
 * A version of the BlockIO interface that stores its data
 * to RAM.
 */

public class RAMBlockIO implements BlockIO
{
  /* The size of one block. */
  protected final static int BLOCK_SIZE = 512;
  
  /* Where we store the blocks of RAM to. */
  protected byte[][] ramBlocks = null;

  /* The number of blocks we can store. This is updated by */
  /* "SetNumBlocks".                                       */
  protected int numBlocks = 0;
  
  
  /**
   * Constructor.
   */
  
  public RAMBlockIO()
  {
  }  
  

  /**
   * Returns the size of one block.
   */
  
  public int getBlockSize()
  {
    return BLOCK_SIZE;
  }
  

  /**
   * Reads data from the passed block. If the data is longer than the
   * block the next linked block will also be read until all the data
   * is read.
   */
  
  public void readBlock( int blockNum, byte[] data ) throws BlockIOException
  {
    int numBlocks = this.getNumBlocks();
    if ( blockNum < numBlocks )
    {
      if ( data.length >= BLOCK_SIZE )
      {
        byte[] block = this.ramBlocks[blockNum];
          
        System.arraycopy( block, 0, data, 0, BLOCK_SIZE );
      }
      else
      {
        throw new BlockIOException( "Passed data block to store read data in is too small." );
      }  
    }
    else
    {
      throw new BlockIOException( "Block " + blockNum + " is too large. Biggest block is " + numBlocks + "." );
    }
  }
  
  
  /**
   * Writes data to the passed block. If the data is larger than the
   * block size then the data will span multiple blocks.
   */
  
  public void writeBlock( int blockNum, byte[] data ) throws BlockIOException
  {
     int numBlocks = this.getNumBlocks();
    if ( blockNum < numBlocks )
    {
      if ( data.length >= BLOCK_SIZE )
      {
        byte[] block = this.ramBlocks[blockNum];
        
        System.arraycopy( data, 0, block, 0, BLOCK_SIZE );
      }
      else
      {
        throw new BlockIOException( "Passed data block to containing data to write is too small." );
      }  
    }
    else
    {
      throw new BlockIOException( "Block " + blockNum + " is too large. Biggest block is " + numBlocks + "." );
    }
  }
  
  
  /**
   * Closes this object.
   */
  
  public void close() throws BlockIOException
  {
  }
  
  
  /**
   * Returns the number of blocks currently available.
   */
  
  public int getNumBlocks()
  {
    return this.numBlocks;
  }
  
  
  /**
   * Set the number of blocks available.
   */
  
  public void setNumBlocks( int newNumBlocks ) throws BlockIOException
  {
    int oldNumBlocks = this.getNumBlocks();
  
    List tmpRAMBlocks = new ArrayList();
    
    if ( newNumBlocks > oldNumBlocks )
    {
      for ( int loop = 0; loop < oldNumBlocks; loop++ )
      {
        byte[] ramBlock = this.ramBlocks[loop];
        tmpRAMBlocks.add( ramBlock );
      }
    
      for ( int loop = oldNumBlocks; loop < newNumBlocks; loop++ )
      {
        byte[] ramBlock = new byte[BLOCK_SIZE];
        tmpRAMBlocks.add( ramBlock );
      }
    }
    else
    {
      for ( int loop = 0; loop < newNumBlocks; loop++ )
      {
        byte[] ramBlock = this.ramBlocks[loop];
        tmpRAMBlocks.add( ramBlock );
      }
    }
    
    this.ramBlocks = ( byte[][] ) tmpRAMBlocks.toArray( new byte[0][0] );
    
    this.numBlocks = newNumBlocks;
  }        
 
  
  /**
   * Flushes any cached blocks to disk.
   */
  
  public void flush() throws BlockIOException
  {
  }
}
