package au.com.explodingsheep.diskDOM.stringStore;


import au.com.explodingsheep.diskDOM.simpleFileSystem.*;
import au.com.explodingsheep.diskDOM.identifier.*;
import au.com.explodingsheep.diskDOM.helpers.IntToBytes;


/**
 * A StringStore that uses a SimpleFileSystem instance to store its strings.
 * We write the strings to one or more blocks. The first 4 bytes point
 * to the next block. The remaining bytes in the block hold as much
 * data as will fit. The string is terminated by a zero.
 */

public class SimpleFileSystemStringStore implements StringStore
{
  /* The size of one block. We write strings to one or more blocks. */
  protected final static int BLOCK_SIZE = 16;

  /* Means a block doesn't point to the next block. */
  protected final static int NO_NEXT_BLOCK = -1;

  /* The SimpleFileSystem we write the strings to. */
  protected SimpleFileSystem sfs = null;
  
  /* Handle to the file where we read and write blocks. */
  protected SimpleFileSystemHandle handle = null;
  
  /* Tracks which blocks are used. */
  protected BlockMapManager bmm = null;
  
  /* An attempt to speed up the code slightly. */
  protected byte[] fourBytes = new byte[4];

  
  /**
   * Constructor.
   */
  
  public SimpleFileSystemStringStore( SimpleFileSystem sfs ) throws StringStoreException
  {
    if ( sfs != null )
    {
      this.sfs = sfs;
      
      try
      {
        this.handle = this.sfs.openFile( "SimpleFileSystemStringStore.blocks" );
        this.bmm = new SimpleFileSystemBlockMapManager( sfs, "SimpleFileSystemStringStore.blockmap" );
      }
      catch ( SimpleFileSystemException sfse )
      {
        throw new StringStoreException( sfse );
      }  
      catch ( BlockMapManagerException bmme )
      {
        throw new StringStoreException( bmme );
      }  
    }
    else
    {
      throw new StringStoreException( "Passed SimpleFileSystem instance cannot be null." );
    }
  }

  
  /**
   * Turns the passed string into a byte array.
   */
  
  protected byte[] stringToByteArray( String s )
  {
    int stringLength = s.length();
    byte[] data = new byte[stringLength + 1];
    if ( s != null )
    {
      for ( int loop = 0; loop < stringLength; loop++ )
      {
        byte b = ( byte ) s.charAt( loop );
        data[loop] = b;
      }
      data[stringLength] = 0;
    }
    else
    {
      data = new byte[0];
    }
    return data;
  }
  
  
  /**
   * Turns the passed byte array into a string.
   */
  
  protected String byteArrayToString( byte[] data )
  {
    String result = "";
    if ( data != null )
    {
      int loop = 0;
      byte b = data[loop];
      while ( b != 0 )
      {
        result += ( char ) b;
        loop++;
        b = data[loop];
      }
    }
    return result;
  }


  /**
   * Allocates more blocks in the BlockIO so we have space to write 
   * things.
   */
  
  protected void allocateMoreSpace() throws StringStoreException
  {
    try
    {
      int newNumBlocks;
      int numBlocks = this.bmm.getNumBlocks();
      if ( numBlocks == 0 )
        newNumBlocks = 8;
      else
        newNumBlocks = numBlocks * 2;
      
      byte[] emptyBlock = new byte[BLOCK_SIZE];
      int fileSize = this.sfs.size( this.handle );
      this.sfs.seek( this.handle, fileSize );
      for ( int loop = numBlocks; loop < newNumBlocks; loop++ )
      {
        this.sfs.write( this.handle, emptyBlock );
      }
      
      this.bmm.setNumBlocks( newNumBlocks );
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new StringStoreException( sfse );
    }  
    catch ( BlockMapManagerException bmme )
    {
      throw new StringStoreException( bmme );
    }  
  }


  /**
   * Frees any currently allocated blocks that follow
   * the passed block that hold a string. However we don't
   * release the first block.
   */
  
  protected void releaseBlocks( int firstBlockNum ) throws StringStoreException
  {
    if ( firstBlockNum != -1 )
    {
      try
      {
        this.sfs.seek( this.handle, firstBlockNum * BLOCK_SIZE );
        this.sfs.read( this.handle, this.fourBytes );
        int nextBlockNum = IntToBytes.bytesToInt( this.fourBytes );
        while ( nextBlockNum != NO_NEXT_BLOCK )
        {
          this.sfs.seek( this.handle, nextBlockNum * BLOCK_SIZE);
          this.sfs.read( this.handle, this.fourBytes );
          this.bmm.unlockBlock( nextBlockNum );
          nextBlockNum = IntToBytes.bytesToInt( this.fourBytes );
        }
      }
      catch ( SimpleFileSystemException sfse )
      {
        throw new StringStoreException( sfse );
      }  
      catch ( BlockMapManagerException bmme )
      {
        throw new StringStoreException( bmme );
      }  
    }
  }
        

  /*-------------------------- StringStore methods ---------------------------*/
  
  
  /**
   * Closes this object.
   */
  
  public void close() throws StringStoreException
  {
    try
    {
      this.sfs.closeFile( this.handle );
      this.bmm.close();
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new StringStoreException( sfse );
    }  
    catch ( BlockMapManagerException bmme )
    {
      throw new StringStoreException( bmme );
    }  
  }
  

  /**
   * Saves a string to some persistent spot, and
   * returns a key that uniquely identifies the string.
   */
  
  public Identifier appendString( String s ) throws StringStoreException
  {
    Identifier result = null;

    try
    {
      if ( s == null )
      {
        s = "";
      }
      
      int firstBlockNum = this.bmm.getFreeBlockNum();
      if ( firstBlockNum == -1 )
      {
        this.allocateMoreSpace();
        firstBlockNum = this.bmm.getFreeBlockNum();
      }
      this.bmm.lockBlock( firstBlockNum );
          
      result = new IntIdentifier( firstBlockNum );

      /* Write a -1 into the "next block number" position of the newly */
      /* allocated block.                                               */
      byte[] emptyBlock = new byte[BLOCK_SIZE];
      IntToBytes.intToBytes( NO_NEXT_BLOCK, emptyBlock );
      this.sfs.seek( this.handle, firstBlockNum * BLOCK_SIZE );
      this.sfs.write( this.handle, emptyBlock );
      
      this.setString( result, s );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new StringStoreException( bmme );
    }  
    catch ( SimpleFileSystemException sfse )
    {
      throw new StringStoreException( sfse );
    }

    return result;
  }


  /**
   * Saves a string to some persistent spot, and
   * returns a key that uniquely identifies the string.
   */
  
  public void setString( Identifier identifier, String s ) throws StringStoreException
  {
    IntIdentifier intIdentifier = ( IntIdentifier ) identifier;
    int blockNum = intIdentifier.getValue();
    
    if ( blockNum != NO_NEXT_BLOCK )
    {
      this.releaseBlocks( blockNum );
    }
    
    try
    {
      if ( s == null )
      {
        s = "";
      }
      
      byte[] stringAsBytes = this.stringToByteArray( s );
      byte[] block = new byte[BLOCK_SIZE];
      int numBytesWritten = 0;
      while ( numBytesWritten < stringAsBytes.length )
      {
        int nextFreeBlockNum = NO_NEXT_BLOCK;
        if ( stringAsBytes.length - numBytesWritten > BLOCK_SIZE - 4 )
        {
          nextFreeBlockNum = this.bmm.getFreeBlockNum();
          if ( nextFreeBlockNum == -1 )
          {
            this.allocateMoreSpace();
            nextFreeBlockNum = this.bmm.getFreeBlockNum();
          }
          this.bmm.lockBlock( nextFreeBlockNum );
        }
        
        IntToBytes.intToBytes( nextFreeBlockNum, block );
      
        /* How many bytes should we write into the block. */
        int numBytesInBlock = BLOCK_SIZE - 4;
        if ( numBytesInBlock > stringAsBytes.length - numBytesWritten )
        {
          numBytesInBlock = stringAsBytes.length - numBytesWritten;
        }
    
        for ( int loop = 0; loop < numBytesInBlock; loop++ )
        {
          block[loop + 4] = stringAsBytes[numBytesWritten];
          numBytesWritten++;
        }
        
        this.sfs.seek( this.handle, blockNum * BLOCK_SIZE );
        this.sfs.write( this.handle, block );
        
        blockNum = nextFreeBlockNum;
      }
    }
    catch ( SimpleFileSystemException sfse )
    {
      throw new StringStoreException( sfse );
    }
    catch ( BlockMapManagerException bmme )
    {
      throw new StringStoreException( bmme );
    }  
  }


  /**
   * Retrieves a string from the persistent spot.
   */
  
  public String getString( Identifier identifier ) throws StringStoreException
  {
    String result = "";
    byte[] block = new byte[BLOCK_SIZE];
    IntIdentifier ii = ( IntIdentifier ) identifier;
    int blockNum = ii.getValue();
    while ( blockNum != NO_NEXT_BLOCK )
    {
      try
      {
        this.sfs.seek( this.handle, blockNum * BLOCK_SIZE );
        this.sfs.read( this.handle, block );

        int loop = 4;
        while ( ( loop < BLOCK_SIZE ) && ( block[loop] != 0 ) )
        {
          result += ( char ) block[loop];
          loop++;
        }
        
        blockNum = IntToBytes.bytesToInt( block );
      }        
      catch ( SimpleFileSystemException sfse )
      {
        throw new StringStoreException( sfse );
      }  
    }
    return result;
  }      
}
